@@ -1,6 +1,7 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,15 +23,6 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare
@@ -38,31 +30,31 @@
   (disable-warning var))
 
 
-#{compiler
+(private
+ compiler
   compiler-arguments process-command-line dump-nodes dump-undefined-globals
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   compiler-cleanup-hook check-global-exports disabled-warnings check-global-imports
-  foreign-type-table-size file-io-only parse-easy-ffi
+  file-io-only undefine-shadowed-macros
   unit-name insert-timer-checks used-units inline-max-size
-  debugging perform-lambda-lifting! disable-stack-overflow-checking no-c-syntax-checks
-  register-ffi-macro ffi-include-path-list ffi-dont-include
-  foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
+  debugging perform-lambda-lifting! disable-stack-overflow-checking
+  foreign-declarations emit-trace-info block-compilation line-number-database-size
   target-heap-size target-stack-size target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   target-initial-heap-size postponed-initforms
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  broken-constant-nodes inline-substitutions-enabled enable-sharp-greater-read-syntax
+  broken-constant-nodes inline-substitutions-enabled compiler-macros-enabled
   emit-profile profile-lambda-list profile-lambda-index profile-info-vector-name
-  direct-call-ids foreign-type-table first-analysis emit-closure-info emit-line-info
+  direct-call-ids foreign-type-table first-analysis emit-closure-info
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call emit-unsafe-marker
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! source-filename compressed-literals literal-compression-threshold
-  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  transform-direct-lambdas! source-filename
+  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -71,18 +63,21 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database dump-exported-globals
   default-declarations units-used-by-default words-per-flonum default-debugging-declarations
-  default-profiling-declarations default-optimization-passes compressed-literals-initializer
+  default-profiling-declarations default-optimization-passes
   inline-max-size file-requirements use-import-table lookup-exports-file
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   chop-separator chop-extension display-real-name-table display-line-number-database explicit-use-flag
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   export-list do-lambda-lifting compiler-warning export-file-name
-  foreign-argument-conversion foreign-result-conversion}
+  foreign-argument-conversion foreign-result-conversion)
 
 
 (include "tweaks")
-(include "parameters")
+
+(define-constant default-profile-name "PROFILE")
+(define-constant default-inline-max-size 10)
+(define-constant funny-message-timeout 60000)
 
 
 ;;; Compile a complete source file:
@@ -126,14 +121,14 @@
 	[hgrowth (memq 'heap-growth options)]
 	[hshrink (memq 'heap-shrinkage options)]
 	[kwstyle (memq 'keyword-style options)]
-	[lcthreshold (memq 'compress-literals options)]
+	[uses-units '()]
 	[uunit (memq 'unit options)]
 	[a-only (memq 'analyze-only options)]
 	[dynamic (memq 'dynamic options)]
 	[dumpnodes #f]
-	[ffi-mode (memq 'ffi options)]
-	[ffi-parse-mode (memq 'ffi-parse options)]
 	[quiet (memq 'quiet options)]
+	[start-time #f]
+	(upap #f)
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
 
     (define (cputime) (##sys#fudge 6))
@@ -156,7 +151,7 @@
 	(printf "(iteration ~s)~%" pass)
 	(display-analysis-database db) ) )
 
-    (define (print mode dbgmode xs)
+    (define (print-expr mode dbgmode xs)
       (when (print-header mode dbgmode)
 	(for-each pretty-print xs) ) )
 
@@ -195,18 +190,17 @@
 	(printf "milliseconds needed for ~a: \t~s~%" pass (- (cputime) time0)) ) )
 
     (define (read-form in)
-      (if (or ffi-mode ffi-parse-mode)
-	  (let ([s (read-string #f in)])
-	    (if (string=? s "")
-		#!eof
-		(cond [ffi-mode
-		       `(declare 
-			  (foreign-declare ,s)
-			  (foreign-parse ,s) ) ]
-		      [else
-		       `(declare
-			  (foreign-parse ,s) ) ] ) ) )
-	  (##sys#read in infohook) ) )
+      (##sys#read in infohook) )
+
+    (define (analyze pass node . args)
+      (let-optionals args ((no 0) (contf #t))
+        (let ((db (analyze-expression node)))
+	  (when upap
+	    (upap pass db node
+		  (cut get db <> <>)
+		  (cut put! db <> <> <>)
+		  no contf) )
+	  db) ) )
 
     (when uunit
       (set! unit-name (string->c-identifier (stringify (option-arg uunit)))) )
@@ -218,6 +212,7 @@
        (collect-options 'debug) ) )
     (set! dumpnodes (memq '|D| debugging-chicken))
     (when (memq 'lambda-lift options) (set! do-lambda-lifting #t))
+    (when (memq 'disable-compiler-macros options) (set! compiler-macros-enabled #f))
     (when (memq 't debugging-chicken) (##sys#start-timer))
     (when (memq 'b debugging-chicken) (set! time-breakdown #t))
     (and-let* ((xfile (memq 'emit-exports options)))
@@ -246,10 +241,8 @@
     (when (memq 'disable-interrupts options) (set! insert-timer-checks #f))
     (when (memq 'fixnum-arithmetic options) (set! number-type 'fixnum))
     (when (memq 'block options) (set! block-compilation #t))
-    (when (memq 'disable-c-syntax-checks options) (set! no-c-syntax-checks #t))
     (when (memq 'emit-external-prototypes-first options) (set! external-protos-first #t))
     (when (memq 'inline options) (set! inline-max-size default-inline-max-size))
-    (when (memq 'track-scheme options) (set! emit-line-info #t))
     (and-let* ([inlimit (memq 'inline-limit options)])
       (set! inline-max-size 
 	(let ([arg (option-arg inlimit)])
@@ -259,17 +252,14 @@
       (when verbose (printf "Identifiers and symbols are case insensitive~%~!"))
       (register-feature! 'case-insensitive)
       (case-sensitive #f) )
+    (when (memq 'compress-literals options)
+      (compiler-warning 'usage "`the -compress-literals' option is obsolete") )
     (when kwstyle
       (let ([val (option-arg kwstyle)])
 	(cond [(string=? "prefix" val) (keyword-style #:prefix)]
 	      [(string=? "none" val) (keyword-style #:none)]
 	      [(string=? "suffix" val) (keyword-style #:suffix)]
 	      [else (quit "invalid argument to `-keyword-style' option")] ) ) )
-    (when lcthreshold
-      (let ([t (option-arg lcthreshold)])
-	(set! literal-compression-threshold 
-	  (or (string->number t)
-	      (quit "invalid argument to `-compress-literals' option: ~A" t) ) ) ) )
     (set! verbose-mode verbose)
     (set! ##sys#read-error-with-line-number #t)
     (set! ##sys#include-pathnames
@@ -278,14 +268,18 @@
 	      ipath) )
     (when (and outfile filename (string=? outfile filename))
       (quit "source- and output-filename are the same") )
+    (set! uses-units
+      (map string->symbol 
+	   (append-map
+	    (cut string-split <> ",")
+	    (collect-options 'uses))))
+    (when (memq 'keep-shadowed-macros options)
+      (set! undefine-shadowed-macros #f) )
 
     ;; Handle feature options:
-    (for-each register-feature! (collect-options 'feature))
-
-    ;; Handle FFI defines and include-paths:
-    (for-each register-ffi-macro (collect-options 'ffi-define))
-    (when (memq 'ffi-no-include options) (set! ffi-dont-include #t))
-    (set! ffi-include-path-list (append (collect-options 'ffi-include-path) ffi-include-path-list))
+    (for-each 
+     register-feature!
+     (append-map (cut string-split <> ",") (collect-options 'feature)))
 
     ;; Load extensions:
     (set! ##sys#features (cons #:compiler-extension ##sys#features))
@@ -299,6 +293,7 @@
     (set! ##sys#features (cons '#:compiling ##sys#features))
     (set! ##sys#features (cons #:match ##sys#features))
     (##sys#provide 'match) 
+    (set! upap (user-post-analysis-pass))
 
     ;; Insert postponed initforms:
     (set! initforms (append initforms postponed-initforms))
@@ -315,13 +310,18 @@
 		     (else (quit "no filename available for `-extension' option")) ) ) ) ) ) ) )
 
     ;; Append required extensions to initforms:
-    (let ([ids (map string->symbol (collect-options 'require-extension))])
+    (let ([ids (lset-difference 
+		eq?
+		(map string->symbol
+		     (append-map
+		      (cut string-split <> ",")
+		      (collect-options 'require-extension)))
+		uses-units)])
       (set! initforms
 	(append initforms (map (lambda (r) `(##core#require-extension ',r)) ids)) ) )
 
     (when (memq 'run-time-macros options)
       (set! ##sys#enable-runtime-macros #t) )
-    (enable-sharp-greater-read-syntax)
     (set! target-heap-size
       (if hsize
 	  (arg-val (option-arg hsize))
@@ -372,7 +372,7 @@
 	  ((not filename)
 	   (unless quiet
 	     (print-version #t)
-	     (display "\n\nEnter \"chicken -help\" for information on how to use it.\n") ) )
+	     (display "\nEnter \"chicken -help\" for information on how to use it.\n") ) )
 	  (else
 
 	   ;; Display header:
@@ -383,6 +383,7 @@
 	   (debugging 'r "debugging options" debugging-chicken)
 	   (debugging 'r "target heap size" target-heap-size)
 	   (debugging 'r "target stack size" target-stack-size)
+	   (set! start-time (cputime))
 
 	   ;; Read toplevel expressions:
 	   (set! ##sys#line-number-database (make-vector line-number-database-size '()))
@@ -404,24 +405,26 @@
 			     (append (map string->expr prelude)
 				     (reverse forms)
 				     (map string->expr postlude) ) ) )
-			(let* ([f (car files)]
-			       [in (check-and-open-input-file f)] )
-			  (do ([x (read-form in) (read-form in)])
-			      ((eof-object? x) 
-			       (close-checked-input-file in f) )
-			    (set! forms (cons x forms)) ) ) ) ] ) ) )
+			(let* ((f (car files))
+			       (in (check-and-open-input-file f)) )
+			  (fluid-let ((##sys#current-source-filename f))
+			    (let ((x1 (read-form in)) )
+			      (do ((x x1 (read-form in)))
+				  ((eof-object? x) 
+				   (close-checked-input-file in f) )
+				(set! forms (cons x forms)) ) ) ) ) ) ] ) ) )
 
 	   ;; Start compilation passes:
 	   (let ([proc (user-preprocessor-pass)])
 	     (when proc
 	       (when verbose (printf "User preprocessing pass...~%~!"))
-	       (set! forms (cons (first forms) (map proc (cdr forms)))) ) )
+	       (set! forms (map proc forms))))
 
-	   (print "source" '|1| forms)
+	   (print-expr "source" '|1| forms)
 	   (begin-time)
-	   (let ([us (collect-options 'uses)])
-	     (unless (null? us)
-	       (set! forms (cons `(declare (uses ,@us)) forms)) ) )
+	   (unless (null? uses-units)
+	     (set! ##sys#explicit-library-modules (append ##sys#explicit-library-modules uses-units))
+	     (set! forms (cons `(declare (uses ,@uses-units)) forms)) )
 	   (let* ([exps0 (map canonicalize-expression (append initforms forms))]
 		  [pvec (gensym)]
 		  [plen (length profile-lambda-list)]
@@ -435,23 +438,11 @@
 				  ',(if unit-name #f profile-name))))
 			     '() )
 			 (map (lambda (pl)
-				`(##core#inline 
-				  "C_i_setslot"
+				`(##sys#set-profile-info-vector!
 				  ,profile-info-vector-name
-				  ',(* profile-info-entry-size (car pl)) 
+				  ',(car pl)
 				  ',(cdr pl) ) )
 			      profile-lambda-list)
-			 (let ([is (fold (lambda (clf r)
-					   `(let ([,(gensym) 
-						   (set! ,(car clf)
-						     (##sys#read-from-string ',(cdr clf)))])
-					      ,r) )
-					 '(##core#undefined) 
-					 compressed-literals) ] )
-			   (if compressed-literals-initializer
-			       `((##core#set! ,compressed-literals-initializer 
-					      (lambda () ,is) ) )
-			       (list is) ) )
 			 exps0
 			 (if (and (not unit-name) (not dynamic))
 			     cleanup-forms
@@ -472,11 +463,16 @@
 	     (when (and unit-name dynamic)
 	       (compiler-warning 'usage "library unit `~a' compiled in dynamic mode" unit-name) )
 
+	     (when (and unsafe (feature? 'compiling-extension))
+	       (compiler-warning 
+		'style
+		"compiling extensions in unsafe mode is bad practice and should be avoided as it may be surprising to an unsuspecting user") )
+
 	     (set! ##sys#line-number-database line-number-database-2)
 	     (set! line-number-database-2 #f)
 
 	     (end-time "canonicalization")
-	     (print "canonicalized" '|2| exps)
+	     (print-expr "canonicalized" '|2| exps)
 
 	     (when (memq 'check-syntax options) (exit))
 
@@ -493,12 +489,12 @@
 				   (canonicalize-begin-body exps) ) ) ) ] 
 		    [proc (user-pass-2)] )
 	       (when (debugging 'M "; requirements:")
-		 (pretty-print (hash-table->alist file-requirements)))
+		 (pretty-print (##sys#hash-table->alist file-requirements)))
 	       (when proc
 		 (when verbose (printf "Secondary user pass...~%"))
 		 (begin-time)
 		 (set! first-analysis #f)
-		 (let ([db (analyze-expression node0)])
+		 (let ([db (analyze 'user node0)])
 		   (print-db "analysis (u)" '|0| db 0)
 		   (end-time "pre-analysis (u)")
 		   (begin-time)
@@ -510,7 +506,7 @@
 	       (when do-lambda-lifting
 		 (begin-time)
 		 (set! first-analysis #f)
-		 (let ([db (analyze-expression node0)])
+		 (let ([db (analyze 'lift node0)])
 		   (print-db "analysis" '|0| db 0)
 		   (end-time "pre-analysis")
 		   (begin-time)
@@ -534,7 +530,7 @@
 		 (let loop ([i 1] [node2 node1] [progress #t])
 
 		   (begin-time)
-		   (let ([db (analyze-expression node2)])
+		   (let ([db (analyze 'opt node2 i progress)])
 		     (when first-analysis
 		       (when use-import-table (check-global-imports db))
 		       (check-global-exports db)
@@ -561,7 +557,7 @@
 				     (loop (add1 i) node2 #t) ]
 				    [optimize-leaf-routines
 				     (begin-time)
-				     (let ([db (analyze-expression node2)])
+				     (let ([db (analyze 'leaf node2)])
 				       (end-time "analysis")
 				       (begin-time)
 				       (let ([progress (transform-direct-lambdas! node2 db)])
@@ -576,25 +572,23 @@
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
+			      (when (and ##sys#warnings-enabled (> (- (cputime) start-time) funny-message-timeout))
+				(display "(do not worry - still compiling...)\n") )
 			      (when export-file-name
 				(dump-exported-globals db export-file-name) )
-			      (let ([upap (user-post-analysis-pass)])
-				(when upap 
-				  (upap db
-					(lambda (k p) (cut get db <> <>))
-					(lambda (k p x) (cut put! db <> <>)) ) ) )
 			      (when a-only (exit 0))
 			      (print-node "closure-converted" '|9| node3)
 
 			      (begin-time)
-			      (receive (node literals lambdas) (prepare-for-code-generation node3 db)
+			      (receive (node literals lliterals lambdas)
+				  (prepare-for-code-generation node3 db)
 				(end-time "preparation")
 
                                 (begin-time)
 				(let ((out (if outfile (open-output-file outfile) (current-output-port))) )
 				  (unless quiet
 				    (printf "generating `~A' ...~%" outfile) )
-				  (generate-code literals lambdas out filename dynamic db)
+				  (generate-code literals lliterals lambdas out filename dynamic db)
 				  (when outfile (close-output-port out)))
                                 (end-time "code generation")
                                 (when (memq 't debugging-chicken) (##sys#display-times (##sys#stop-timer)))
