@@ -1,6 +1,7 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,38 +23,27 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare (unit backend))
 
 
-;; The unspeakable namespace hack
-
-#{compiler
+(private compiler
   compiler-arguments process-command-line find-early-refs
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings 
   foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations debugging cleanup
-  foreign-type-table-size file-io-only namespace-table
+  file-io-only
   unit-name insert-timer-checks used-units inlining external-variables
-  foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
+  foreign-declarations emit-trace-info block-compilation line-number-database-size
   target-heap-size target-stack-size target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   target-initial-heap-size disable-stack-overflow-checking
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used 
-  mutable-constants emit-line-info
+  mutable-constants encode-literal
   broken-constant-nodes inline-substitutions-enabled
   direct-call-ids foreign-type-table first-analysis block-variable-literal?
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database 
@@ -64,7 +54,7 @@
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub 
   expand-foreign-lambda*
   transform-direct-lambdas! target-include-file emit-unsafe-marker
-  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -75,13 +65,12 @@
   scan-free-variables external-protos-first emit-closure-info
   topological-sort print-version print-usage initialize-analysis-database
   generate-external-variables real-name real-name2 unique-id
-  default-declarations units-used-by-default words-per-flonum
+  default-declarations units-used-by-default words-per-flonum big-fixnum?
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   default-optimization-iterations generate-foreign-callback-header generate-foreign-callback-stub-prototypes
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion quick-namespace-list setup-quick-namespace-list
-  namespace-lookup compute-namespace-size}
+  foreign-argument-conversion foreign-result-conversion)
 
 (include "tweaks")
 
@@ -103,9 +92,6 @@
    (lambda (x) (display x output))
    (intersperse lst #\space) ) )
 
-(define (compute-namespace-size n)
-  37)					; Arbitrary...
-
 
 ;;; Unique id/prefix:
 
@@ -114,31 +100,9 @@
    (sprintf "C_~X_~A_" (random #x1000000) (current-seconds)) ) )
 
 
-;;; Check name for namespace:
-;;
-;; This stuff is basically not needed. The namespace thingy once looked good, but
-;; can not replace a proper module system. It's still available, and might be
-;; handy for certain hacks.
-
-(define quick-namespace-list '())
-
-(define (setup-quick-namespace-list)
-  (for-each 
-   (lambda (ns)
-     (set! quick-namespace-list (append (cdr ns) quick-namespace-list)) )
-   namespace-table) )
-
-(define (namespace-lookup sym)
-  (and (memq sym quick-namespace-list)
-       (let loop ([nslist namespace-table] [i 0])
-	 (cond [(null? nslist) (bomb "symbol not in namespace" sym)]
-	       [(memq sym (cdar nslist)) i]
-	       [else (loop (cdr nslist) (add1 i))] ) ) ) )
-
-
 ;;; Generate target code:
 
-(define (generate-code literals lambdas out source-file dynamic db)
+(define (generate-code literals lliterals lambdas out source-file dynamic db)
   (let ()
 
     ;; Some helper procedures
@@ -167,7 +131,11 @@
 	       ((eof) (gen "C_SCHEME_END_OF_FILE"))
 	       (else (bomb "bad immediate")) ) )
 
-	    ((##core#literal) (gen "lf[" (first params) #\]))
+	    ((##core#literal) 
+	     (let ((lit (first params)))
+	       (if (vector? lit)
+		   (gen "((C_word)li" (vector-ref lit 0) ")") 
+		   (gen "lf[" (first params) #\])) ) )
 
 	    ((if)
 	     (gen #t "if(C_truep(")
@@ -297,11 +265,6 @@
 		    (empty-closure (and customizable (zero? (lambda-literal-closure-size (find-lambda call-id)))))
 		    (fn (car subs)) )
 	       (when name
-		 (when emit-line-info
-		   (match name
-		     ((file (? number? ln) _)
-		      (gen #t "#line " ln " \"" (slashify file) "\"") )
-		     (_ #f) ) )
 		 (if emit-trace-info
 		     (gen #t "C_trace(\"" (slashify name-str) "\");")
 		     (gen #t "/* " (uncommentify name-str) " */") ) )
@@ -485,9 +448,13 @@
 	    n) )
       (match (##sys#decode-seconds (current-seconds) #f)
 	[#(_ min hour mday mon year _ _ _ _)
-	  (gen "/* Generated from " source-file " by the Chicken compiler" #t
+	  (gen "/* Generated from " source-file " by the CHICKEN compiler" #t
+	       "   http://www.call-with-current-continuation.org" #t
 	       "   " (+ 1900 year) #\- (pad0 (add1 mon)) #\- (pad0 mday) #\space (pad0 hour) #\: (pad0 min) #t
-	       "   " (chicken-version #t) #t
+	       (string-intersperse
+		(map (cut string-append "   " <> "\n") 
+		     (string-split (chicken-version #t) "\n") ) 
+		"")
 	       "   command line: ")
 	  (gen-list compiler-arguments)
 	  (gen #t)
@@ -495,9 +462,6 @@
 		[else 
 		 (gen "   used units: ")
 		 (gen-list used-units) ] )
-	  (unless unit-name
-	    (gen #t "   default heap size: " default-default-target-heap-size #t)
-	    (gen "   default nursery (stack) size: " default-default-target-stack-size) )
 	  (gen #t "*/" #t #t "#include \"" target-include-file "\"")
 	  (when external-protos-first
 	    (generate-foreign-callback-stub-prototypes foreign-callback-stubs) )
@@ -519,7 +483,25 @@
 		#t "C_externimport void C_ccall C_" uu "_toplevel(C_word c,C_word d,C_word k) C_noret;"))
 	 used-units)
 	(unless (zero? n)
-	  (gen #t #t "static C_TLS C_word lf[" n"];") ) ) )
+	  (gen #t #t "static C_TLS C_word lf[" n "];") )
+	(gen #t "static double C_possibly_force_alignment;")
+	(do ((i 0 (add1 i))
+	     (llits lliterals (cdr llits)))
+	    ((null? llits))
+	  (let* ((ll (##sys#lambda-info->string (car llits)))
+		 (llen (string-length ll)))
+	    (gen #t "static C_char C_TLS li" i "[] C_aligned={C_lihdr(" 
+		 (arithmetic-shift llen -16) #\,
+		 (bitwise-and #xff (arithmetic-shift llen -8)) #\,
+		 (bitwise-and #xff llen)
+		 #\))
+	    (do ((n 0 (add1 n)))
+		((>= n llen))
+	      (gen #\, (char->integer (string-ref ll n))) )
+	    (do ((n (- (bitwise-and #xfffff8 (+ llen 7)) llen) (sub1 n))) ; fill up with zeros to align following entry
+		((zero? n))
+	      (gen ",0") )
+	    (gen "};")))))
   
     (define (prototypes)
       (let ([large-signatures '()])
@@ -574,13 +556,13 @@
 			  (gen ",C_word t" (+ n 1) ") C_noret;") ) ) ]
 		   [else 
 		    (gen #\))
-					;(when customizable (gen " C_c_regparm"))
+		    ;;(when customizable (gen " C_c_regparm"))
 		    (unless direct (gen " C_noret"))
 		    (gen #\;) ] ) ) )
 	 lambdas) 
 	(for-each
 	 (lambda (s)
-	   (gen #t "typedef void (C_cdecl *C_proc" s ")(C_word")
+	   (gen #t "typedef void (*C_proc" s ")(C_word")
 	   (for-each gen (make-list s ",C_word"))
 	   (gen ") C_noret;") )
 	 large-signatures) ) )
@@ -659,10 +641,10 @@
 	(for-each (emitter #t) nsrv) ) )
   
     (define (literal-frame)
-      (do ([i 0 (+ i 1)]
+      (do ([i 0 (add1 i)]
 	   [lits literals (cdr lits)] )
 	  ((null? lits))
-	(gen-lit (car lits) (sprintf "lf[~s]" i) #t) ) )
+	(gen-lit (car lits) (sprintf "lf[~s]" i)) ) )
 
     (define (bad-literal lit)
       (bomb "type of literal not supported" lit) )
@@ -671,16 +653,13 @@
       (cond [(immediate? lit) 0]
 	    [(string? lit) 0]
 	    [(number? lit) words-per-flonum]
-	    [(symbol? lit) (fx+ (if (##sys#fudge 31) 1 0) 9)]		; size of symbol, and possibly a bucket
+	    [(symbol? lit) 10]		; size of symbol, and possibly a bucket
 	    [(pair? lit) (+ 3 (literal-size (car lit)) (literal-size (cdr lit)))]
 	    [(vector? lit) (+ 1 (vector-length lit) (reduce + 0 (map literal-size (vector->list lit))))]
 	    [(block-variable-literal? lit) 0]
 	    [(##sys#immediate? lit) (bad-literal lit)]
 	    [(##core#inline "C_lambdainfop" lit) 0]
-	    [(##sys#bytevector? lit)
-	     (if (##sys#permanent? lit)
-		 0
-		 (+ 2 (words (##sys#size lit))) ) ]
+	    [(##sys#bytevector? lit) (+ 2 (words (##sys#size lit))) ] ; drops "permanent" property!
 	    [(##sys#generic-structure? lit)
 	     (let ([n (##sys#size lit)])
 	       (let loop ([i 0] [s (+ 2 n)])
@@ -689,118 +668,50 @@
 		     (loop (add1 i) (+ s (literal-size (##sys#slot lit i)))) ) ) ) ]
 	    [else (bad-literal lit)] ) )
 
-    ;; This is currently not needed but will be handy for optimized literal lists/vector constructors...
-    (define (imm-lit lit)
-      (cond [(fixnum? lit) (string-append "C_fix(" (number->string lit) ")")]
-	    [(eq? #t lit) "C_SCHEME_TRUE"]
-	    [(eq? #f lit) "C_SCHEME_FALSE"]
-	    [(null? lit) "C_SCHEME_END_OF_LIST"]
-	    [(eq? lit (void)) "C_SCHEME_UNDEFINED"]
-	    [(char? lit) (string-append "C_make_character(" (number->string (char->integer lit)) ")")]
-	    [(eof-object? lit) "C_SCHEME_END_OF_FILE"]
-	    [(and (number? lit) (eq? 'fixnum number-type))
-	     (let ([flit (##sys#flo2fix lit)])
-	       (compiler-warning 'type "coerced inexact literal number `~S' to fixnum `~S'" lit flit)
-	       (string-append "C_fix(" (number->string flit) ")") ) ]
-	    [else #f] ) )
-
-    (define (gen-lit lit to lf)
-      (cond ((fixnum? lit)
-	     (if (eq? 'flonum number-type)
-		 (gen #t to "=C_flonum(C_heaptop," lit ");")
-		 (gen #t to "=C_fix(" lit ");") ) )
+    (define (gen-lit lit to)
+      ;; we do simple immediate literals directly to avoid a function call:
+      (cond ((and (fixnum? lit) (not (big-fixnum? lit)))
+	     (gen #t to "=C_fix(" lit ");") )
 	    ((block-variable-literal? lit))
 	    ((eq? lit (void))
 	     (gen #t to "=C_SCHEME_UNDEFINED;") )
-	    ((number? lit)
-	     (cond [(eq? 'fixnum number-type)
-		    (let ([flit (##sys#flo2fix lit)])
-		      (compiler-warning 'type "coerced inexact literal number `~S' to fixnum `~S'" lit flit)
-		      (gen #t to "=C_fix(" flit ");") ) ]
-		   [else
-		    (let ((str (number->string lit)))
-		      (if (and (> (string-length str) 1) (char-alphabetic? (string-ref str 1)))	; inf or nan?
-			  (gen #t to "=C_flonum(C_heaptop, C_strtod(\"" lit "\", NULL));")
-			  (gen #t to "=C_flonum(C_heaptop," lit ");"))) ] ) )
 	    ((boolean? lit) 
 	     (gen #t to #\= (if lit "C_SCHEME_TRUE" "C_SCHEME_FALSE") #\;) )
 	    ((char? lit)
 	     (gen #t to "=C_make_character(" (char->integer lit) ");") )
-	    ((null? lit) 
-	     (gen #t to "=C_SCHEME_END_OF_LIST;") )
-	    ((string? lit) (gen-string-like-lit to lit "C_static_string" #t))
-	    ((##core#inline "C_lambdainfop" lit) (gen-string-like-lit to lit "C_static_lambda_info" #t))
-	    ((pair? lit)
-	     (cond ((and (proper-list? lit) (pair? (cdr lit)))
-		    (do ((len 0 (add1 len))
-			 (lst lit (cdr lst)) )
-			((null? lst)
-			 (gen #t to "=C_h_list(" len)
-			 (do ((k (sub1 len) (sub1 k)))
-			     ((< k 0) (gen ");" #t "C_drop(" len ");"))
-			   (gen ",C_pick(" k #\)) ) )
-		      (gen-lit (car lst) "tmp" #f)
-		      (gen #t "C_save(tmp);") ) )
-		   (else
-		    (gen-lit (car lit) "tmp" #f)
-		    (gen #t "C_save(tmp);")
-		    (gen-lit (cdr lit) "tmp" #f)
-		    (gen #t to "=C_h_pair(C_restore,tmp);") ) ) )
-	    ((vector? lit) (gen-vector-like-lit to lit "C_h_vector"))
-	    ((symbol? lit)
+	    ((symbol? lit)		; handled slightly specially (see C_h_intern_in)
 	     (let* ([str (##sys#slot lit 1)]
 		    [cstr (c-ify-string str)]
-		    [len (##sys#size str)] 
-		    [nsi (namespace-lookup lit)] )
+		    [len (##sys#size str)] )
 	       (gen #t to "=")
-	       (if nsi
-		   (if lf
-		       (gen "C_h_intern_in(&" to #\, len #\, cstr ",stable" nsi ");")
-		       (gen "C_intern_in(C_heaptop," len #\, cstr ",stable" nsi ");") )
-		   (if lf
-		       (gen "C_h_intern(&" to #\, len #\, cstr ");")
-		       (gen "C_intern(C_heaptop," len #\, cstr ");") ) ) ) )
-	    ((##sys#immediate? lit) (bad-literal lit))
-	    ((##sys#bytevector? lit)
-	     (if (##sys#permanent? lit)
-		 (gen-string-like-lit to lit "C_pbytevector" #f)
-		 (gen-string-like-lit to lit "C_bytevector" #t) ) )
-	    ((##sys#generic-structure? lit) (gen-vector-like-lit to lit "C_h_structure"))
-	    (else (bad-literal lit)) ) )
+	       (gen "C_h_intern(&" to #\, len #\, cstr ");") ) )
+	    ((null? lit) 
+	     (gen #t to "=C_SCHEME_END_OF_LIST;") )
+	    ((and (not (##sys#immediate? lit))
+		  (##core#inline "C_lambdainfop" lit)))
+	    ((or (fixnum? lit) (not (##sys#immediate? lit)))
+	     (gen #t to "=C_decode_literal(C_heaptop,")
+	     (gen-string-constant (encode-literal lit))
+	     (gen ");") )
+	    (else (bad-literal lit))))
 
-    (define (gen-string-like-lit to lit conser top)
-      (let* ([len (##sys#size lit)]
+    (define (gen-string-constant str)
+      (let* ([len (##sys#size str)]
 	     [ns (fx/ len 80)]
 	     [srest (modulo len 80)] )
-	(gen #t to #\= conser #\()
-	(when top (gen "C_heaptop,"))
-	(gen len #\,)
 	(do ([i ns (sub1 i)]
 	     [offset 0 (+ offset 80)] )
 	    ((zero? i)
 	     (when (or (zero? len) (not (zero? srest)))
-	       (gen (c-ify-string (string-like-substring lit offset len))) )
-	     (gen ");") )
-	  (gen (c-ify-string (string-like-substring lit offset (+ offset 80))) #t) ) ) )
-  
+	       (gen (c-ify-string (string-like-substring str offset len))) ) )
+	  (gen (c-ify-string (string-like-substring str offset (+ offset 80))) #t) ) ) )
+ 
     (define (string-like-substring s start end)
       (let* ([len (- end start)]
 	     [s2 (make-string len)] )
 	(##sys#copy-bytes s s2 start 0 len)
 	s2) )
 
-    (define (gen-vector-like-lit to lit conser)
-      (let ([len (##sys#size lit)])
-	(do ([j 0 (+ j 1)]
-	     [n len (- n 1)] )
-	    ((zero? n)
-	     (gen #t to #\= conser #\( len)
-	     (do ([j (- len 1) (- j 1)])
-		 ((< j 0) (gen ");" #t "C_drop(" len ");"))
-	       (gen ",C_pick(" j #\)) ) )
-	  (gen-lit (##sys#slot lit j) "tmp" #f)
-	  (gen #t "C_save(tmp);") ) ) )
-
     (define (procedures)
       (for-each
        (lambda (ll)
@@ -860,20 +771,11 @@
 		   ((zero? j))
 		 (gen #t "C_word t" i #\;) ) )
 	   (cond [(eq? 'toplevel id) 
-		  (do ([i 0 (add1 i)]
-		       [ns namespace-table (cdr ns)] )
-		      ((null? ns))
-		    (gen #t "C_SYMBOL_TABLE *stable" i #\;) )
 		  (let ([ldemand (fold (lambda (lit n) (+ n (literal-size lit))) 0 literals)]
 			[llen (length literals)] )
 		    (gen #t "C_word *a;"
 			 #t "if(toplevel_initialized) C_kontinue(t1,C_SCHEME_UNDEFINED);"
 			 #t "else C_toplevel_entry(C_text(\"" topname "\"));")
-		    (do ([i 0 (add1 i)]
-			 [ns namespace-table (cdr ns)] )
-			((null? ns))
-		      (gen #t "stable" i "=C_new_symbol_table(\""
-			   (caar ns) "\"," (compute-namespace-size (cdar ns)) ");") )
 		    (when disable-stack-overflow-checking
 		      (gen #t "C_disable_overflow_check=1;") )
 		    (unless unit-name
@@ -998,13 +900,12 @@
     (prototypes)
     (generate-foreign-callback-stubs foreign-callback-stubs db)
     (trampolines)
-    (setup-quick-namespace-list)
     (procedures)
     (emit-procedure-table-info lambdas source-file)
     (trailer) ) )
 
 
-;;; Emit prrocedure table:
+;;; Emit procedure table:
 
 (define (emit-procedure-table-info lambdas sf)
   (gen #t #t "#ifdef C_ENABLE_PTABLES"
@@ -1100,7 +1001,9 @@
        (gen #t)
        (when rname
 	 (gen #t "/* from " (cleanup rname) " */") )
-       (when body (gen #t "#define return(x) C_cblock C_r = (" rconv "(x))); goto C_return; C_cblockend"))
+       (when body
+	 (gen #t "#define return(x) C_cblock C_r = (" rconv 
+	      "(x))); goto C_ret; C_cblockend"))
        (if cps
 	   (gen #t "C_noret_decl(" id ")"
 		#t "static void C_ccall " id "(C_word C_c,C_word C_self,C_word C_k,")
@@ -1121,13 +1024,13 @@
 	       "=(" (foreign-type-declaration type "") #\)
 	       (foreign-argument-conversion type) "C_a" index ");") )
 	types (iota n) names)
-       (when callback (gen #t "int C_dummy=C_save_callback_continuation(&C_a,C_k);"))
+       (when callback (gen #t "int C_level=C_save_callback_continuation(&C_a,C_k);"))
        (cond [body
 	      (gen #t body
-		   #t "C_return:")
+		   #t "C_ret:")
 	      (gen #t "#undef return" #t)
 	      (cond [callback
-		     (gen #t "C_k=C_restore_callback_continuation();"
+		     (gen #t "C_k=C_restore_callback_continuation2(C_level);"
 			  #t "C_kontinue(C_k,C_r);") ]
 		    [cps (gen #t "C_kontinue(C_k,C_r);")]
 		    [else (gen #t "return C_r;")] ) ]
@@ -1140,7 +1043,7 @@
 	      (unless (eq? rtype 'void) (gen #\)))
 	      (gen ");")
 	      (cond [callback
-		     (gen #t "C_k=C_restore_callback_continuation();"
+		     (gen #t "C_k=C_restore_callback_continuation2(C_level);"
 			  #t "C_kontinue(C_k,C_r);") ]
 		    [cps (gen "C_kontinue(C_k,C_r);")]
 		    [else (gen #t "return C_r;")] ) ] )
@@ -1163,11 +1066,11 @@
 		  byte unsigned-byte)
 	    ns]
 	   [(float double c-pointer unsigned-integer unsigned-integer32 long integer integer32 unsigned-long 
-		   nonnull-c-pointer number integer64)
+		   nonnull-c-pointer number integer64 c-string-list c-string-list*)
 	    (string-append ns "+3") ]
-	   [(c-string c-string*)
+	   [(c-string c-string* unsigned-c-string unsigned-c-string unsigned-c-string*)
 	    (string-append ns "+2+(" var "==NULL?1:C_bytestowords(C_strlen(" var ")))") ]
-	   [(nonnull-c-string nonnull-c-string* symbol)
+	   [(nonnull-c-string nonnull-c-string* nonnull-unsigned-c-string nonnull-unsigned-c-string* symbol)
 	    (string-append ns "+2+C_bytestowords(C_strlen(" var "))") ]
 	   [else
 	    (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type)) 
@@ -1187,8 +1090,8 @@
 	 (when rname
 	   (gen #t "/* from " (cleanup rname) " */") )
 	 (generate-foreign-callback-header "" stub)
-	 (gen #\{ #t "C_word x, *a=C_alloc(" sizestr ");")
-	 (gen #t "C_callback_adjust_stack_limits(a);")
+	 (gen #\{ #t "C_word x,s=" sizestr ",*a=C_alloc(s);")
+	 (gen #t "C_callback_adjust_stack(a,s);") ; make sure content is below stack_bottom as well
 	 (for-each
 	  (lambda (v t)
 	    (gen #t "x=" (foreign-result-conversion t "a") v ");"
@@ -1225,8 +1128,8 @@
 	[str (lambda (ts) (string-append ts " " target))] )
     (case type
       [(scheme-object) (str "C_word")]
-      [(char byte) (str "char")]
-      [(unsigned-char unsigned-byte) (str "unsigned char")]
+      [(char byte) (str "C_char")]
+      [(unsigned-char unsigned-byte) (str "unsigned C_char")]
       [(unsigned-int unsigned-integer) (str "unsigned int")]
       [(unsigned-int32 unsigned-integer32) (str "C_u32")]
       [(int integer bool) (str "int")]
@@ -1239,9 +1142,12 @@
       [(float) (str "float")]
       [(double number) (str "double")]
       ;; pointer and nonnull-pointer are DEPRECATED
-      [(pointer nonnull-pointer c-pointer nonnull-c-pointer scheme-pointer nonnull-scheme-pointer)
-       (str "void *")]
-      [(byte-vector nonnull-byte-vector u8vector nonnull-u8vector) (str "unsigned char *")]
+      [(pointer nonnull-pointer) (str "void *")]
+      [(c-pointer nonnull-c-pointer scheme-pointer nonnull-scheme-pointer) (str "void *")]
+      [(c-string-list c-string-list*) "C_char **"]
+      ;; byte-vector and nonnull-byte-vector are DEPRECATED
+      [(byte-vector nonnull-byte-vector) (str "unsigned char *")]
+      [(blob nonnull-blob u8vector nonnull-u8vector) (str "unsigned char *")]
       [(u16vector nonnull-u16vector) (str "unsigned short *")]
       [(s8vector nonnull-s8vector) (str "char *")]
       [(u32vector nonnull-u32vector) (str "unsigned int *")]
@@ -1249,7 +1155,9 @@
       [(s32vector nonnull-s32vector) (str "int *")]
       [(f32vector nonnull-f32vector) (str "float *")]
       [(f64vector nonnull-f64vector) (str "double *")]
-      [(nonnull-c-string c-string nonnull-c-string* c-string* symbol) (str "char *")]
+      [(nonnull-c-string c-string nonnull-c-string* c-string* 
+			 nonnull-unsigned-c-string nonnull-unsigned-c-string* unsigned-c-string*
+			 symbol) (str "char *")]
       [(void) (str "void")]
       [else
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
@@ -1258,7 +1166,7 @@
 	     [(string? type) (str type)]
 	     [(pair? type)
 	      (match type
-		[((or 'pointer 'c-pointer 'nonnull-pointer 'nonnull-c-pointer) ptype)
+		[((or 'pointer 'nonnull-pointer 'c-pointer 'nonnull-c-pointer) ptype)
 		 (foreign-type-declaration ptype (string-append "*" target)) ]
 		[('ref rtype)
 		 (foreign-type-declaration rtype (string-append "&" target)) ]
@@ -1312,10 +1220,16 @@
       ((integer64) "C_num_to_int64(")
       ((long) "C_num_to_long(")
       ((unsigned-integer unsigned-integer32) "C_num_to_unsigned_int(")
-      ((pointer scheme-pointer) "C_data_pointer_or_null(") ; pointer is DEPRECATED
-      ((nonnull-pointer nonnull-scheme-pointer) "C_data_pointer(") ; nonnull-pointer is DEPRECATED
+      ;; pointer and nonnull-pointer are DEPRECATED
+      ((pointer) "C_data_pointer_or_null(")
+      ((nonnull-pointer) "C_data_pointer(")
+      ((scheme-pointer) "C_data_pointer_or_null(")
+      ((nonnull-scheme-pointer) "C_data_pointer(")
       ((c-pointer) "C_c_pointer_or_null(")
       ((nonnull-c-pointer) "C_c_pointer_nn(")
+      ((blob) "C_c_bytevector_or_null(")
+      ((nonnull-blob) "C_c_bytevector(")
+      ;; byte-vector and nonnull-byte-vector are DEPRECATED
       ((byte-vector) "C_c_bytevector_or_null(")
       ((nonnull-byte-vector) "C_c_bytevector(")
       ((u8vector) "C_c_u8vector_or_null(")
@@ -1334,8 +1248,9 @@
       ((nonnull-f32vector) "C_c_f32vector(")
       ((f64vector) "C_c_f64vector_or_null(")
       ((nonnull-f64vector) "C_c_f64vector(")
-      ((c-string c-string*) "C_string_or_null(")
-      ((nonnull-c-string nonnull-c-string* symbol) "C_c_string(")
+      ((c-string c-string* unsigned-c-string unsigned-c-string*) "C_string_or_null(")
+      ((nonnull-c-string nonnull-c-string* nonnull-unsigned-c-string 
+			 nonnull-unsigned-c-string* symbol) "C_c_string(")
       ((bool) "C_truep(")
       (else
        (cond [(and (symbol? type) (##sys#hash-table-ref foreign-type-table type))
@@ -1343,8 +1258,11 @@
 		   (foreign-argument-conversion (if (vector? t) (vector-ref t 0) t)) ) ]
 	     [(pair? type)
 	      (match type
-		[((or 'pointer 'c-pointer) ptype) "C_c_pointer_or_null("]
-		[((or 'nonnull-pointer 'nonnull-c-pointer) ptype) "C_c_pointer_nn("]
+                ;; pointer and nonnull-pointer are DEPRECATED
+		[('pointer ptype) "C_c_pointer_or_null("]
+		[('nonnull-pointer ptype) "C_c_pointer_nn("]
+		[('c-pointer ptype) "C_c_pointer_or_null("]
+		[('nonnull-c-pointer ptype) "C_c_pointer_nn("]
 		[`(instance ,cname ,sname) "C_c_pointer_or_null("]
 		[`(nonnull-instance ,cname ,sname) "C_c_pointer_nn("]
 		[`(function ,rtype ,@argtypes) "C_c_pointer_or_null("]
@@ -1368,12 +1286,15 @@
       ((unsigned-short) "C_fix(0xffff&(C_word)")
       ((byte) "C_fix((char)")
       ((unsigned-byte) "C_fix(0xff&(C_word)")
-      ((float double integer64) (sprintf "C_flonum(&~a," dest))	;*** suboptimal for int64
+      ((float double) (sprintf "C_flonum(&~a," dest))	;*** suboptimal for int64
       ((number) (sprintf "C_number(&~a," dest))
-      ((nonnull-c-string c-string nonnull-c-pointer c-string* nonnull-c-string* symbol) 
+      ((nonnull-c-string c-string nonnull-c-pointer c-string* nonnull-c-string* 
+			 unsigned-c-string unsigned-c-string* nonnull-unsigned-c-string
+			 nonnull-unsigned-c-string* symbol c-string-list c-string-list*) 
        (sprintf "C_mpointer(&~a,(void*)" dest) )
       ((c-pointer) (sprintf "C_mpointer_or_false(&~a,(void*)" dest))
       ((integer integer32) (sprintf "C_int_to_num(&~a," dest))
+      ((integer64) (sprintf "C_a_double_to_num(&~a," dest))
       ((unsigned-integer unsigned-integer32) (sprintf "C_unsigned_int_to_num(&~a," dest))
       ((long) (sprintf "C_long_to_num(&~a," dest))
       ((unsigned-long) (sprintf "C_unsigned_long_to_num(&~a," dest))
@@ -1402,3 +1323,69 @@
 		[`(enum ,etype) (sprintf "C_int_to_num(&~a," dest)]
 		[else (err)] ) ]
 	     [else (err)] ) ) ) ) )
+
+
+;;; Encoded literals as strings, to be decoded by "C_decode_literal()"
+;; 
+;; - everything hardcoded, using the FFI would be the ugly, but safer method.
+
+(define (encode-literal lit)
+  (define getbits
+    (foreign-lambda* int ((scheme-object lit))
+      "
+#ifdef C_SIXTY_FOUR
+return((C_header_bits(lit) >> (24 + 32)) & 0xff);
+#else
+return((C_header_bits(lit) >> 24) & 0xff);
+#endif
+") )
+  (define getsize
+    (foreign-lambda* int ((scheme-object lit))
+      "return(C_header_size(lit));"))
+  (define (encode-size n)
+    ;; only handles sizes in the 24-bit range!
+    (string (integer->char (bitwise-and #xff (arithmetic-shift n -16)))
+	    (integer->char (bitwise-and #xff (arithmetic-shift n -8)))
+	    (integer->char (bitwise-and #xff n))))
+  (define (finish str)		   ; can be taken out at a later stage
+    (string-append (string #\xfe) str))
+  (finish
+   (cond ((eq? #t lit) "\xff\x06\x01")
+	 ((eq? #f lit) "\xff\x06\x00")
+	 ((char? lit) (string-append "\xff\x0a" (encode-size (char->integer lit))))
+	 ((null? lit) "\xff\x0e")
+	 ((eof-object? lit) "\xff\x3e")
+	 ((eq? (void) lit) "\xff\x1e")
+	 ((fixnum? lit)
+	  (if (not (big-fixnum? lit))
+	      (string-append
+	       "\xff\x01"
+	       (string (integer->char (bitwise-and #xff (arithmetic-shift lit -24)))
+		       (integer->char (bitwise-and #xff (arithmetic-shift lit -16)))
+		       (integer->char (bitwise-and #xff (arithmetic-shift lit -8)))
+		       (integer->char (bitwise-and #xff lit)) ) )
+	      (string-append "\xff\x55" (number->string lit) "\x00") ) )
+	 ((number? lit)
+	  (string-append "\x55" (number->string lit) "\x00") )
+	 ((symbol? lit)
+	  (let ((str (##sys#slot lit 1)))
+	    (string-append 
+	     "\x01" 
+	     (encode-size (string-length str))
+	     str) ) )
+	 ((##sys#immediate? lit)
+	  (bomb "invalid literal - can not encode" lit))
+	 ((##core#inline "C_byteblockp" lit)
+	  (##sys#string-append ; relies on the fact that ##sys#string-append doesn't check
+	   (string-append
+	    (string (integer->char (getbits lit)))
+	    (encode-size (getsize lit)) )
+	   lit) )
+	 (else
+	  (let ((len (getsize lit)))
+	    (string-intersperse
+	     (cons*
+	      (string (integer->char (getbits lit)))
+	      (encode-size len)
+	      (list-tabulate len (lambda (i) (encode-literal (##sys#slot lit i)))))
+	     ""))))) )
