@@ -1,6 +1,7 @@
 ;;;; chicken-more-macros.scm - More syntax extensions
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,15 +23,6 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (##sys#provide 'chicken-more-macros)
@@ -64,17 +56,17 @@
 		     `(begin
 			(define ,setr
 			  (lambda (x val)
-			    (##sys#check-structure x ',name)
+			    (##core#check (##sys#check-structure x ',name))
 			    (##sys#block-set! x ,i val) ) )
 			(define ,getr
 			  ,(if setters
 			       `(getter-with-setter
 				 (lambda (x) 
-				   (##sys#check-structure x ',name)
+				   (##core#check (##sys#check-structure x ',name))
 				   (##sys#block-ref x ,i) )
 				 ,setr)
 			       `(lambda (x)
-				  (##sys#check-structure x ',name)
+				  (##core#check (##sys#check-structure x ',name))
 				  (##sys#block-ref x ,i) ) ) ) )
 		     (mapslots (##sys#slot slots 1) (fx+ i 1)) ) ) ) ) ) ) ) ) )
 
@@ -86,9 +78,12 @@
        (begin
 	 (##sys#check-syntax 'receive vars 'lambda-list)
 	 (##sys#check-syntax 'receive rest '(_ . _))
-	 `(##sys#call-with-values 
-	   (lambda () ,(car rest))
-	   (lambda ,vars ,@(cdr rest)) ) ) ) ) )
+	 (if (and (pair? vars) (null? (cdr vars)))
+	     `(let ((,(car vars) ,(car rest)))
+		,@(cdr rest))
+	     `(##sys#call-with-values 
+	       (lambda () ,(car rest))
+	       (lambda ,vars ,@(cdr rest)) ) ) ) ) ) )
 
 (##sys#register-macro
  'time 
@@ -114,13 +109,16 @@
        [read read]
        [reverse reverse] )
    (lambda (filename)
-     `(begin
-	,@(with-input-from-file (##sys#resolve-include-filename filename #t)
-	    (lambda ()
-	      (do ([x (read) (read)]
-		   [xs '() (cons x xs)] )
-		  ((eof-object? x) 
-		   (reverse xs))) ) ) ) ) ) )
+     (let ((path (##sys#resolve-include-filename filename #t)))
+       (when (load-verbose) (print "; including " path " ..."))
+       `(begin
+	  ,@(with-input-from-file path
+	      (lambda ()
+		(fluid-let ((##sys#current-source-filename path))
+		  (do ([x (read) (read)]
+		       [xs '() (cons x xs)] )
+		      ((eof-object? x) 
+		       (reverse xs))) ) ) ) ) ) ) ) )
 
 (##sys#register-macro
  'assert
@@ -166,13 +164,15 @@
 		,@(map (lambda (ot id) `(##core#set! ,ot ,id))
 		       old-tmps ids)
 		,@(map (lambda (id nt) `(##core#set! ,id ,nt))
-		       ids new-tmps))
+		       ids new-tmps)
+		(##sys#void) )
 	      (lambda () ,@body)
 	      (lambda ()
 		,@(map (lambda (nt id) `(##core#set! ,nt ,id))
 		       new-tmps ids)
 		,@(map (lambda (id ot) `(##core#set! ,id ,ot))
-		       ids old-tmps) ) ) ) ) ) ) )
+		       ids old-tmps)
+		(##sys#void) ) ) ) ) ) ) )
 
 (##sys#register-macro
  'eval-when
@@ -230,17 +230,21 @@
  (lambda (test . body)
    `(if ,test (##core#undefined) (begin ,@body)) ) )
 
-(let* ([map map]
-       [assign
+(let* ((map map)
+       (assign
 	(lambda (vars exp)
 	  (##sys#check-syntax 'set!-values/define-values vars '#(symbol 0))
-	  (if (null? vars)
-	      `(##sys#call-with-values (lambda () ,exp) (lambda () (##core#undefined)))
-	      (let ([aliases (map gensym vars)])
-		`(##sys#call-with-values
-		  (lambda () ,exp)
-		  (lambda ,aliases
-		    ,@(map (lambda (v a) `(##core#set! ,v ,a)) vars aliases) ) ) ) ) ) ] )
+	  (cond ((null? vars)
+		 ;; may this be simply "exp"?
+		 `(##sys#call-with-values (lambda () ,exp) (lambda () (##core#undefined))) )
+		((null? (cdr vars))
+		 `(##core#set! ,(car vars) ,exp)) 
+		(else
+		 (let ([aliases (map gensym vars)])
+		   `(##sys#call-with-values
+		     (lambda () ,exp)
+		     (lambda ,aliases
+		       ,@(map (lambda (v a) `(##core#set! ,v ,a)) vars aliases) ) ) ) ) ) ) ) )
   (##sys#register-macro 'set!-values assign)
   (##sys#register-macro 'define-values assign) )
 
@@ -282,10 +286,15 @@
        (let fold ([llists llists]
 		  [exps (map (lambda (x) (cadr x)) vbindings)]
 		  [llists2 llists2] )
-	 (if (null? llists)
-	     `(let ,(map (lambda (v) (##sys#list v (lookup v))) vars) ,@body)
-	     `(##sys#call-with-values (lambda () ,(car exps))
-				      (lambda ,(car llists2) ,(fold (cdr llists) (cdr exps) (cdr llists2))) ) ) ) ) ) ) )
+	 (cond ((null? llists)
+		`(let ,(map (lambda (v) (##sys#list v (lookup v))) vars) ,@body) )
+	       ((and (pair? (car llists2)) (null? (cdar llists2)))
+		`(let ((,(caar llists2) ,(car exps)))
+		   ,(fold (cdr llists) (cdr exps) (cdr llists2)) ) )
+	       (else
+		`(##sys#call-with-values
+		  (lambda () ,(car exps))
+		  (lambda ,(car llists2) ,(fold (cdr llists) (cdr exps) (cdr llists2))) ) ) ) ) ) ) ) )
 
 (##sys#register-macro-2
  'let*-values
@@ -346,14 +355,6 @@
    (##sys#check-syntax 'define-constant form '(symbol _))
    `(##core#define-constant ',(car form) ,(cadr form)) ) )
 
-(##sys#register-macro-2			; DEPRECATED
- 'critical-section
- (lambda (form)
-   `(##sys#dynamic-wind
-	##sys#disable-interrupts
-	(lambda () ,@form)
-	##sys#enable-interrupts) ) )
-
 (##sys#register-macro-2
  'and-let*
    (lambda (forms)
@@ -375,54 +376,27 @@
 			       (if ,var ,(fold bs2) #f) ) ) ] ) ) ) ) ) ) ) )
 
 (##sys#register-macro-2
- 'cond-expand
-   (lambda (clauses)
+ 'select
+ (let ((gensym gensym))
+   (lambda (form)
+     (let ((exp (car form))
+	   (body (cdr form)) )
+       (let ((tmp (gensym)))
+	 `(let ((,tmp ,exp))
+	    ,(let expand ((clauses body))
+	       (if (not (pair? clauses))
+		   '(##core#undefined)
+		   (let ((clause (##sys#slot clauses 0))
+			 (rclauses (##sys#slot clauses 1)) )
+		     (##sys#check-syntax 'select clause '#(_ 1))
+		     (if (eq? 'else (car clause))
+			 `(begin ,@(cdr clause))
+			 `(if (or ,@(map (lambda (x) `(eqv? ,tmp ,x)) 
+					 (car clause) ) )
+			      (begin ,@(cdr clause)) 
+			      ,(expand rclauses) ) ) ) ) ) ) ) ) ) ) )
 
-     (define (err x) 
-       (##sys#error "syntax error in `cond-expand' form" x (cons 'cond-expand clauses)) )
-
-     (define (test fx)
-       (cond ((symbol? fx) (##sys#feature? fx))
-	     ((not (pair? fx)) (err fx))
-	     (else
-	      (let ((rest (##sys#slot fx 1)))
-		(case (##sys#slot fx 0)
-		  ((and)
-		   (or (eq? rest '())
-		       (if (pair? rest)
-			   (and (test (##sys#slot rest 0))
-				(test `(and ,@(##sys#slot rest 1))) )
-			   (err fx) ) ) )
-		  ((or) 
-		   (and (not (eq? rest '()))
-			(if (pair? rest)
-			    (or (test (##sys#slot rest 0))
-				(test `(or ,@(##sys#slot rest 1))) )
-			    (err fx) ) ) )
-		  ((not) (not (test (cadr fx))))
-		  (else (err fx)) ) ) ) ) )
-
-     (let expand ((cls clauses))
-       (cond ((eq? cls '())
-	      (##sys#apply
-	       ##sys#error "no matching clause in `cond-expand' form" 
-	       (map (lambda (x) (car x)) clauses) ) )
-	     ((not (pair? cls)) (err cls))
-	     (else
-	      (let ((clause (##sys#slot cls 0))
-		    (rclauses (##sys#slot cls 1)) )
-		(if (not (pair? clause)) 
-		    (err clause)
-		    (let ((id (##sys#slot clause 0)))
-		      (cond ((eq? id 'else)
-			     (let ((rest (##sys#slot clause 1)))
-			       (if (eq? rest '())
-				   '(##core#undefined)
-				   `(begin ,@rest) ) ) )
-			    ((test id) `(begin ,@(##sys#slot clause 1)))
-			    (else (expand rclauses)) ) ) ) ) ) ) ) ) )
-
-(##sys#register-macro-2
+(##sys#register-macro-2			; DEPRECATED
  'switch
  (let ((gensym gensym))
    (lambda (form)
@@ -601,12 +575,17 @@
 ;;; - If REST-ARG has 1 element, return that element.
 ;;; - If REST-ARG has >1 element, error.
 
-(define-macro (:optional rest default-exp)
+(define-macro (optional rest default-exp)
   (let ([var (gensym)])
     `(let ((,var ,rest))
-       (cond ((null? ,var) ,default-exp)
-	     ((null? (cdr ,var)) (car ,var))
-	     (else (##core#check (##sys#error (##core#immutable '"too many optional arguments") ,var)))))))
+       (if (null? ,var) 
+	   ,default-exp
+	   (if (##core#check (null? (cdr ,var)))
+	       (car ,var)
+	       (##sys#error (##core#immutable '"too many optional arguments") ,var))))))
+
+(define-macro (:optional . args)	; DEPRECATED to avoid conflicts with keyword-style prefix
+  `(optional ,@args) )
 
 
 ;;; (LET-OPTIONALS* args ((var1 default1) ... [rest]) body1 ...)
@@ -751,7 +730,6 @@
   (let ([vars (cdr conser)]
 	[slotnames (map car slots)] )
     `(begin
-       (define ,t ',t)
        (define ,conser
 	 (##sys#make-structure 
 	  ',t 
@@ -768,11 +746,11 @@
 		      (setters (memq #:record-setters ##sys#features))
 		      (setr? (pair? (cddr slot))) 
 		      (getr `(lambda (x)
-			       (##sys#check-structure x ',t)
+			       (##core#check (##sys#check-structure x ',t))
 			       (##sys#block-ref x ,i) ) ) )
 		 `(,@(if setr?
 			 `((define (,(caddr slot) x y)
-			     (##sys#check-structure x ',t)
+			     (##core#check (##sys#check-structure x ',t))
 			     (##sys#block-set! x ,i y)) )
 			 '() )
 		   (define ,(cadr slot) 
@@ -874,7 +852,7 @@
 	  (else (syntax-error 'define-extension "invalid clause syntax" cs)) ) ) )
 
 
-;;;; SRFI-31
+;;; SRFI-31
 
 (define-macro (rec head . args)
   (if (pair? head)
@@ -896,7 +874,7 @@
 	'(begin) ) ) )
 
 
-;;;; Register features provided by this file
+;;; Register features provided by this file
 
 (eval-when (compile load eval)
   (register-feature! 'srfi-8 'srfi-16 'srfi-26 'srfi-31 'srfi-15 'srfi-11) )
