@@ -1,6 +1,7 @@
 /* chicken.h - General headerfile for compiler generated executables
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,25 +23,20 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 */
 
-
 /* Configuration: */
 
 #ifndef ___CHICKEN
 #define ___CHICKEN
 
-#if !defined(CMAKE_BUILD) && ((defined(HAVE_CONFIG_H) || defined(HAVE_CHICKEN_CONFIG_H)) && !defined(_MSC_VER))
-# include <chicken-config.h>
+#define C_MAJOR_VERSION       3
+
+/*
+ * N.B. This file MUST not rely upon "chicken-config.h"
+ */
+#if defined(HAVE_CONFIG_H) || defined(HAVE_CHICKEN_CONFIG_H)
+# include "chicken-config.h"
 #endif
 
 #if !defined(__GNUC__) && !defined(__WATCOMC__)
@@ -48,7 +44,7 @@
 #  include <alloca.h>
 # else
 #  ifdef _AIX
-# pragma alloca
+#   pragma alloca
 #  else
 #   ifndef alloca /* predefined by HP cc +Olibcalls */
 char *alloca ();
@@ -76,7 +72,7 @@ char *alloca ();
 #define C_fctexport
 #define C_externimport             C_extern
 #define C_externexport             C_extern
-#if !(defined(C_NO_PIC_NO_DLL) && !defined(PIC))
+#if defined(PIC)
 # if defined(__CYGWIN__) || defined(__MINGW32__)
 #  ifndef C_BUILDING_LIBCHICKEN
 #   undef  C_varextern
@@ -87,16 +83,17 @@ char *alloca ();
 #  define C_fctimport              __declspec(dllexport)
 #  undef  C_externimport
 #  undef  C_externexport
-#  define C_externimport           C_extern __declspec(dllimport)
 #  define C_externexport           C_extern __declspec(dllexport)
 #  undef  C_varextern
 #  undef  C_fctexport
 #  ifdef C_BUILDING_LIBCHICKEN
 #   define C_varextern             C_extern __declspec(dllexport)
 #   define C_fctexport             __declspec(dllexport)
+#   define C_externimport          C_extern __declspec(dllexport)
 #  else
 #   define C_varextern             C_extern __declspec(dllimport)
 #   define C_fctexport             __declspec(dllimport)
+#   define C_externimport          C_extern __declspec(dllimport)
 #  endif
 # elif defined(__WATCOMC__)
 #  undef  C_fctimport
@@ -151,7 +148,6 @@ char *alloca ();
 #endif
 
 #define C_TIMER_INTERRUPTS
-#define C_128_PARAMETERS
 
 #ifdef C_DEFAULT_TARGET_STACK_SIZE
 # define C_resize_stack(n)           C_do_resize_stack(C_DEFAULT_TARGET_STACK_SIZE)
@@ -159,21 +155,24 @@ char *alloca ();
 # define C_resize_stack(n)           C_do_resize_stack(n)
 #endif
 
-#if defined (__alpha__) || defined (__sparc_v9__) || defined (__sparcv9) || defined(__ia64__) || defined(__x86_64__) || defined(__LP64__)
-# define C_SIXTY_FOUR
-#endif
-
-#if defined(__mips64) && (!defined(__GNUC__) || _MIPS_SZPTR == 64)
-# define C_SIXTY_FOUR
+#ifndef C_SIXTY_FOUR
+# if defined (__alpha__) || defined (__sparc_v9__) || defined (__sparcv9) || defined(__ia64__) || defined(__x86_64__) || defined(__LP64__) || defined(__powerpc64__)
+#   define C_SIXTY_FOUR
+# elif defined(__mips64) && (!defined(__GNUC__) || _MIPS_SZPTR == 64)
+#   define C_SIXTY_FOUR
+# endif
 #endif
 
 #if defined(__APPLE__) && defined(__MACH__)
 # define C_MACOSX
-/*
- * Darwin provides gcvt/ecvt/fcvt for compatibility with legacy code.
- * (They don't even have a header definition!)
- * Use snprintf instead.
- */
+#endif
+
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__) || defined(__OpenBSD__)
+# define C_XXXBSD
+#endif
+
+#if defined(C_MACOSX) || defined(__linux__) || defined(C_XXXBSD)
+# define C_GNU_ENV
 #endif
 
 #if defined(_MSC_VER) || defined(__MWERKS__) || defined(__DJGPP__) || defined(__MINGW32__) || defined(__WATCOMC__)
@@ -189,26 +188,51 @@ char *alloca ();
 #include <limits.h>
 #include <time.h>
 
-#ifdef C_SIXTY_FOUR
-# ifdef HAVE_STDINT_H
-#  include <stdint.h>
-# else
-#  include <sys/types.h>
-# endif
-#endif
-
-#ifndef C_NONUNIX
+#if !defined(C_NONUNIX) || defined(__MINGW32__) || defined(__WATCOMC__)
 # include <unistd.h>
+# include <inttypes.h>
+# include <sys/types.h>
 #endif
 
-#if defined(__MWERKS__) && !defined(__INTEL__)
-/* This is a rather crude way of assuming this is MacOS 9 or lower! */
-# define C_MACOS
-# include <alloca.h>
-int strncasecmp(const char *one, const char *two, size_t n);
+#if defined(__MINGW32__)
+# include <sys/param.h>
+#elif defined(__CYGWIN__)
+# include <endian.h>
+#elif defined(__linux__)
+# include <endian.h>
+#elif defined(C_MACOSX) || defined(C_XXXBSD)
+# include <machine/endian.h>
+#elif defined(__hpux__)
+# include <arpa/nameser.h>
+#elif defined(_AIX)
+# include <sys/machine.h>
+#elif defined(__sun__)
+# include <sys/isa_defs.h>
+#elif defined(__svr4__)
+# include <sys/byteorder.h>
 #endif
 
-#ifdef __WATCOMC__
+#if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN
+# define C_BIG_ENDIAN
+#elif defined(BYTE_ORDER) && defined(BIG_ENDIAN) && BYTE_ORDER == BIG_ENDIAN
+# define C_BIG_ENDIAN
+#elif defined(__BIG_ENDIAN__)
+# define C_BIG_ENDIAN
+#elif defined(__sparc__) || defined(__POWERPC__) || defined(__MC68K__) || defined(__mips__)
+# define C_BIG_ENDIAN
+#endif
+
+#if defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN) && __BYTE_ORDER == __LITTLE_ENDIAN
+# define C_LITTLE_ENDIAN
+#elif defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && BYTE_ORDER == LITTLE_ENDIAN
+# define C_LITTLE_ENDIAN
+#elif defined(__LITTLE_ENDIAN__)
+# define C_LITTLE_ENDIAN
+#elif defined (__alpha__) || defined(_M_IX86) || defined(__i386__) || defined(__x86_64__) || defined(__ia64__)
+# define C_LITTLE_ENDIAN
+#endif
+
+#ifdef __MINGW32__
 # include <malloc.h>
 #endif
 
@@ -218,11 +242,19 @@ int strncasecmp(const char *one, const char *two, size_t n);
 # define alloca            _alloca
 # define strncasecmp       strnicmp
 # define isatty            _isatty
+typedef __int8             int8_t;
+typedef unsigned __int8    uint8_t;
+typedef __int16            int16_t;
+typedef unsigned  __int16  uint16_t;
+typedef __int32            int32_t;
+typedef unsigned __int32   uint32_t;
+typedef __int64            int64_t;
+typedef unsigned __int64   uint64_t;
 # pragma warning(disable: 4101)
 #endif
 
-#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__DragonFly__)
-# define C_GNU_ENV
+#ifdef __WATCOMC__
+# include <malloc.h>
 #endif
 
 /* For the easy FFI: */
@@ -235,7 +267,6 @@ int strncasecmp(const char *one, const char *two, size_t n);
 #define ___scheme_pointer   void *
 #define ___byte_vector      unsigned char *
 #define ___symbol           char *
-#define ___callback
 #define ___safe
 #define ___declare(x, y)
 #define ___specialize
@@ -280,8 +311,10 @@ int strncasecmp(const char *one, const char *two, size_t n);
 
 #ifdef C_SIXTY_FOUR
 # define C_MOST_POSITIVE_FIXNUM   0x3fffffffffffffffL
+# define C_WORD_SIZE              64
 #else
 # define C_MOST_POSITIVE_FIXNUM   0x3fffffff
+# define C_WORD_SIZE              32
 #endif
 
 #define C_MOST_NEGATIVE_FIXNUM    (-C_MOST_POSITIVE_FIXNUM - 1)
@@ -326,7 +359,11 @@ int strncasecmp(const char *one, const char *two, size_t n);
 # define C_STRING_TYPE            (0x02000000 | C_BYTEBLOCK_BIT)
 # define C_PAIR_TYPE              (0x03000000)
 # define C_CLOSURE_TYPE           (0x04000000 | C_SPECIALBLOCK_BIT)
-# define C_FLONUM_TYPE            (0x05000000 | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
+# ifdef C_DOUBLE_IS_32_BITS
+#  define C_FLONUM_TYPE            (0x05000000 | C_BYTEBLOCK_BIT)
+# else
+#  define C_FLONUM_TYPE            (0x05000000 | C_BYTEBLOCK_BIT | C_8ALIGN_BIT)
+# endif
 # define C_UNUSED_TYPE            (0x06000000)
 # define C_PORT_TYPE              (0x07000000 | C_SPECIALBLOCK_BIT)
 # define C_STRUCTURE_TYPE         (0x08000000)
@@ -355,13 +392,13 @@ int strncasecmp(const char *one, const char *two, size_t n);
 #define C_SIZEOF_LIST(n)          ((n) * 3 + 1)
 #define C_SIZEOF_PAIR             3
 #define C_SIZEOF_STRING(n)        (C_bytestowords(n) + 2)
-#ifdef C_EXTRA_SYMBOL_SLOT
-# define C_SIZEOF_SYMBOL          4
-#else
-# define C_SIZEOF_SYMBOL          3
-#endif
+#define C_SIZEOF_SYMBOL          4
 #define C_SIZEOF_INTERNED_SYMBOL(n) (C_SIZEOF_SYMBOL + C_SIZEOF_BUCKET + C_SIZEOF_STRING(n))
-#define C_SIZEOF_FLONUM           4
+#ifdef C_DOUBLE_IS_32_BITS
+# define C_SIZEOF_FLONUM           2
+#else
+# define C_SIZEOF_FLONUM           4
+#endif
 #define C_SIZEOF_POINTER          2
 #define C_SIZEOF_TAGGED_POINTER   3
 #define C_SIZEOF_SWIG_POINTER     3
@@ -376,12 +413,7 @@ int strncasecmp(const char *one, const char *two, size_t n);
 #define C_TAGGED_POINTER_TAG      (C_TAGGED_POINTER_TYPE | (C_SIZEOF_TAGGED_POINTER - 1))
 #define C_SWIG_POINTER_TAG        (C_SWIG_POINTER_TYPE | (C_wordstobytes(C_SIZEOF_SWIG_POINTER - 1)))
 #define C_SYMBOL_TAG              (C_SYMBOL_TYPE | (C_SIZEOF_SYMBOL - 1))
-
-#ifdef C_SIXTY_FOUR
-# define C_FLONUM_TAG             (C_FLONUM_TYPE | C_wordstobytes(1))
-#else
-# define C_FLONUM_TAG             (C_FLONUM_TYPE | C_wordstobytes(2))
-#endif
+#define C_FLONUM_TAG             (C_FLONUM_TYPE | sizeof(double))
 
 #ifdef C_SIXTY_FOUR
 # define C_word                   long
@@ -442,6 +474,7 @@ int strncasecmp(const char *one, const char *two, size_t n);
 #define C_RUNTIME_UNSAFE_DLOAD_SAFE_ERROR             33
 #define C_RUNTIME_SAFE_DLOAD_UNSAFE_ERROR             34
 #define C_BAD_ARGUMENT_TYPE_NO_FLONUM_ERROR           35
+#define C_BAD_ARGUMENT_TYPE_NO_CLOSURE_ERROR          36
 
 
 #define CHICKEN_gc_root_ref(root)      (((C_GC_ROOT *)(root))->value)
@@ -460,6 +493,7 @@ int strncasecmp(const char *one, const char *two, size_t n);
 #  define C_cblockend             })
 #  define C_noret                 __attribute__ ((noreturn))
 #  define C_noret_decl(name)
+#  define C_aligned               __attribute__ ((aligned))
 # endif
 # ifdef __i386__
 #  define C_regparm               __attribute__ ((regparm(3)))
@@ -489,6 +523,10 @@ int strncasecmp(const char *one, const char *two, size_t n);
 # define C_ccall
 #endif
 
+#ifndef C_aligned
+# define C_aligned
+#endif
+
 #define C_c_regparm
 
 /* Types: */
@@ -577,8 +615,6 @@ DECL_C_PROC_p3 (48,49,50,51,52,53,54,55,  0,0,1,1,0)
 DECL_C_PROC_p3 (56,57,58,59,60,61,62,63,  0,0,1,1,1)
 DECL_C_PROC_p1 (64,65,  0,1,0,0,0,0,0)
 DECL_C_PROC_p0 (66,  0,1,0,0,0,0,1,0)
-
-#ifdef C_128_PARAMETERS
 DECL_C_PROC_p0 (67,  0,1,0,0,0,0,1,1)
 DECL_C_PROC_p2 (68,69,70,71,  0,1,0,0,0,1)
 DECL_C_PROC_p3 (72,73,74,75,76,77,78,79,  0,1,0,0,1)
@@ -589,14 +625,17 @@ DECL_C_PROC_p3 (104,105,106,107,108,109,110,111,  0,1,1,0,1)
 DECL_C_PROC_p3 (112,113,114,115,116,117,118,119,  0,1,1,1,0)
 DECL_C_PROC_p3 (120,121,122,123,124,125,126,127,  0,1,1,1,1)
 DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
-#endif
 
 
 /* Macros: */
 
+#define C_align4(n)                (((n) + 3) & ~3)
+#define C_align8(n)                (((n) + 7) & ~7)
+#define C_align16(n)               (((n) + 15) & ~15)
+
 /* This is word-size dependent: */
 #ifdef C_SIXTY_FOUR
-# define C_align(n)                (((n) + 7) & ~7)
+# define C_align(n)                C_align8(n)
 # define C_wordstobytes(n)         ((n) << 3)
 # define C_bytestowords(n)         (((n) + 7) >> 3)
 # define C_wordsperdouble(n)       (n)
@@ -604,7 +643,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 # define C_WORD_MAX                LONG_MAX
 # define C_UWORD_MAX               ULONG_MAX
 #else
-# define C_align(n)                (((n) + 3) & ~3)
+# define C_align(n)                C_align4(n)
 # define C_wordstobytes(n)         ((n) << 2)
 # define C_bytestowords(n)         (((n) + 3) >> 2)
 # define C_wordsperdouble(n)       ((n) << 1)
@@ -688,6 +727,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 
 #define C_return(x)                return(x)
 
+#define C_memcpy_slots(t, f, n)    C_memcpy((t), (f), (n) * sizeof(C_word))
 #define C_block_header(x)          (((C_SCHEME_BLOCK *)(x))->header)
 #define C_header_bits(x)           (C_block_header(x) & C_HEADER_BITS_MASK)
 #define C_header_size(x)           (C_block_header(x) & C_HEADER_SIZE_MASK)
@@ -720,7 +760,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_immediatep(x)            ((x) & C_IMMEDIATE_MARK_BITS)
 #define C_mk_bool(x)               ((x) ? C_SCHEME_TRUE : C_SCHEME_FALSE)
 #define C_mk_nbool(x)              ((x) ? C_SCHEME_FALSE : C_SCHEME_TRUE)
-#define C_port_file(p)             ((C_FILEPTR)C_u_i_car(p))
+#define C_port_file(p)             ((C_FILEPTR)C_block_item(p, 0))
 #define C_data_pointer(x)          ((void *)((C_SCHEME_BLOCK *)(x))->data)
 #define C_invert_flag(f)           (!(f))
 #define C_fitsinfixnump(n)         (((n) & C_INT_SIGN_BIT) == (((n) & C_INT_TOP_BIT) << 1))
@@ -766,6 +806,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_zero_length_p(x)        C_mk_bool(C_header_size(x) == 0)
 #define C_boundp(x)               C_mk_bool(((C_SCHEME_BLOCK *)(x))->data[ 0 ] != C_SCHEME_UNBOUND)
 #define C_blockp(x)               C_mk_bool(!C_immediatep(x))
+#define C_forwardedp(x)           C_mk_bool((C_block_header(x) & C_GC_FORWARDING_BIT) != 0)
 #define C_immp(x)                 C_mk_bool(C_immediatep(x))
 #define C_flonump(x)              C_mk_bool(C_block_header(x) == C_FLONUM_TAG)
 #define C_stringp(x)              C_mk_bool(C_header_bits(x) == C_STRING_TYPE)
@@ -790,6 +831,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_anypointerp(x)          C_mk_bool(C_block_header(x) == C_POINTER_TAG || C_block_header(x) == C_TAGGED_POINTER_TAG || C_block_header(x) == C_SWIG_POINTER_TAG)
 #define C_specialp(x)             C_mk_bool(C_header_bits(x) & C_SPECIALBLOCK_BIT)
 #define C_byteblockp(x)           C_mk_bool(C_header_bits(x) & C_BYTEBLOCK_BIT)
+#define C_anyp(x)                 C_SCHEME_TRUE
 #define C_eqp(x, y)               C_mk_bool((x) == (y))
 #define C_vemptyp(x)              C_mk_bool(C_header_size(x) == 0)
 #define C_notvemptyp(x)           C_mk_bool(C_header_size(x) > 0)
@@ -853,9 +895,9 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
                                                                 (C_char *)C_data_pointer(s2) + C_unfix(start2), \
                                                                 C_unfix(len) ) == 0)
 #define C_subvector_copy(v1, v2, start1, end1, start2) \
-                                        (C_memcpy((C_char *)C_data_pointer(v2) + C_unfix(start2), \
+                                        (C_memcpy_slots((C_char *)C_data_pointer(v2) + C_unfix(start2), \
                                                   (C_char *)C_data_pointer(v1) + C_unfix(start1), \
-                                                  sizeof(C_word) * (C_unfix(end1) - C_unfix(start1)) ), C_SCHEME_UNDEFINED)
+						  C_unfix(end1) - C_unfix(start1) ), C_SCHEME_UNDEFINED)
 #define C_words(n)                      C_fix(C_bytestowords(C_unfix(n)))
 #define C_bytes(n)                      C_fix(C_wordstobytes(C_unfix(n)))
 #define C_random_fixnum(n)              C_fix(rand() % C_unfix(n))
@@ -902,6 +944,9 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 # define C_tty_portp(p)                 C_mk_bool(isatty(fileno(C_port_file(p))))
 #endif
 
+#define C_emit_eval_trace_info(x, y, z) C_emit_trace_info2("<eval>", x, y, z)
+#define C_emit_syntax_trace_info(x, y, z) C_emit_trace_info2("<syntax>", x, y, z)
+
 /* These expect C_VECTOR_TYPE to be 0: */
 #define C_vector_to_structure(v)        (((C_SCHEME_BLOCK *)(v))->header |= C_STRUCTURE_TYPE, C_SCHEME_UNDEFINED)
 #define C_vector_to_closure(v)          (((C_SCHEME_BLOCK *)(v))->header |= C_CLOSURE_TYPE, C_SCHEME_UNDEFINED)
@@ -929,6 +974,7 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_a_i_data_mpointer(ptr, n, x)  C_mpointer(ptr, C_data_pointer(x))
 #define C_a_int_to_num(ptr, n, i)       C_int_to_num(ptr, i)
 #define C_a_unsigned_int_to_num(ptr, n, i)  C_unsigned_int_to_num(ptr, i)
+#define C_a_double_to_num(ptr, n)       C_double_to_number(C_flonum(ptr, n))
 #define C_a_i_vector                    C_vector
 #define C_list                          C_a_i_list
 #define C_i_setslot(x, i, y)            (C_mutate(&C_block_item(x, C_unfix(i)), y), C_SCHEME_UNDEFINED)
@@ -997,7 +1043,9 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 # define C_i_not_pair_p                 C_i_not_pair_p_2
 #endif
 
+#define C_i_check_closure(x)            C_i_check_closure_2(x, C_SCHEME_FALSE)
 #define C_i_check_exact(x)              C_i_check_exact_2(x, C_SCHEME_FALSE)
+#define C_i_check_inexact(x)            C_i_check_inexact_2(x, C_SCHEME_FALSE)
 #define C_i_check_number(x)             C_i_check_number_2(x, C_SCHEME_FALSE)
 #define C_i_check_string(x)             C_i_check_string_2(x, C_SCHEME_FALSE)
 #define C_i_check_bytevector(x)         C_i_check_bytevector_2(x, C_SCHEME_FALSE)
@@ -1017,14 +1065,37 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 #define C_u_i_s8vector_ref(x, i)        C_fix(((char *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
 #define C_u_i_u16vector_ref(x, i)       C_fix(((unsigned short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
 #define C_u_i_s16vector_ref(x, i)       C_fix(((short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
+#define C_u_i_u32vector_ref(x, i)       C_fix(((C_u32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
+#define C_u_i_s32vector_ref(x, i)       C_fix(((C_u32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
+#define C_a_i_u32vector_ref(ptr, c, x, i)  C_unsigned_int_to_num(ptr, ((C_u32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
+#define C_a_i_s32vector_ref(ptr, c, x, i)  C_int_to_num(ptr, ((C_s32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ])
 
 #define C_u_i_u8vector_set(x, i, v)     ((((unsigned char *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
 #define C_u_i_s8vector_set(x, i, v)     ((((char *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
-#define C_u_i_u16vector_set(x, i, v)     ((((unsigned short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
-#define C_u_i_s16vector_set(x, i, v)     ((((short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
-#define C_u_i_u32vector_set(x, i, v)     ((((C_u32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_num_to_unsigned_int(v)), C_SCHEME_UNDEFINED)
-#define C_u_i_s32vector_set(x, i, v)     ((((C_s32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_num_to_int(v)), C_SCHEME_UNDEFINED)
+#define C_u_i_u16vector_set(x, i, v)    ((((unsigned short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
+#define C_u_i_s16vector_set(x, i, v)    ((((short *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_unfix(v)), C_SCHEME_UNDEFINED)
+#define C_u_i_u32vector_set(x, i, v)    ((((C_u32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_num_to_unsigned_int(v)), C_SCHEME_UNDEFINED)
+#define C_u_i_s32vector_set(x, i, v)    ((((C_s32 *)C_data_pointer(C_block_item((x), 1)))[ C_unfix(i) ] = C_num_to_int(v)), C_SCHEME_UNDEFINED)
 
+#define C_u_i_bit_setp(x, i)            C_mk_bool((C_unfix(x) & (1 << C_unfix(i))) != 0)
+
+#ifdef C_BIG_ENDIAN
+# ifdef C_SIXTY_FOUR
+#  define C_lihdr(x, y, z)              ((C_LAMBDA_INFO_TYPE >> 56) & 0xff), \
+                                        0, 0, 0, 0, (x), (y), (z)
+# else
+#  define C_lihdr(x, y, z)              ((C_LAMBDA_INFO_TYPE >> 24) & 0xff), \
+                                        (x), (y), (z)
+# endif
+#else
+# ifdef C_SIXTY_FOUR
+#  define C_lihdr(x, y, z)              (z), (y), (x), 0, 0, 0, 0, \
+                                        ((C_LAMBDA_INFO_TYPE >> 56) & 0xff)
+# else
+#  define C_lihdr(x, y, z)              (z), (y), (x), \
+                                        ((C_LAMBDA_INFO_TYPE >> 24) & 0xff)
+# endif
+#endif
 
 #define C_end_of_main
 
@@ -1032,8 +1103,9 @@ DECL_C_PROC_p0 (128,  1,0,0,0,0,0,0,0)
 # ifndef C_WINDOWS_GUI
 #  define C_main_entry_point            int main(int argc, char *argv[]) { return CHICKEN_main(argc, argv, (void*)C_toplevel); } C_end_of_main
 # else
-#  define C_main_entry_point            int WINAPI WinMain(HINSTANCE me, HINSTANCE you, LPSTR cmdline, int show) \
-                                          { return CHICKEN_main(0, NULL, C_toplevel); } C_end_of_main
+#  define C_main_entry_point            \
+  int WINAPI WinMain(HINSTANCE me, HINSTANCE you, LPSTR cmdline, int show) \
+  { return CHICKEN_main(0, NULL, (void *)C_toplevel); } C_end_of_main
 # endif
 #else
 # define C_main_entry_point
@@ -1059,7 +1131,9 @@ C_varextern C_TLS void *C_restart_address;
 C_varextern C_TLS int C_entry_point_status;
 
 C_varextern C_TLS void (C_fcall *C_restart_trampoline)(void *proc) C_regparm C_noret;
-C_varextern C_TLS void (*C_post_gc_hook)(int mode);
+C_varextern C_TLS void (*C_pre_gc_hook)(int mode);
+C_varextern C_TLS void (*C_post_gc_hook)(int mode, long ms);
+C_varextern C_TLS void (*C_panic_hook)(C_char *msg);
 
 C_varextern C_TLS int
   C_abort_on_thread_exceptions,
@@ -1094,13 +1168,16 @@ C_fctexport void *CHICKEN_new_gc_root();
 C_fctexport void CHICKEN_delete_gc_root(void *root);
 C_fctexport void *CHICKEN_global_lookup(char *name);
 C_fctexport int CHICKEN_is_running();
+C_fctexport void CHICKEN_interrupt();
 
 C_fctexport void C_check_nursery_minimum(C_word size);
 C_fctexport int C_fcall C_save_callback_continuation(C_word **ptr, C_word k);
 C_fctexport C_word C_fcall C_restore_callback_continuation(void);
+C_fctexport C_word C_fcall C_restore_callback_continuation2(int level);
 C_fctexport C_word C_fcall C_callback(C_word closure, int argc);
 C_fctexport C_word C_fcall C_callback_wrapper(void *proc, int argc);
-C_fctexport void C_fcall C_callback_adjust_stack_limits(C_word *base);
+C_fctexport void C_fcall C_callback_adjust_stack_limits(C_word *base); /* DEPRECATED */
+C_fctexport void C_fcall C_callback_adjust_stack(C_word *base, int size);
 C_fctexport void CHICKEN_parse_command_line(int argc, char *argv[], C_word *heap, C_word *stack, C_word *symbols);
 C_fctexport void C_fcall C_toplevel_entry(C_char *name) C_regparm;
 C_fctexport C_word C_fcall C_enable_interrupts(void) C_regparm;
@@ -1191,6 +1268,7 @@ C_fctexport int C_fcall C_in_heapp(C_word x) C_regparm;
 C_fctexport int C_fcall C_in_fromspacep(C_word x) C_regparm;
 C_fctexport void C_fcall C_trace(C_char *name) C_regparm;
 C_fctexport C_word C_fcall C_emit_trace_info(C_word x, C_word y, C_word t) C_regparm;
+C_fctexport C_word C_fcall C_emit_trace_info2(char *raw, C_word x, C_word y, C_word t) C_regparm;
 C_fctexport C_word C_fcall C_hash_string(C_word str) C_regparm;
 C_fctexport C_word C_fcall C_hash_string_ci(C_word str) C_regparm;
 C_fctexport C_word C_halt(C_word msg);
@@ -1200,6 +1278,8 @@ C_fctexport C_word C_fcall C_set_gc_report(C_word flag) C_regparm;
 C_fctexport C_word C_fcall C_start_timer(void) C_regparm;
 C_fctexport C_word C_exit_runtime(C_word code);
 C_fctexport C_word C_fcall C_display_flonum(C_word port, C_word n) C_regparm;
+C_fctexport C_word C_fcall C_set_print_precision(C_word n) C_regparm;
+C_fctexport C_word C_fcall C_get_print_precision(void) C_regparm;
 C_fctexport C_word C_fcall C_read_char(C_word port) C_regparm;
 C_fctexport C_word C_fcall C_peek_char(C_word port) C_regparm;
 C_fctexport C_word C_fcall C_execute_shell_command(C_word string) C_regparm;
@@ -1278,6 +1358,7 @@ C_fctexport void C_ccall C_peek_unsigned_integer(C_word c, C_word closure, C_wor
 C_fctexport void C_ccall C_decode_seconds(C_word c, C_word closure, C_word k, C_word secs, C_word mode) C_noret;
 C_fctexport void C_ccall C_software_type(C_word c, C_word closure, C_word k) C_noret;
 C_fctexport void C_ccall C_machine_type(C_word c, C_word closure, C_word k) C_noret;
+C_fctexport void C_ccall C_machine_byte_order(C_word c, C_word closure, C_word k) C_noret;
 C_fctexport void C_ccall C_software_version(C_word c, C_word closure, C_word k) C_noret;
 C_fctexport void C_ccall C_build_platform(C_word c, C_word closure, C_word k) C_noret;
 C_fctexport void C_ccall C_c_runtime(C_word c, C_word closure, C_word k) C_noret;
@@ -1304,13 +1385,14 @@ C_fctexport C_word C_fcall C_a_i_bytevector(C_word **a, int c, C_word x) C_regpa
 C_fctexport C_word C_fcall C_i_eqvp(C_word x, C_word y) C_regparm;
 C_fctexport C_word C_fcall C_i_symbolp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_pairp(C_word x) C_regparm;
-C_fctexport C_word C_fcall C_i_atomp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_vectorp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_closurep(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_portp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_stringp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_numberp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_integerp(C_word x) C_regparm;
+C_fctexport C_word C_fcall C_i_flonump(C_word x) C_regparm;
+C_fctexport C_word C_fcall C_i_finitep(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_locativep(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_fixnum_min(C_word x, C_word y) C_regparm;
 C_fctexport C_word C_fcall C_i_fixnum_max(C_word x, C_word y) C_regparm;
@@ -1363,7 +1445,9 @@ C_fctexport C_word C_fcall C_i_memv(C_word x, C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_member(C_word x, C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_length(C_word lst) C_regparm;
 C_fctexport C_word C_fcall C_i_inexact_to_exact(C_word n) C_regparm;
+C_fctexport C_word C_fcall C_i_check_closure_2(C_word x, C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_i_check_exact_2(C_word x, C_word loc) C_regparm;
+C_fctexport C_word C_fcall C_i_check_inexact_2(C_word x, C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_i_check_number_2(C_word x, C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_i_check_string_2(C_word x, C_word loc) C_regparm;
 C_fctexport C_word C_fcall C_i_check_bytevector_2(C_word x, C_word loc) C_regparm;
@@ -1399,6 +1483,7 @@ C_fctexport C_word C_fcall C_a_i_flonum_negate(C_word **a, int c, C_word n1) C_r
 C_fctexport C_word C_fcall C_a_i_bitwise_and(C_word **a, int c, C_word n1, C_word n2) C_regparm;
 C_fctexport C_word C_fcall C_a_i_bitwise_ior(C_word **a, int c, C_word n1, C_word n2) C_regparm;
 C_fctexport C_word C_fcall C_a_i_bitwise_not(C_word **a, int c, C_word n1) C_regparm;
+C_fctexport C_word C_fcall C_i_bit_setp(C_word n, C_word i) C_regparm;
 C_fctexport C_word C_fcall C_a_i_bitwise_xor(C_word **a, int c, C_word n1, C_word n2) C_regparm;
 C_fctexport C_word C_fcall C_a_i_arithmetic_shift(C_word **a, int c, C_word n1, C_word n2) C_regparm;
 C_fctexport C_word C_fcall C_a_i_exp(C_word **a, int c, C_word n) C_regparm;
@@ -1411,6 +1496,11 @@ C_fctexport C_word C_fcall C_a_i_acos(C_word **a, int c, C_word n) C_regparm;
 C_fctexport C_word C_fcall C_a_i_atan(C_word **a, int c, C_word n) C_regparm;
 C_fctexport C_word C_fcall C_a_i_atan2(C_word **a, int c, C_word n1, C_word n2) C_regparm;
 C_fctexport C_word C_fcall C_a_i_sqrt(C_word **a, int c, C_word n) C_regparm;
+C_fctexport C_word C_fcall C_i_o_fixnum_plus(C_word x, C_word y) C_regparm;
+C_fctexport C_word C_fcall C_i_o_fixnum_difference(C_word x, C_word y) C_regparm;
+C_fctexport C_word C_fcall C_i_o_fixnum_and(C_word x, C_word y) C_regparm;
+C_fctexport C_word C_fcall C_i_o_fixnum_ior(C_word x, C_word y) C_regparm;
+C_fctexport C_word C_fcall C_i_o_fixnum_xor(C_word x, C_word y) C_regparm;
 
 C_fctexport C_word C_fcall C_i_foreign_char_argumentp(C_word x) C_regparm;
 C_fctexport C_word C_fcall C_i_foreign_fixnum_argumentp(C_word x) C_regparm;
@@ -1427,6 +1517,7 @@ C_fctexport C_word C_fcall C_i_foreign_unsigned_integer_argumentp(C_word x) C_re
 
 C_fctexport C_char *C_lookup_procedure_id(void *ptr);
 C_fctexport void *C_lookup_procedure_ptr(C_char *id);
+C_fctexport C_word C_dunload(C_word name);
 
 #ifdef C_SIXTY_FOUR
 C_fctexport void C_ccall C_peek_signed_integer_32(C_word c, C_word closure, C_word k, C_word v, C_word index) C_noret;
@@ -1436,6 +1527,8 @@ C_fctexport void C_ccall C_peek_unsigned_integer_32(C_word c, C_word closure, C_
 # define C_peek_unsigned_integer_32  C_peek_unsigned_integer
 #endif
 
+C_fctexport C_word C_fcall C_decode_literal(C_word **ptr, C_char *str) C_regparm;
+
 /* defined in eval.scm: */
 C_fctexport  void  CHICKEN_get_error_message(char *buf,int bufsize);
 C_fctexport  int  CHICKEN_load(char * filename);
