@@ -1,11 +1,13 @@
 ;;;; compiler.scm - The CHICKEN Scheme compiler
 ;
 ;
+;
 ; "This is insane. What we clearly want to do is not exactly clear, and is rooted in NCOMPLR."
 ;
 ;
 ;-----------------------------------------------------------------------------------------------------------
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,15 +30,6 @@
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
 ;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
-;
 ;
 ; Supported syntax:
 ;
@@ -63,20 +56,18 @@
 ; (disable-warning <class> ...)
 ; (always-bound {<name>})
 ; (foreign-declare {<string>})
-; (foreign-parse {<string>})
 ; (block)
 ; (separate)
 ; (run-time-macros)
 ; (export {<name>})
-; (compress-literals [<threshold>])
 ; (safe-globals)
-; (namespace <name> {<symbol})
 ; (custom-declare (<tag> <name> <filename> <arg> ...) <string> ...)
 ; (data <tag1> <exp1> ...)
 ; (post-process <string> ...)
 ; (emit-exports <string>)
 ; (keep-shadowed-macros)
 ; (import <symbol-or-string> ...)
+; (unused <symbol> ...)
 ;
 ;   <type> = fixnum | generic
 ;
@@ -206,7 +197,7 @@
 ;   foldable -> <boolean>                    If true: variable names foldable standard-binding
 ;   boxed -> <boolean>                       If true: variable has to be boxed after closure-conversion
 ;   contractable -> <boolean>                If true: variable names contractable procedure
-;   inlineable -> <boolean>                  If true: variable names potentially inlinable procedure
+;   inlinable -> <boolean>                   If true: variable names potentially inlinable procedure
 ;   collapsable -> <boolean>                 If true: variable refers to collapsable constant
 ;   removable -> <boolean>                   If true: variable is not used
 ;   replacable -> <variable>                 Variable can be replaced by another variable
@@ -233,20 +224,13 @@
 
 (declare
  (unit compiler)
- (disable-warning var)
- (foreign-declare #<<EOF
-#ifdef CMAKE_BUILD
-# include "chicken-paths.h"
-# include "stack-size.h"
-#elif defined(C_USE_C_DEFAULTS)
-# include "chicken-defaults.h"
-#else
-# define C_INSTALL_HOME NULL
+ (disable-warning var) )
+
+#>
+#ifndef C_INSTALL_SHARE_HOME
+# define C_INSTALL_SHARE_HOME NULL
 #endif
 
-#ifdef CMAKE_BUILD
-# include "stack-size.h"
-#endif
 #ifndef C_DEFAULT_TARGET_STACK_SIZE
 # define C_DEFAULT_TARGET_STACK_SIZE 0
 #endif
@@ -254,20 +238,19 @@
 #ifndef C_DEFAULT_TARGET_HEAP_SIZE
 # define C_DEFAULT_TARGET_HEAP_SIZE 0
 #endif
-EOF
-) )
+<#
 
 
-#{compiler
+(private compiler
   compiler-arguments process-command-line explicit-use-flag inline-list not-inline-list
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home decompose-lambda-list external-to-pointer defconstant-bindings constant-declarations
-  foreign-type-table-size copy-node! error-is-extended-binding toplevel-scope toplevel-lambda-id
+  copy-node! error-is-extended-binding toplevel-scope toplevel-lambda-id
   unit-name insert-timer-checks used-units external-variables require-imports-flag custom-declare-alist
   profile-info-vector-name finish-foreign-result pending-canonicalizations
-  foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
+  foreign-declarations emit-trace-info block-compilation line-number-database-size
   always-bound-to-procedure block-globals make-block-variable-literal block-variable-literal? block-variable-literal-name
   target-heap-size target-stack-size valid-c-identifier?
   target-initial-heap-size internal-bindings source-filename dump-nodes source-info->string
@@ -275,17 +258,17 @@ EOF
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   broken-constant-nodes inline-substitutions-enabled loop-lambda-names expand-profile-lambda
-  profile-lambda-list profile-lambda-index emit-profile expand-profile-lambda no-c-syntax-checks
-  direct-call-ids foreign-type-table first-analysis callback-names namespace-table disabled-warnings
+  profile-lambda-list profile-lambda-index emit-profile expand-profile-lambda
+  direct-call-ids foreign-type-table first-analysis callback-names disabled-warnings
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database! scan-toplevel-assignments
-  compiler-warning import-table use-import-table
+  compiler-warning import-table use-import-table compiler-macro-table compiler-macros-enabled
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call inline-max-size
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub 
   expand-foreign-lambda* data-declarations emit-control-file-item expand-foreign-primitive
-  process-declaration external-protos-first basic-literal? emit-line-info
+  process-declaration external-protos-first basic-literal?
   transform-direct-lambdas! expand-foreign-callback-lambda* debugging emit-unsafe-marker
-  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -293,23 +276,25 @@ EOF
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database export-list csc-control-file
-  estimate-foreign-result-location-size compressed-literals-initializer parse-easy-ffi
+  estimate-foreign-result-location-size unused-variables
   expand-foreign-callback-lambda default-optimization-passes default-optimization-passes-when-trying-harder
   units-used-by-default words-per-flonum disable-stack-overflow-checking
-  parameter-limit eq-inline-operator optimizable-rest-argument-operators check-c-syntax postponed-initforms
+  parameter-limit eq-inline-operator optimizable-rest-argument-operators postponed-initforms
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   make-random-name final-foreign-type real-name-table real-name set-real-name! safe-globals-flag
-  location-pointer-map literal-compression-threshold compressed-literals compressable-literal
-  lookup-exports-file undefine-shadowed-macros
+  location-pointer-map literal-rewrite-hook
+  lookup-exports-file undefine-shadowed-macros process-lambda-documentation emit-syntax-trace-info
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   process-custom-declaration do-lambda-lifting file-requirements emit-closure-info export-file-name
-  foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result}
+  foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result
+  big-fixnum?)
 
 (eval-when (compile eval)
   (match-error-control #:fail) )
 
+
 (include "tweaks")
-(include "parameters")
+
 
 (define-inline (gensym-f-id) (gensym 'f_))
 
@@ -319,7 +304,7 @@ EOF
   (define default-target-stack-size #f) )
 
 (eval-when (load)
-  (define-foreign-variable installation-home c-string "C_INSTALL_HOME")
+  (define-foreign-variable installation-home c-string "C_INSTALL_SHARE_HOME")
   (define-foreign-variable default-target-heap-size int "C_DEFAULT_TARGET_HEAP_SIZE")
   (define-foreign-variable default-target-stack-size int "C_DEFAULT_TARGET_STACK_SIZE") )
 
@@ -330,6 +315,16 @@ EOF
 (define user-pass-2 (make-parameter #f))
 (define user-post-analysis-pass (make-parameter #f))
 
+(define-constant foreign-type-table-size 301)
+(define-constant analysis-database-size 3001)
+(define-constant default-line-number-database-size 997)
+(define-constant inline-table-size 301)
+(define-constant constant-table-size 301)
+(define-constant file-requirements-size 301)
+(define-constant real-name-table-size 997)
+(define-constant import-table-size 997)
+(define-constant default-inline-max-size 10)
+
 
 ;;; Global variables containing compilation parameters:
 
@@ -357,21 +352,15 @@ EOF
 (define block-globals '())
 (define source-filename #f)
 (define export-list #f)
-(define compressed-literals '())
-(define literal-compression-threshold #f)
-(define compressed-literals-initializer #f)
 (define safe-globals-flag #f)
 (define explicit-use-flag #f)
 (define disable-stack-overflow-checking #f)
-(define namespace-table '())
 (define require-imports-flag #f)
-(define no-c-syntax-checks #f)
 (define emit-unsafe-marker #f)
 (define external-protos-first #f)
 (define do-lambda-lifting #f)
 (define inline-max-size -1)
 (define emit-closure-info #t)
-(define emit-line-info #f)
 (define export-file-name #f)
 (define import-table #f)
 (define use-import-table #f)
@@ -427,6 +416,10 @@ EOF
 (define not-inline-list '())
 (define file-requirements #f)
 (define postponed-initforms '())
+(define unused-variables '())
+(define compiler-macro-table #f)
+(define compiler-macros-enabled #t)
+(define literal-rewrite-hook #f)
 
 
 ;;; Initialize globals:
@@ -443,7 +436,9 @@ EOF
       (set! constant-table (make-vector constant-table-size '())) )
   (set! profile-info-vector-name (make-random-name 'profile-info))
   (set! real-name-table (make-vector real-name-table-size '()))
-  (set! file-requirements (make-hash-table eq?))
+  (if file-requirements
+      (vector-fill! file-requirements '())
+      (set! file-requirements (make-vector file-requirements-size '())) )
   (if import-table
       (vector-fill! import-table '())
       (set! import-table (make-vector import-table-size '())) )
@@ -473,43 +468,57 @@ EOF
       [('quote x) x]
       [x x] ) )
 
+  (define (resolve-atom x ae me dest)
+    (cond [(and constants-used (##sys#hash-table-ref constant-table x)) 
+	   => (lambda (val) (walk (car val) ae me dest)) ]
+	  [(and inline-table-used (##sys#hash-table-ref inline-table x))
+	   => (lambda (val)
+		(walk val ae me dest)) ]
+	  [(assq x foreign-variables)
+	   => (lambda (fv) 
+		(let* ([t (second fv)]
+		       [ft (final-foreign-type t)] 
+		       [body `(##core#inline_ref (,(third fv) ,t))] )
+		  (foreign-type-convert-result
+		   (finish-foreign-result ft body)
+		   t) ) ) ]
+	  [(assq x location-pointer-map)
+	   => (lambda (a)
+		(let* ([t (third a)]
+		       [ft (final-foreign-type t)] 
+		       [body `(##core#inline_loc_ref (,t) ,(second a))] )
+		  (foreign-type-convert-result
+		   (finish-foreign-result ft body)
+		   t) ) ) ]
+	  [else #f] ) )
+
+  (define (walk-literal x ae me dest)
+    (if literal-rewrite-hook
+	(literal-rewrite-hook x (cut walk <> ae me dest))
+	`(quote ,x) ) )
+
   (define (walk x ae me dest)
     (cond ((symbol? x)
-	   (cond [(assq x ae) 
-		  => (lambda (n)
-		       (walk (##sys#macroexpand-1-local (cdr n) me) ae me dest) ) ]
-		 [(and constants-used (##sys#hash-table-ref constant-table x)) 
-		  => (lambda (val) (walk (car val) ae me dest)) ]
-		 [(and inline-table-used (##sys#hash-table-ref inline-table x))
-		  => (lambda (val) (walk val ae me dest)) ]
-		 [(assq x foreign-variables)
-		  => (lambda (fv) 
-		       (let* ([t (second fv)]
-			      [ft (final-foreign-type t)] 
-			      [body `(##core#inline_ref (,(third fv) ,t))] )
-			 (foreign-type-convert-result
-			  (finish-foreign-result ft body)
-			  t) ) ) ]
-		 [(assq x location-pointer-map)
-		  => (lambda (a)
-		       (let* ([t (third a)]
-			      [ft (final-foreign-type t)] 
-			      [body `(##core#inline_loc_ref (,t) ,(second a))] )
-			 (foreign-type-convert-result
-			  (finish-foreign-result ft body)
-			  t) ) ) ]
-		 [else x] ) )
-	  ((and (not-pair? x) (constant? x)) `(quote ,x))
-	  ((not-pair? x) (quit "syntax error - illegal atomic form `~s'" x))
+	   (cond ((assq x ae) => 
+		  (lambda (a)
+		    (let ((alias (cdr a)))
+		      (or (resolve-atom alias ae me dest)
+			  alias) ) ) )
+		 ((resolve-atom x ae me dest))
+		 (else (##sys#alias-global-hook x))) )
+	  ((and (not-pair? x) (constant? x)) 
+	   (walk-literal x ae me dest) )
+	  ((not-pair? x) (syntax-error "illegal atomic form" x))
 	  ((symbol? (car x))
 	   (let* ([head (car x)]
 		  [rest (cdr x)]
 		  [ln (get-line x)]
 		  [name (resolve head ae)] )
+	     (emit-syntax-trace-info x #f)
 	     (unless (proper-list? x)
 	       (if ln
-		   (quit "syntax error in line ~s - malformed expression `~s'" ln x)
-		   (quit "syntax error - malformed expression `~s'" x) ) )
+		   (syntax-error (sprintf "(in line ~s) - malformed expression" ln) x)
+		   (syntax-error "malformed expression" x)))
 	     (set! ##sys#syntax-error-culprit x)
 	     (let* ([x2 (cons name rest)]
 		    [xexpanded (##sys#macroexpand-1-local x2 me)] )
@@ -517,7 +526,8 @@ EOF
 		      (when ln (update-line-number-database! xexpanded ln))
 		      (walk xexpanded ae me dest) ]
 		     [(and inline-table-used (##sys#hash-table-ref inline-table name))
-		      => (lambda (val) (walk (cons val (cdr x)) ae me dest)) ]
+		      => (lambda (val)
+			   (walk (cons val (cdr x)) ae me dest)) ]
 		     [else
 		      (case name
 
@@ -531,18 +541,7 @@ EOF
 
 			((quote)
 			 (##sys#check-syntax 'quote x '(quote _))
-			 (let* ([lit (cadr x)] 
-				[cf (and literal-compression-threshold
-					 (compressable-literal lit literal-compression-threshold) ) ] )
-			   (if cf
-			       (let ([var (gensym 'lf)])
-				 (debugging 'o "compressing literal of size" cf)
-				 (set! compressed-literals
-				   (alist-cons var (write-to-string lit) compressed-literals) )
-				 (set! always-bound (cons var always-bound))
-				 (set! block-globals (cons var block-globals))
-				 var)
-			       x) ) )
+			 (walk-literal (cadr x) ae me dest) )
 
 			((##core#check)
 			 (if unsafe
@@ -559,23 +558,16 @@ EOF
 				    (set! block-globals (cons var block-globals))
 				    var) ] ) ) )
 
-			((##core#undefined ##core#callunit ##core#primitive ##core#inline_ref ##core#inline_loc_ref) x)
+			((##core#undefined ##core#callunit ##core#primitive ##core#inline_ref 
+					   ##core#inline_loc_ref) x)
 
 			((##core#require-for-syntax)
 			 (let ([ids (map eval (cdr x))])
 			   (apply ##sys#require ids)
-			   (hash-table-update! 
+			   (##sys#hash-table-update! 
 			    file-requirements 'syntax-requirements (cut lset-union eq? <> ids)
 			    (lambda () ids) )
-			   (let ([rs (##sys#lookup-runtime-requirements ids)])
-			     (hash-table-update!
-			      file-requirements 'runtime-requirements (cut lset-union eq? <> rs)
-			      (lambda () rs))
-			     (walk
-			      (if (null? rs)
-				  '(##core#undefined)
-				  `(##sys#require ,@(map (lambda (x) `',x) rs)) )
-			      ae me dest) ) ) )
+			   '(##core#undefined) ) )
 
 			((##core#require-extension)
 			 (walk
@@ -614,7 +606,7 @@ EOF
 			   `(let ,(map (lambda (alias b)
 					 (list alias (walk (cadr b) ae me (car b))) )
 				       aliases bindings)
-			      ,(walk (##sys#canonicalize-body (cddr x) (cut assq <> ae2) me)
+			      ,(walk (##sys#canonicalize-body (cddr x) (cut assq <> ae2) me dest)
 				     ae2
 				     me dest) ) ) )
 
@@ -631,21 +623,27 @@ EOF
 			   (decompose-lambda-list
 			    llist
 			    (lambda (vars argc rest)
-			      (let* ([aliases (map gensym vars)]
-				     [ae2 (append (map cons vars aliases) ae)]
-				     [body 
-				      (walk 
-				       (##sys#canonicalize-body obody (cut assq <> ae2) me)
-				       ae2
-				       me #f) ]
-				     [llist2 
+			      (let* ((aliases (map gensym vars))
+				     (ae2 (append (map cons vars aliases) ae))
+				     (body0 (##sys#canonicalize-body obody (cut assq <> ae2) me dest))
+				     (body (walk body0 ae2 me #f))
+				     (llist2 
 				      (build-lambda-list
 				       aliases argc
-				       (and rest (list-ref aliases (posq rest vars))) ) ] )
+				       (and rest (list-ref aliases (posq rest vars))) ) )
+				     (l `(lambda ,llist2 ,body)) )
 				(set-real-names! aliases vars)
-				(cond [(and dest emit-profile (eq? 'lambda name))
-				       (expand-profile-lambda dest llist2 body) ]
-				      [else `(lambda ,llist2 ,body)] ) ) ) ) ) )
+				(cond ((or (not dest) 
+					   (not (eq? dest (resolve dest ae)))) ; global?
+				       l)
+				      ((and emit-profile (eq? 'lambda name))
+				       (expand-profile-lambda dest llist2 body) )
+				      (else
+				       (match body0
+					 (('begin (or (? string? doc) ('quote doc)) _ . more)
+					  (process-lambda-documentation
+					   dest doc l) )
+					 (_ l) ) ) ) ) ) ) ) )
 
 			((##core#named-lambda)
 			 (walk `(lambda ,@(cddr x)) ae me (cadr x)) )
@@ -657,7 +655,7 @@ EOF
 				(ae2 (append (map cons vars aliases) ae))
 				[body 
 				 (walk 
-				  (##sys#canonicalize-body obody (cut assq <> ae2) me)
+				  (##sys#canonicalize-body obody (cut assq <> ae2) me dest)
 				  ae2
 				  me #f) ] )
 			   (set-real-names! aliases vars)
@@ -669,11 +667,12 @@ EOF
 				[var (resolve var0 ae)]
 				[ln (get-line x)]
 				[val (walk (caddr x) ae me var0)] )
-			   (when (and safe-globals-flag (eq? var var0))
-			     (set! always-bound-to-procedure
-			       (lset-adjoin eq? always-bound-to-procedure var))
-			     (set! always-bound (lset-adjoin eq? always-bound var)) )
-			   (when (eq? var var0)
+			   (when (eq? var var0) ; global?
+			     (set! var (##sys#alias-global-hook var))
+			     (when safe-globals-flag
+			       (set! always-bound-to-procedure
+				 (lset-adjoin eq? always-bound-to-procedure var))
+			       (set! always-bound (lset-adjoin eq? always-bound var)) )
 			     (when (macro? var)
 			       (compiler-warning 
 				'var "assigned global variable `~S' is a macro ~A"
@@ -760,7 +759,6 @@ EOF
 				[name (if (pair? (cdddr x))
 					  (cadr (fourth x))
 					  (symbol->string var) ) ] )
-			   (check-c-syntax name 'define-foreign-variable)
 			   (set! foreign-variables
 			     (cons (list var type (if (string? name) name (symbol->string name)))
 				   foreign-variables))
@@ -779,7 +777,9 @@ EOF
 				    (walk
 				     `(begin
 					(##core#set! ,arg ,(first conv))
-					(##core#set! ,ret ,(if (pair? (cdr conv)) (second conv) '##sys#values)) ) 
+					(##core#set! 
+					 ,ret 
+					 ,(if (pair? (cdr conv)) (second conv) '##sys#values)) ) 
 				     ae me dest) ) ]
 				 [else
 				  (##sys#hash-table-set! foreign-type-table name type)
@@ -811,7 +811,11 @@ EOF
 			   `(let (,(let ([size (words (estimate-foreign-result-location-size type))])
 				     ;; Add 2 words: 1 for the header, 1 for double-alignment:
 				     ;; Note: C_a_i_bytevector takes number of words, not bytes
-				     (list store `(##core#inline_allocate ("C_a_i_bytevector" ,(+ 2 size)) ',size)) ) )
+				     (list 
+				      store
+				      `(##core#inline_allocate
+					("C_a_i_bytevector" ,(+ 2 size))
+					',size)) ) )
 			      ,(walk
 				`(begin
 				   ,@(if init
@@ -838,7 +842,9 @@ EOF
 			 (let* ([name (cadr (second x))]
 				[valexp (third x)]
 				[val (handle-exceptions ex
-					 (quit "error in constant evaluation of ~S for named constant ~S" valexp name)
+					 ;; could show line number here
+					 (quit "error in constant evaluation of ~S for named constant ~S" 
+					       valexp name)
 				       (if (collapsable-literal? valexp)
 					   valexp
 					   (eval `(let ,defconstant-bindings ,valexp)) ) ) ] )
@@ -848,11 +854,6 @@ EOF
 				  (##sys#hash-table-set! constant-table name (list val))
 				  '(##core#undefined) ]
 				 [else
-				  (unless (basic-literal? val)
-				    (compiler-warning 
-				     'const
-				     "value for constant binding appears not to be a valid literal: ~s"
-				     val) )
 				  (let ([var (gensym "constant")])
 				    (##sys#hash-table-set! constant-table name (list var))
 				    (set! mutable-constants (alist-cons var val mutable-constants))
@@ -877,11 +878,12 @@ EOF
 			     (let ([name (cadr name)])
 			       (if (valid-c-identifier? name)
 				   (set! callback-names (cons name callback-names))
-				   (quit "name `~S' of external definition is not a valid C identifier" name) ) )
+				   (quit "name `~S' of external definition is not a valid C identifier"
+					 name) ) )
 			     (when (or (not (proper-list? vars)) 
 				       (not (proper-list? atypes))
 				       (not (= (length vars) (length atypes))) )
-			       (##sys#syntax-error-hook 
+			       (syntax-error 
 				"non-matching or invalid argument list to foreign callback-wrapper"
 				vars atypes) )
 			     `(##core#foreign-callback-wrapper
@@ -903,13 +905,25 @@ EOF
 					  ,(foreign-type-convert-argument
 					    `(let ()
 					       ,@(match rtype
-						   ((or '(const nonnull-c-string) 'nonnull-c-string)
+						   ((or '(const nonnull-c-string) 
+							'(const nonnull-unsigned-c-string)
+							'nonnull-unsigned-c-string
+							'nonnull-c-string)
 						    `((##sys#make-c-string (let () ,@(cddr lam)))))
-						   ((or '(const c-string*) 'c-string*)
-						    (##sys#syntax-error-hook
-						     "`c-string*' is not a valid resulte type for callback procedures"
+						   ((or '(const c-string*)
+							'(const unsigned-c-string*)
+							'unsigned-c-string*
+							'c-string*
+							'c-string-list
+							'c-string-list*)
+						    (syntax-error
+						     "not a valid result type for callback procedures"
+						     rtype
 						     name) )
-						   ((or 'c-string '(const c-string))
+						   ((or 'c-string
+							'(const unsigned-c-string)
+							'unsigned-c-string
+							'(const c-string))
 						    `((let ((r (let () ,@(cddr lam))))
 							(and r (##sys#make-c-string r)) ) ) )
 						   (_ (cddr lam)) ) )
@@ -946,18 +960,29 @@ EOF
 					       (walk `(##sys#make-locative ,sym 0 #f 'location) ae me #f) ] )
 					(walk `(##sys#make-locative ,sym 0 #f 'location) ae me #f) ) ) ]
 				 
+				 ((and compiler-macros-enabled
+				       compiler-macro-table
+				       (##sys#hash-table-ref compiler-macro-table name)) =>
+				  (lambda (cm)
+				    (let ((cx (cm x)))
+				      (if (equal? cx x)
+					  (handle-call)
+					  (walk cx ae me dest)))))
+
 				 [else (handle-call)] ) ) ) ) ] ) ) ) )
 
 	  ((not (proper-list? x))
-	   (quit "syntax error - malformed expression `~s'" x) )
+	   (syntax-error "malformed expression" x) )
 
 	  ((constant? (car x))
+	   (emit-syntax-trace-info x #f)
 	   (compiler-warning 'syntax "literal in operator position: ~S" x) 
 	   (mapwalk x ae me) )
 
 	  ((and (pair? (car x)) (eq? 'lambda (caar x)))
 	   (let ([lexp (car x)]
 		 [args (cdr x)] )
+	     (emit-syntax-trace-info x #f)
 	     (##sys#check-syntax 'lambda lexp '(lambda lambda-list . #(_ 1)))
 	     (let ([llist (cadr lexp)])
 	       (if (and (proper-list? llist) (= (length llist) (length args)))
@@ -968,12 +993,15 @@ EOF
 			 (,var ,@(cdr x)) )
 		      ae me dest) ) ) ) ) )
 	  
-	  (else (mapwalk x ae me)) ) )
+	  (else
+	   (emit-syntax-trace-info x #f)
+	   (mapwalk x ae me)) ) )
   
   (define (mapwalk xs ae me)
     (map (lambda (x) (walk x ae me #f)) xs) )
 
   (when (memq 'c debugging-chicken) (newline) (pretty-print exp))
+  (##sys#clear-trace-buffer)
   ;; Process visited definitions and main expression:
   (walk 
    `(begin
@@ -990,11 +1018,11 @@ EOF
   (define (check-decl spec minlen . maxlen)
     (let ([n (length (cdr spec))])
       (if (or (< n minlen) (> n (:optional maxlen 99999)))
-	  (quit "syntax error in declaration: `~s'" spec) ) ) )  
+	  (syntax-error "invalid declaration" spec) ) ) )  
   (call-with-current-continuation
    (lambda (return)
      (unless (pair? spec)
-       (quit "invalid specification syntax: ~S" spec) )
+       (syntax-error "invalid declaration specification" spec) )
      (case (car spec)
        ((uses)
 	(let ([us (cdr spec)])
@@ -1002,14 +1030,14 @@ EOF
 	  (when use-import-table
 	    (for-each lookup-exports-file us) )
 	  (when (pair? us)
-	    (hash-table-update! file-requirements 'uses (cut lset-union eq? us <>) (lambda () us))
+	    (##sys#hash-table-update! file-requirements 'uses (cut lset-union eq? us <>) (lambda () us))
 	    (let ((units (map (lambda (u) (string->c-identifier (stringify u))) us)))
 	      (set! used-units (append used-units units)) ) ) ) )
        ((unit)
 	(check-decl spec 1 1)
 	(let* ([u (cadr spec)]
 	       [un (string->c-identifier (stringify u))] )
-	  (hash-table-set! file-requirements 'unit u)
+	  (##sys#hash-table-set! file-requirements 'unit u)
 	  (when (and unit-name (not (string=? unit-name un)))
 	    (compiler-warning 'usage "unit was already given a name (new name is ignored)") )
 	  (set! unit-name un) ) )
@@ -1058,18 +1086,10 @@ EOF
 	(let ([fds (cdr spec)])
 	  (if (every string? fds)
 	      (set! foreign-declarations (append foreign-declarations fds))
-	      (quit "syntax error in declaration: `~S'" spec) )
-	  (check-c-syntax (apply string-append fds) 'foreign-declare) ) )
-       ((foreign-parse)
-	(let ([fds (cdr spec)])
-	  (if (every string? fds)
-	      (return 
-	       (##sys#compiler-toplevel-macroexpand-hook 
-		`(begin ,@(parse-easy-ffi (string-intersperse fds ""))) ) )
-	      (quit "syntax error in declaration: `~S'" spec) ) ) )
+	      (syntax-error "invalid declaration" spec) ) ) )
        ((custom-declare)
 	(if (or (not (list? spec)) (not (list? (cadr spec))) (< (length (cadr spec)) 3))
-	    (quit "syntax error in declaration: `~S'" spec)
+	    (syntax-error "invalid declaration" spec)
 	    (process-custom-declaration (cadr spec) (cddr spec)) ) )
        ((c-options)
 	(emit-control-file-item `(c-options ,@(cdr spec))) )
@@ -1081,8 +1101,9 @@ EOF
 	   `(post-process ,@(map (cut string-substitute "\\$@" file <>) (cdr spec))) ) ) )
        ((block) (set! block-compilation #t))
        ((separate) (set! block-compilation #f))
-       ((check-c-syntax) (set! no-c-syntax-checks #f))
        ((keep-shadowed-macros) (set! undefine-shadowed-macros #f))
+       ((unused)
+	(set! unused-variables (append (cdr spec) unused-variables)))
        ((not)
 	(check-decl spec 1)
 	(case (second spec)
@@ -1110,7 +1131,6 @@ EOF
 	   (check-decl spec 1 1)
 	   (case (cadr spec)
 	     [(interrupts-enabled) (set! insert-timer-checks #f)]
-	     [(check-c-syntax) (set! no-c-syntax-checks #t)]
 	     [(safe) 
 	      (set! unsafe #t)
 	      (##match#set-error-control #:unspecified) ]
@@ -1125,16 +1145,10 @@ EOF
 	(let ((syms (cdr spec)))
 	  (set! block-globals (lset-difference eq? block-globals syms))
 	  (set! export-list (lset-union eq? syms (or export-list '())))))
-       ((compress-literals)
-	(set! literal-compression-threshold 
-	  (or (and (pair? (cdr spec)) (number? (cadr spec)) (cadr spec))
-	      default-literal-compression-threshold) ) 
-	(when (and (list? spec) (= 3 (length spec)))
-	  (set! compressed-literals-initializer (third spec)) ) )
        ((emit-exports)
-	(if (null? (cdr spec))
-	    (quit "invalid `emit-exports' declaration" spec)
-	    (set! export-file-name (cadr spec)) ) )
+	(cond ((null? (cdr spec))
+	       (quit "invalid `emit-exports' declaration" spec) )
+	      ((not export-file-name) (set! export-file-name (cadr spec))) ) )
        ((emit-external-prototypes-first)
 	(set! external-protos-first #t) )
        ((lambda-lift) (set! do-lambda-lifting #t))
@@ -1150,15 +1164,6 @@ EOF
 	    (if (number? n)
 		n
 		(quit "invalid argument to `inline-limit' declaration" spec) ) ) ) )
-       ((namespace)
-	(check-decl spec 2)
-	(let* ([syms (cdr spec)]
-	       [ns (car syms)] )
-	  (if (every symbol? syms)
-	      (let ([oldsyms (or (and-let* ([a (assq ns namespace-table)]) (cdr a)) '())])
-		(set! namespace-table 
-		  (alist-update! ns (lset-union eq? oldsyms (cdr syms)) namespace-table eq?) ) )
-	      (quit "invalid arguments to `namespace' declaration: ~S" spec) ) ) )
        ((constant)
 	(let ((syms (cdr spec)))
 	  (if (every symbol? syms)
@@ -1241,7 +1246,6 @@ EOF
 	 [body (apply string-append (map cadr (cdddr exp)))]
  	 [argtypes (map car args)]
 	 [argnames (map cadr args)] )
-    (check-c-syntax body 'foreign-lambda*)
     (create-foreign-stub rtype #f argtypes argnames body #f #f) ) )
 
 (define (expand-foreign-callback-lambda* exp)
@@ -1250,7 +1254,6 @@ EOF
 	 [body (apply string-append (map cadr (cdddr exp)))]
  	 [argtypes (map car args)]
 	 [argnames (map cadr args)] )
-    (check-c-syntax body 'foreign-callback-lambda*)
     (create-foreign-stub rtype #f argtypes argnames body #t #t) ) )
 
 (define (expand-foreign-primitive exp)
@@ -1260,21 +1263,18 @@ EOF
 	 [body (apply string-append (map cadr (if hasrtype (cdddr exp) (cddr exp))))]
  	 [argtypes (map car args)]
 	 [argnames (map cadr args)] )
-    (check-c-syntax body 'foreign-primitive)
     (create-foreign-stub rtype #f argtypes argnames body #f #t) ) )
 
 
 ;;; Traverse expression and update line-number db with all contained calls:
 
 (define (update-line-number-database! exp ln)
-
   (define (mapupdate xs)
     (let loop ((xs xs))
       (if (pair? xs)
 	  (begin
 	    (walk (car xs))
 	    (loop (cdr xs)) ) ) ) )
-
   (define (walk x)
     (cond ((not-pair? x))
 	  ((symbol? (car x))
@@ -1284,7 +1284,6 @@ EOF
 		 (##sys#hash-table-set! ##sys#line-number-database name (alist-cons x ln old)) )
 	     (mapupdate (cdr x)) ) )
 	  (else (mapupdate x)) ) )
-
   (walk exp) )
 
 
@@ -1643,7 +1642,10 @@ EOF
 
 	 ;; If this is the first analysis and the variable is global and has no references and we are
 	 ;;  in block mode, then issue warning:
-	 (when (and first-analysis global (null? references))
+	 (when (and first-analysis 
+		    global
+		    (null? references)
+		    (not (memq sym unused-variables)))
 	   (when assigned-locally
 	     (compiler-warning 'var "local assignment to unused variable `~S' may be unintended" sym) )
 	   (when (and (or block-compilation
@@ -1657,7 +1659,7 @@ EOF
 
 	 ;; Make 'contractable, if it has a procedure as known value, has only one use and one call-site and
 	 ;;  if the lambda has no free non-global variables or is an internal lambda. Make 'inlinable if
-	 ;;  use/call count is 1:
+	 ;;  use/call count is not 1:
 	 (when value
 	   (let ((valparams (node-parameters value)))
 	     (when (and (eq? '##core#lambda (node-class value))
@@ -1839,8 +1841,7 @@ EOF
 						     (= (length refs) (length sites)) 
 						     (proper-list? llist) ) ] )
 					  (when (and name custom (not (= (length llist) (length (cdr subs)))))
-					    (compiler-warning
-					     'call 
+					    (quit
 					     "known procedure called with wrong number of arguments: ~A" 
 					     (source-info->string name) ) )
 					  (register-direct-call! id)
@@ -2059,6 +2060,7 @@ EOF
   
 (define (prepare-for-code-generation node db)
   (let ([literals '()]
+	[lambda-info-literals '()]
         [lambdas '()]
         [temporaries 0]
         [allocated 0]
@@ -2250,11 +2252,11 @@ EOF
 
 	  ((quote)
 	   (let ((c (first params)))
-	     (cond ((fixnum? c)
+	     (cond ((and (fixnum? c) (not (big-fixnum? c)))
 		    (immediate-literal c) )
 		   ((number? c)
 		    (cond ((eq? 'fixnum number-type)
-			   (cond ((integer? c)
+			   (cond ((and (integer? c) (not (big-fixnum? c)))
 				  (compiler-warning 
 				   'type 
 				   "coerced inexact literal number `~S' to fixnum ~S" c (inexact->exact c))
@@ -2271,15 +2273,22 @@ EOF
 
     (define (literal x)
       (cond [(immediate? x) (immediate-literal x)]
-	    [(and (number? x) (inexact? x) 
-		  (list-index (lambda (y) (and (number? y) (inexact? y) (= x y))) literals) )
-	     => values]
-            [(posq x literals) => values]
+	    [(number? x)
+	     (or (and (inexact? x) 
+		      (list-index (lambda (y) (and (number? y) (inexact? y) (= x y)))
+				  literals) )
+		 (new-literal x)) ]
+	    ((##core#inline "C_lambdainfop" x)
+	     (let ((i (length lambda-info-literals)))
+	       (set! lambda-info-literals 
+		 (append lambda-info-literals (list x))) ;*** see below
+	       (vector i) ) )
+            [(posq x literals) => identity]
 	    [else (new-literal x)] ) )
 
     (define (new-literal x)
       (let ([i (length literals)])
-	(set! literals (append literals (list x))) ; could be optimized
+	(set! literals (append literals (list x))) ;*** could (should) be optimized
 	i) )
 
     (define (blockvar-literal var)
@@ -2291,18 +2300,20 @@ EOF
 	  (new-literal (make-block-variable-literal var)) ) )
     
     (define (immediate-literal x)
-      (make-node '##core#immediate
-		 (cond ((fixnum? x) `(fix ,x))
-		       ((boolean? x) `(bool ,x))
-		       ((char? x) `(char ,x))
-		       ((null? x) '(nil))
-		       ((eof-object? x) '(eof))
-		       (else (bomb "bad immediate (prepare)")) )
-		 '() ) )
+      (if (eq? (void) x)
+	  (make-node '##core#undefined '() '())
+	  (make-node '##core#immediate
+		     (cond ((fixnum? x) `(fix ,x))
+			   ((boolean? x) `(bool ,x))
+			   ((char? x) `(char ,x))
+			   ((null? x) '(nil))
+			   ((eof-object? x) '(eof))
+			   (else (bomb "bad immediate (prepare)")) )
+		     '() ) ) )
     
     (debugging 'p "preparation phase...")
     (let ((node2 (walk node '() #f '())))
       (debugging 'o "fast box initializations" fastinits)
       (debugging 'o "fast global references" fastrefs)
       (debugging 'o "fast global assignments" fastsets)
-      (values node2 literals lambdas) ) ) )
+      (values node2 literals lambda-info-literals lambdas) ) ) )
