@@ -1,6 +1,7 @@
 ;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Hen -*-
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,24 +23,13 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare
   (block)
-  (uses extras srfi-1 srfi-13 regex utils))
+  (uses data-structures ports srfi-1 srfi-13 utils extras ))
 
 #>
-#include "chicken-defaults.h"
-
 #ifndef C_TARGET_CC
 # define C_TARGET_CC  C_INSTALL_CC
 #endif
@@ -52,6 +42,10 @@
 # define C_TARGET_CFLAGS  C_INSTALL_CFLAGS
 #endif
 
+#ifndef C_TARGET_LDFLAGS
+# define C_TARGET_LDFLAGS  C_INSTALL_LDFLAGS
+#endif
+
 #ifndef C_TARGET_BIN_HOME
 # define C_TARGET_BIN_HOME  C_INSTALL_BIN_HOME
 #endif
@@ -71,6 +65,14 @@
 #ifndef C_TARGET_SHARE_HOME
 # define C_TARGET_SHARE_HOME  C_INSTALL_SHARE_HOME
 #endif
+
+#ifndef C_TARGET_RUN_LIB_HOME
+# define C_TARGET_RUN_LIB_HOME    C_TARGET_LIB_HOME
+#endif
+
+#ifndef C_CHICKEN_PROGRAM
+# define C_CHICKEN_PROGRAM     "chicken"
+#endif
 <#
 
 (define-foreign-variable INSTALL_BIN_HOME c-string "C_INSTALL_BIN_HOME")
@@ -80,6 +82,8 @@
 (define-foreign-variable TARGET_CXX c-string "C_TARGET_CXX")
 (define-foreign-variable TARGET_CFLAGS c-string "C_TARGET_CFLAGS")
 (define-foreign-variable INSTALL_CFLAGS c-string "C_INSTALL_CFLAGS")
+(define-foreign-variable TARGET_LDFLAGS c-string "C_TARGET_LDFLAGS")
+(define-foreign-variable INSTALL_LDFLAGS c-string "C_INSTALL_LDFLAGS")
 (define-foreign-variable INSTALL_MORE_LIBS c-string "C_INSTALL_MORE_LIBS")
 (define-foreign-variable INSTALL_MORE_STATIC_LIBS c-string "C_INSTALL_MORE_STATIC_LIBS")
 (define-foreign-variable INSTALL_SHARE_HOME c-string "C_INSTALL_SHARE_HOME")
@@ -93,13 +97,14 @@
 (define-foreign-variable TARGET_LIB_HOME c-string "C_TARGET_LIB_HOME")
 (define-foreign-variable TARGET_INCLUDE_HOME c-string "C_TARGET_INCLUDE_HOME")
 (define-foreign-variable TARGET_STATIC_LIB_HOME c-string "C_TARGET_STATIC_LIB_HOME")
+(define-foreign-variable TARGET_RUN_LIB_HOME c-string "C_TARGET_RUN_LIB_HOME")
+(define-foreign-variable CHICKEN_PROGRAM c-string "C_CHICKEN_PROGRAM")
 
 
 ;;; Parameters:
 
-(define win (eq? (build-platform) 'msvc))
-(define cygwin (eq? (build-platform) 'cygwin))
 (define mingw (eq? (build-platform) 'mingw32))
+(define msvc (eq? (build-platform) 'msvc))
 (define osx (eq? (software-version) 'macosx))
 (define hpux-hppa (and (eq? (software-version) 'hpux)
                        (eq? (machine-type) 'hppa)))
@@ -109,83 +114,61 @@
   (exit 64) )
 
 (define chicken-prefix (getenv "CHICKEN_PREFIX"))
-(define cmake-build (##sys#fudge 38))
 (define arguments (command-line-arguments))
 (define host-mode (member "-host" arguments))
+(define cross-chicken (##sys#fudge 39))
 
 (define (prefix str dir default)
   (if chicken-prefix
       (make-pathname (list chicken-prefix dir) str)
       default) )
 
-(define home
-  (or (getenv "CHICKEN_HOME") 
-      (if (and win (not cmake-build))
-	  (quit "`CHICKEN_HOME' environment variable not set - please set it to the directory where CHICKEN is installed")
-	  (quotewrap (prefix "" "share" (if host-mode INSTALL_SHARE_HOME TARGET_SHARE_HOME)))
-      )
-  )
-)
-
-(define (homize str) (make-pathname home str))
-
 (define (quotewrap str)
   (if (string-any char-whitespace? str)
       (string-append "\"" str "\"") 
       str) )
 
+(define home
+  (quotewrap 
+   (prefix "" "share" (if host-mode INSTALL_SHARE_HOME TARGET_SHARE_HOME))))
+
 (define translator
-  (if (and win (not cmake-build))
-      (quotewrap (homize "chicken"))
-      (quotewrap 
-       (prefix "chicken" "bin"
-	       (make-pathname
-		(if host-mode INSTALL_BIN_HOME TARGET_BIN_HOME)
-		"chicken")))))
-
-(if win
-    (begin
-      (define compiler "cl")
-      (define c++-compiler "cl")
-      (define linker "link")
-      (define c++-linker "link")
-      (define object-extension "obj")
-      (define library-extension "lib")
-      (define link-output-flag "/out:")
-      (define compile-output-flag "/Fo")
-      (define executable-extension "exe")
-      (define shared-library-extension "dll")
-      (define nonstatic-compilation-options '("/DPIC")) )
-    (begin
-      (define compiler (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
-      (define c++-compiler (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
-      (define linker (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
-      (define c++-linker (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
-      (define object-extension "o")
-      (define library-extension "a")
-      (define link-output-flag "-o ")
-      (define executable-extension "")
-      (define compile-output-flag "-o ")
-      (define nonstatic-compilation-options '())
-      (define shared-library-extension (cond ((or cygwin mingw) "dll")
-                                             ;(hpux-hppa "sl")
-                                             (else "so")))))
+  (quotewrap 
+   (prefix "chicken" "bin"
+	   (make-pathname
+	    (if host-mode INSTALL_BIN_HOME TARGET_BIN_HOME)
+	    CHICKEN_PROGRAM))))
 
+(define compiler (quotewrap (if host-mode INSTALL_CC TARGET_CC)))
+(define c++-compiler (quotewrap (if host-mode INSTALL_CXX TARGET_CXX)))
+(define linker (quotewrap (if msvc "link" (if host-mode INSTALL_CC TARGET_CC))))
+(define c++-linker (quotewrap (if msvc "link" (if host-mode INSTALL_CXX TARGET_CXX))))
+(define object-extension (if msvc "obj" "o"))
+(define library-extension (if msvc "lib" "a"))
+(define link-output-flag (if msvc "-out:" "-o "))
+(define executable-extension (if msvc "exe" ""))
+(define compile-output-flag (if msvc "-Fo" "-o "))
+(define nonstatic-compilation-options '())
+(define shared-library-extension ##sys#load-dynamic-extension)
 (define default-translation-optimization-options '())
+(define pic-options (if (or mingw msvc) '("-DPIC") '("-fPIC" "-DPIC")))
 
-(if win
-    (begin
-      (define (cleanup-filename s) (string-translate s "/" "\\")) ; we need this to please the MSVC tools
-      (define default-compilation-optimization-options '("/nologo"))
-      (define default-linking-optimization-options '("/nologo"))
-      (define best-linking-optimization-options '("/nologo"))
-      (define best-compilation-optimization-options '("/O2" "/nologo")) )
-    (begin
-      (define (cleanup-filename s) s)
-      (define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
-      (define best-compilation-optimization-options default-compilation-optimization-options)
-      (define default-linking-optimization-options '())
-      (define best-linking-optimization-options '()) ) )
+(define default-library (string-append
+                         (if msvc "libchicken-static." "libchicken.")
+                         library-extension))
+(define default-unsafe-library (string-append
+                                (if msvc "libuchicken-static." "libuchicken.")
+                                library-extension))
+
+(define cleanup-filename
+  (if (not mingw)
+      (lambda (s) (quotewrap s)) ; allow filenames w/ whitespace
+      (lambda (s) s)))
+
+(define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
+(define best-compilation-optimization-options default-compilation-optimization-options)
+(define default-linking-optimization-options (string-split (if host-mode INSTALL_LDFLAGS TARGET_LDFLAGS)))
+(define best-linking-optimization-options default-linking-optimization-options)
 
 (define-constant simple-options
   '(-explicit-use -no-trace -no-warnings -no-usual-integrations -optimize-leaf-routines -unsafe
@@ -193,14 +176,14 @@
     -check-syntax -case-insensitive -benchmark-mode -shared -run-time-macros -no-lambda-info
     -lambda-lift -dynamic -disable-stack-overflow-checks -emit-debug-info -check-imports
     -emit-external-prototypes-first -inline -extension -release -static-extensions
-    -analyze-only -keep-shadowed-macros) )
+    -analyze-only -keep-shadowed-macros -disable-compiler-macros) )
 
 (define-constant complex-options
   '(-debug -output-file -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
     -optimize-level -include-path -database-size -extend -prelude -postlude -prologue -epilogue 
     -inline-limit -profile-name -disable-warning -import -require-static-extension
     -feature -debug-level -heap-growth -heap-shrinkage -heap-initial-size -emit-exports
-    -compress-literals) )
+    -compress-literals) )		; DEPRECATED
 
 (define-constant shortcuts
   '((-h "-help")
@@ -241,103 +224,36 @@
 (define show-libs #f)
 (define dry-run #f)
 
-(if win
-    (begin
-      (define extra-libraries "")
-      (define extra-shared-libraries "")
-      (if (not cmake-build)
-        (begin				; flat destination directory
-          (define default-library-files
-	    (map homize '("libchicken-static.lib")))
-          (define default-shared-library-files
-	    (map homize '("libchicken.lib")))
-          (define unsafe-library-files
-	    (map homize '("libuchicken-static.lib")))
-          (define unsafe-shared-library-files
-	    (map homize '("libuchicken.lib")))
-          (define gui-library-files
-	    (map homize '("libchickengui-static.lib")))
-          (define gui-shared-library-files
-	    (map homize '("libchickengui.lib")))
-        )
-        (begin
-          (define default-library-files 
-	    (list
-	     (quotewrap 
-	      (prefix "libchicken-s.lib" "lib"
-		      (string-append
-		       (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		       "\\libchicken-s.lib"))))
-	    )
-	  (define default-shared-library-files (list
-            (quotewrap 
-	     (prefix "libchicken.lib" "lib"
-		     (string-append
-		      (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		      "\\libchicken.lib")))
-          ))
-          (define unsafe-library-files (list
-            (quotewrap
-	     (prefix "libuchicken-s.lib" "lib"
-		     (string-append
-		      (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		      "\\libuchicken-s.lib")))
-          ))
-          (define unsafe-shared-library-files (list
-            (quotewrap
-	     (prefix "libuchicken.lib" "lib"
-		     (string-append
-		      (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		      "\\libuchicken.lib")))
-          ))
-          (define gui-library-files (list
-            (quotewrap 
-	     (prefix "libchickengui-s.lib" "lib"
-		     (string-append
-		      (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		      "\\libchickengui-s.lib")))
-          ))
-          (define gui-shared-library-files (list
-            (quotewrap
-	     (prefix "libchickengui.lib" "lib"
-		     (string-append
-		      (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		      "\\libchickengui.lib")))
-          ))
-        )
-      )
-    )
-    (begin
-      ; will more libs ever need quotewrapping?
-      (define extra-libraries
-	(if host-mode
-	    INSTALL_MORE_STATIC_LIBS
-	    TARGET_MORE_STATIC_LIBS))
-      (define extra-shared-libraries 
-	(if host-mode 
-	    INSTALL_MORE_LIBS
-	    TARGET_MORE_LIBS))
-      (define default-library-files (list
-        (quotewrap
-	 (prefix "libchicken.a" "lib"
-		 (string-append
-		  (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		  "/libchicken.a")))
-      ))
-      (define default-shared-library-files '("-lchicken"))
-      (define unsafe-library-files (list
-        (quotewrap 
-	 (prefix "libuchicken.a" "lib"
-		 (string-append 
-		  (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
-		  "/libuchicken.a")))
-      ))
-      (define unsafe-shared-library-files '("-luchicken"))
-      (define gui-library-files default-library-files)
-      (define gui-shared-library-files default-shared-library-files)
-    )
-)
-
+(define extra-libraries
+  (if host-mode
+      INSTALL_MORE_STATIC_LIBS
+      TARGET_MORE_STATIC_LIBS))
+(define extra-shared-libraries 
+  (if host-mode 
+      INSTALL_MORE_LIBS
+      TARGET_MORE_LIBS))
+(define default-library-files 
+  (list
+   (quotewrap
+    (prefix default-library "lib"
+	    (string-append
+	     (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
+	     (string-append "/" default-library)))) ))
+(define default-shared-library-files (if msvc
+                                         (list (string-append "libchicken." library-extension))
+                                         '("-lchicken")))
+(define unsafe-library-files
+  (list
+   (quotewrap 
+    (prefix default-unsafe-library "lib"
+	    (string-append 
+	     (if host-mode INSTALL_LIB_HOME TARGET_LIB_HOME)
+	     (string-append "/" default-unsafe-library)))) ))
+(define unsafe-shared-library-files (if msvc
+                                        (list (string-append "libuchicken." library-extension))
+                                        '("-luchicken")))
+(define gui-library-files default-library-files)
+(define gui-shared-library-files default-shared-library-files)
 (define library-files default-library-files)
 (define shared-library-files default-shared-library-files)
 
@@ -349,37 +265,34 @@
     (and (not (member id '("/usr/include" "")))
 	 id) ) )
 
-(define compile-options
-  (if win
-      (if (not cmake-build)
-	  (cons* "/I%CHICKEN_HOME%" "/DC_NO_PIC_NO_DLL" (if (eq? (c-runtime) 'dynamic) '("/MD") '()))
-	  (cons* (string-append 
-		  "/I" 
-		  (if host-mode INSTALL_INCLUDE_HOME TARGET_INCLUDE_HOME)
-		 "/DC_NO_PIC_NO_DLL" (if (eq? (c-runtime) 'dynamic) '("/MD") '())) ) )
-      (if include-dir (list "-I" include-dir) '())) )
-
-(define compile-only-flag
-  (if win "/c" "-c") )
+(define compile-options '())
+(define builtin-compile-options
+  (if include-dir (list (conc "-I" (quotewrap include-dir))) '()))
 
+(define compile-only-flag "-c")
 (define translation-optimization-options default-translation-optimization-options)
 (define compilation-optimization-options default-compilation-optimization-options)
 (define linking-optimization-options default-linking-optimization-options)
 
-(define link-options
-  (cond [win '()]
-	[(or osx hpux-hppa) 
-	 (list (conc "-L" (quotewrap
-			   (prefix "" "lib"
-				   (if host-mode 
-				       INSTALL_LIB_HOME
-				       TARGET_LIB_HOME)) )))]
-	[else 
-	 (let ((p (quotewrap (prefix "" "lib"
-				     (if host-mode
-					 INSTALL_LIB_HOME
-					 TARGET_LIB_HOME)))))
-	   (list (conc "-L" p " -Wl,-R" p) ) ) ] ) )
+(define library-dir
+  (prefix "" "lib"
+         (if host-mode
+             INSTALL_LIB_HOME
+             TARGET_LIB_HOME)) )
+
+(define link-options '())
+(define builtin-link-options
+  (cond ((or osx hpux-hppa mingw)
+	 (list (conc "-L" (quotewrap library-dir))))
+        (msvc
+         (list (conc "-LIBPATH:" (quotewrap library-dir))))
+	(else 
+	 (list
+	  (conc "-L" (quotewrap library-dir))
+	  (conc " -Wl,-R" (quotewrap (prefix "" "lib"
+					     (if host-mode
+						 INSTALL_LIB_HOME
+						 TARGET_RUN_LIB_HOME)))) ) ) ) )
 
 (define target-filename #f)
 (define verbose #f)
@@ -431,7 +344,7 @@
   Syntax related options:
 
     -i -case-insensitive        don't preserve case of read symbols    
-    -K -keyword-style STYLE     allow alternative keyword syntax (prefix or suffix)
+    -K -keyword-style STYLE     allow alternative keyword syntax (prefix, suffix or none)
     -run-time-macros            macros are made available at run-time
 
   Translation options:
@@ -473,6 +386,7 @@
     -disable-stack-overflow-checks  disables detection of stack-overflows
     -inline                     enable inlining
     -inline-limit               set inlining threshold
+    -disable-compiler-macros    disable expansion of compiler macros
 
   Configuration options:
 
@@ -527,6 +441,7 @@
     -ldflags                    show required linker flags and exit
     -libs                       show required libraries and exit
     -cc-name                    show name of default C compiler used
+    -cxx-name                   show name of default C++ compiler used
     -ld-name                    show name of default linker used
     -dry-run                    just show commands executed, don't run them 
                                  (implies `-v')
@@ -535,7 +450,6 @@
 
     -debug MODES                display debugging output for the given modes
     -compiler PATHNAME          use other compiler than default `chicken'
-    -compress-literals NUMBER   compile literals above threshold as strings
     -disable-c-syntax-checks    disable syntax checks of C code fragments
     -raw                        do not generate implicit init- and exit code			       
     -emit-external-prototypes-first  emit protoypes for callbacks before foreign
@@ -570,18 +484,20 @@
 
   (define (shared-build lib)
     (set! translate-options (cons* "-feature" "chicken-compile-shared" translate-options))
-    (if win
-	(begin
-	  (set! compile-options (cons* "/DPIC" "/DC_SHARED" compile-options)) 
-	  (set! link-options (cons* "/dll" link-options)))
-	(begin
-	  (set! compile-options (cons* "-fPIC" "-DPIC" "-DC_SHARED" compile-options)) 
-	  (set! link-options
-	    (cons* "-fPIC" (if osx (if lib "-dynamiclib" "-bundle") "-shared") link-options))))
+    (set! compile-options (append pic-options '("-DC_SHARED") compile-options))
+    (set! link-options
+      (cons (cond
+             (osx (if lib "-dynamiclib" "-bundle"))
+             (msvc "-dll")
+             (else "-shared")) link-options))
     (set! shared #t) )
 
   (let loop ([args args])
     (cond [(null? args)
+           ;Builtin search directory options do not override explict options
+           (set! compile-options (append compile-options builtin-compile-options))
+           (set! link-options (append link-options builtin-link-options))
+           ;
 	   (when inquiry-only
 	     (when show-cflags (print* (compiler-options) #\space))
 	     (when show-ldflags (print* (linker-options) #\space))
@@ -606,8 +522,8 @@
 		  (unless target-filename
 		    (set! target-filename
 		      (if shared
-			  (pathname-replace-extension (last scheme-files) shared-library-extension)
-			  (pathname-replace-extension (last scheme-files) executable-extension) ) ) )
+			  (pathname-replace-extension (first scheme-files) shared-library-extension)
+			  (pathname-replace-extension (first scheme-files) executable-extension) ) ) )
 		  (run-translation) ] )
 	   (unless translate-only 
 	     (run-compilation)
@@ -615,7 +531,7 @@
 	       (when (member target-filename scheme-files)
 		 (printf "Warning: output file will overwrite source file `~A' - renaming source to `~A.old'~%"
 			 target-filename target-filename)
-		 (unless (zero? (system* (sprintf "mv ~A ~A.old" target-filename target-filename)))
+		 (unless (zero? ($system (sprintf "mv ~A ~A.old" target-filename target-filename)))
 		   (exit last-exit-code) ) )
 	       (run-linking)) ) ]
 	  [else
@@ -652,6 +568,7 @@
 		(set! inquiry-only #t)
 		(set! show-ldflags #t) ]
 	       [(-cc-name) (print compiler) (exit 0)]
+	       [(-cxx-name) (print c++-compiler) (exit 0)]
 	       [(-ld-name) (print linker) (exit 0)]
 	       [(-home) (print home) (exit 0)]
 	       [(-libs)
@@ -668,8 +585,9 @@
 	       [(-v3)
 		(set! verbose #t)
 		(t-options "-verbose")
-		(set! compile-options (cons "-v" compile-options))
-		(set! link-options (cons "-v" link-options)) ]
+                (if (not msvc)
+                    (set! compile-options (cons* "-v" "-Q" compile-options)))
+		(set! link-options (cons (if msvc "-VERBOSE" "-v") link-options)) ]
 	       [(|-A| -analyze-only)
 		(set! translate-only #t)
 		(t-options "-analyze-only") ]
@@ -689,9 +607,16 @@
 		(set! rest (cdr rest)) ]
 	       [(-windows |-W|)
 		(set! gui #t)
-		(when win
-		  (set! link-options (cons* "kernel32.lib" "user32.lib" "gdi32.lib" link-options))
-		  (set! compile-options (cons "-DC_WINDOWS_GUI" compile-options))) ]
+		(cond
+                 (mingw
+		  (set! link-options
+		    (cons* "-lkernel32" "-luser32" "-lgdi32" "-mwindows"
+			   link-options))
+		  (set! compile-options (cons "-DC_WINDOWS_GUI" compile-options)))
+                 (msvc
+                  (set! link-options
+                    (cons* "kernel32.lib" "user32.lib" "gdi32.lib" link-options))
+		  (set! compile-options (cons "-DC_WINDOWS_GUI" compile-options)))) ]
 	       [(-framework)
 		(check s rest)
 		(when osx 
@@ -739,7 +664,7 @@
 		(set! compile-options (append compile-options (string-split (car rest))))
 		(set! rest (cdr rest)) ]
 	       [(-strip)
-		(unless win (set! link-options (append link-options "-s")))]
+		(set! link-options (append link-options (list "-s")))]
 	       [(|-L|)
 		(check s rest)
 		(set! link-options (append link-options (string-split (car rest))))
@@ -754,6 +679,9 @@
 		  (set! link-options (append link-options (list (string-append "-Wl,-R" (car rest)))))
 		  (set! rest (cdr rest)) ) ]
 	       [(-host) #f]
+	       [(-) 
+		(set! target-filename (make-pathname #f "a" executable-extension))
+		(set! scheme-files (append scheme-files '("-")))]
 	       [else
 		(when (memq s '(-unsafe -benchmark-mode))
 		  (when (eq? s '-benchmark-mode)
@@ -778,10 +706,7 @@
 		      [(and (> (string-length arg) 1)
 			    (char=? #\- (string-ref arg 0)) )
 		       (cond [(char=? #\l (string-ref arg 1))
-			      (if win
-				  (set! object-files
-				    (append object-files (list (string-append (substring arg 2) ".lib"))) )
-				  (set! link-options (append link-options (list arg))) ) ]
+			      (set! link-options (append link-options (list arg))) ]
  			     [(char=? #\L (string-ref arg 1))
  			      (set! link-options (append link-options (list arg))) ]
  			     [(char=? #\I (string-ref arg 1))
@@ -802,20 +727,20 @@
 			     [else (quit "invalid option `~A'" s)] ) ]
 		      [(file-exists? arg)
 		       (let-values ([(dirs name ext) (decompose-pathname arg)])
-			 (cond [(not ext) (set! scheme-files (cons arg scheme-files))]
+			 (cond [(not ext) (set! scheme-files (append scheme-files (list arg)))]
 			       [(member ext '("h" "c"))
-				(set! c-files (cons arg c-files)) ]
+				(set! c-files (append c-files (list arg))) ]
 			       [(member ext '("cpp" "C" "cc" "cxx" "hpp"))
 				(when osx (set! compile-options (cons "-no-cpp-precomp" compile-options)))
 				(set! cpp-mode #t)
-				(set! c-files (cons arg c-files)) ]
+				(set! c-files (append c-files (list arg))) ]
 			       [(member ext '("m" "M" "mm"))
 				(set! objc-mode #t)
-				(set! c-files (cons arg c-files)) ]
+				(set! c-files (append c-files (list arg))) ]
 			       [(or (string=? ext object-extension)
 				    (string=? ext library-extension) )
-				(set! object-files (cons arg object-files)) ]
-			       [else (set! scheme-files (cons arg scheme-files))] ) ) ]
+				(set! object-files (append object-files (list arg))) ]
+			       [else (set! scheme-files (append scheme-files (list arg)))] ) ) ]
 		      [else
 		       (let ([f2 (string-append arg ".scm")])
 			 (if (file-exists? f2)
@@ -833,7 +758,7 @@
        (when (and (file-exists? cscf)
 		  (let ([x (with-input-from-file cscf read-line)])
 		    (or (eof-object? x) (string=? "#%eof" x)) ) )
-	 (delete-file* cscf) )
+	 ($delete-file cscf) )
        (let ([fc (pathname-replace-extension
 		  (if (= 1 (length scheme-files))
 		      target-filename
@@ -842,13 +767,13 @@
 			(objc-mode "m")
 			(else "c") ) ) ] )
 	 (unless (zero?
-		  (system* 
+		  ($system 
 		   (string-intersperse 
-		    (cons* translator f 
+		    (cons* translator (cleanup-filename f) 
 			   (append 
 			    (if to-stdout 
 				'("-to-stdout")
-				`("-output-file" ,fc) )
+				`("-output-file" ,(cleanup-filename fc)) )
 			    (if (or static static-libs static-extensions)
 				(map (lambda (e) (conc "-uses " e)) required-extensions)
 				'() )
@@ -864,38 +789,40 @@
 	       (for-each
 		(match-lambda
  		  [('post-process commands ...)
- 		   (for-each system* commands) ]
+ 		   (for-each $system commands) ]
  		  [('c-options opts ...)
  		   (set! compile-options (append compile-options opts)) ]
  		  [('link-options opts ...)
  		   (set! link-options (append link-options opts)) ]
 		  [x (error "invalid entry in csc control file" x)] )
 		(read-file) ) ) )
-	   (delete-file* cscf) ) ) ) )
-   (reverse scheme-files) )
-  (unless keep-files (for-each delete-file* generated-scheme-files)) )
+	   ($delete-file cscf) ) ) ) )
+   scheme-files)
+  (unless keep-files (for-each $delete-file generated-scheme-files)) )
 
 
 ;;; Compile all C files:
 
 (define (run-compilation)
-  (for-each
-   (lambda (f)
-     (let ([fo (pathname-replace-extension f object-extension)])
-       (unless (zero?
-		(system*
-		 (string-intersperse
-		  (list (cond (cpp-mode c++-compiler)
-			      (else compiler) )
-			(cleanup-filename f)
-			(string-append compile-output-flag (cleanup-filename fo)) 
-			compile-only-flag
-			(compiler-options) ) ) ) )
-	 (exit last-exit-code) )
-       (set! generated-object-files (cons fo generated-object-files))
-       (set! object-files (cons fo object-files)) ) )
-   (reverse c-files) )
-  (unless keep-files (for-each delete-file* generated-c-files)) )
+  (let ((ofiles '()))
+    (for-each
+     (lambda (f)
+       (let ([fo (pathname-replace-extension f object-extension)])
+	 (unless (zero?
+		  ($system
+		   (string-intersperse
+		    (list (cond (cpp-mode c++-compiler)
+				(else compiler) )
+			  (cleanup-filename f)
+			  (string-append compile-output-flag (cleanup-filename fo)) 
+			  compile-only-flag
+			  (compiler-options) ) ) ) )
+	   (exit last-exit-code) )
+	 (set! generated-object-files (cons fo generated-object-files))
+	 (set! ofiles (cons fo ofiles))))
+     c-files)
+    (set! object-files (append (reverse ofiles) object-files)) ; put generated object files first
+    (unless keep-files (for-each $delete-file generated-c-files)) ) )
 
 (define (compiler-options)
   (string-intersperse
@@ -909,27 +836,36 @@
 ;;; Link object files and libraries:
 
 (define (run-linking)
-  (let ([files (map cleanup-filename
-		    (append (reverse object-files)
-			    (nth-value 0 (static-extension-info))
-			    (if (or static static-libs)
-				(if gui gui-library-files library-files)
-				(if gui gui-shared-library-files shared-library-files) ) ) ) ] )
+  (let ((files (map cleanup-filename
+		    (append object-files
+			    (nth-value 0 (static-extension-info)) ) ) )
+	(target (cleanup-filename target-filename)))
     (unless (zero?
-	     (system*
+	     ($system
 	      (string-intersperse 
 	       (cons* (cond (cpp-mode c++-linker)
 			    (else linker) )
-		      (append 
+		      (append
 		       files
-		       (list (string-append link-output-flag (cleanup-filename target-filename)) 
+		       (list (string-append link-output-flag target)
 			     (linker-options)
 			     (linker-libraries #f) ) ) ) ) ) )
       (exit last-exit-code) )
-    (when (and win (not static) (not static-libs) (not shared))
-      (delete-file* (pathname-replace-extension target-filename "exp"))
-      (delete-file* (pathname-replace-extension target-filename "lib")) )
-    (unless keep-files (for-each delete-file* generated-object-files)) ) )
+    (when (and osx (or (not cross-chicken) host-mode))
+      (unless (zero? ($system 
+		      (string-append
+		       "install_name_tool -change libchicken.dylib "
+		       (quotewrap 
+			(make-pathname
+			 (prefix "" "lib"
+				 (if host-mode
+				     INSTALL_LIB_HOME
+				     TARGET_RUN_LIB_HOME))
+			 "libchicken.dylib") )
+		       " " 
+		       target) ) )
+	(exit last-exit-code) ) )
+    (unless keep-files (for-each $delete-file generated-object-files)) ) )
 
 (define (static-extension-info)
   (let ((rpath (repository-path)))
@@ -952,30 +888,24 @@
    (string-intersperse
     (append linking-optimization-options link-options
 	    (nth-value 1 (static-extension-info)) ) )
-   (if (and static (not win) (not osx)) " -static" "") ) )
+   (if (and static (not mingw) (not msvc) (not osx)) " -static" "") ) )
 
 (define (linker-libraries #!optional staticexts)
   (string-intersperse
    (append
     (if staticexts (nth-value 0 (static-extension-info)) '())
-    (cons
-     (if (or static static-libs) extra-libraries extra-shared-libraries)
-     (if (or static static-libs) 
-	 (if gui gui-library-files library-files)
-	 (if gui gui-shared-library-files shared-library-files) ) ) ) ) )
+    (if (or static static-libs)
+        (if gui gui-library-files library-files)
+        (if gui gui-shared-library-files shared-library-files))
+    (if (or static static-libs)
+        (list extra-libraries)
+        (list extra-shared-libraries)))))
 
 
 ;;; Helper procedures:
 
 (define-constant +hairy-chars+ '(#\\ #\#))
 
-(define (quote-option x)
-  (if (any (lambda (c)
-	     (or (char-whitespace? c) (memq c +hairy-chars+)) )
-	   (string->list x) )
-      (cleanup x)
-      x) )
-
 (define (cleanup s)
   (let* ((q #f)
 	 (s (list->string
@@ -991,23 +921,30 @@
 	(string-append "\"" (string-translate* s '(("\"" . "\\\""))) "\"")
 	s) ) )
 
+(define (quote-option x)
+  (if (any (lambda (c)
+	     (or (char-whitespace? c) (memq c +hairy-chars+)) )
+	   (string->list x) )
+      (cleanup x)
+      x) )
+
 (define last-exit-code #f)
 
-(define (system* str)
+(define ($system str)
   (when verbose (print str))
   (set! last-exit-code
     (if dry-run 
 	0
-	(system str)) )
+	(if (zero? (system str))
+	    0
+	    1)))
   (unless (zero? last-exit-code)
     (printf "*** Shell command terminated with exit status ~S: ~A~%" last-exit-code str) )
   last-exit-code)
 
-(define (delete-file* str)
+(define ($delete-file str)
   (when verbose 
-    (if win
-	(print "del " str) 
-	(print "rm " str) ) )
+    (print "rm " str) )
   (unless dry-run (delete-file str) ))
 
 
