@@ -1,6 +1,7 @@
-;;;; csi.scm - Simple interpreter stub for CHICKEN
+;;;; csi.scm - Interpreter stub for CHICKEN
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,18 +23,9 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
-(declare (uses match))
+(declare (uses match srfi-69 ports))
 
 (declare
   (usual-integrations)
@@ -46,29 +38,26 @@
 #else
 # define _getcwd(buf, len)       NULL
 #endif
-
-#ifdef __DJGPP__
-# include <unistd.h>
-#endif
 EOF
 ) )
 
 (include "chicken-more-macros")
 (include "banner")
-(include "parameters")
 
-#{csi 
+(private csi 
   print-usage print-banner
   run hexdump del 
   parse-option-string chop-separator lookup-script-file
   report describe dump hexdump bytevector-data get-config
   deldups tty-input?
   history-list history-count history-add history-ref
-  trace-indent trace-indent-level traced-procedure-entry traced-procedure-exit}
+  trace-indent trace-indent-level traced-procedure-entry traced-procedure-exit)
 
 (declare
-  (hide parse-option-string bytevector-data banner member* canonicalize-args do-trace do-untrace
-	traced-procedures describer-table
+  (always-bound
+    ##sys#windows-platform)
+  (hide parse-option-string bytevector-data member* canonicalize-args do-trace do-untrace
+	traced-procedures describer-table dirseparator?
 	findall trace-indent command-table do-break do-unbreak broken-procedures) )
 
 
@@ -90,8 +79,11 @@ EOF
 
     -h  -help  --help           display this text and exit
     -v  -version                display version and exit
+        -release                print release number and exit
     -i  -case-insensitive       enable case-insensitive reading
     -e  -eval EXPRESSION        evaluate given expression
+    -p  -print EXPRESSION       evaluate and print result(s)
+    -P  -pretty-print EXPRESSION  evaluate and print result(s) prettily
     -D  -feature SYMBOL         register feature identifier
     -q  -quiet                  do not print banner
     -n  -no-init                do not load initialization file `")
@@ -102,6 +94,7 @@ EOF
     -w  -no-warnings            disable all warnings
     -k  -keyword-style STYLE    enable alternative keyword-syntax (none, prefix or suffix)
     -s  -script PATHNAME        use interpreter for shell scripts
+        -ss PATHNAME            shell script with `main' procedure
     -R  -require-extension NAME require extension before executing code
     -I  -include-path PATHNAME  add PATHNAME to include path
     --                          ignore all following options
@@ -109,7 +102,7 @@ EOF
 ") )
 
 (define (print-banner)
-  (print banner (chicken-version #t) "\n" copyright) )
+  (print +banner+ (chicken-version #t) "\n") )
 
 
 ;;; Reader for REPL history:
@@ -121,22 +114,24 @@ EOF
     (lambda (char port)
       (cond [(or (char=? #\) char) (char-whitespace? char))
 	     `',(history-ref (fx- history-count 1)) ]
-	    [(char-numeric? char)
-	     (let ([n (read port)])
-	       (if (integer? n)
-		   `',(history-ref n)
-		   (##sys#error "invalid syntax in `#<number>' (REPL history)" n) ) ) ]
 	    [else (old-hook char port)] ) ) ) )
 
+(set! ##sys#sharp-number-hook
+  (lambda (port n)
+    `',(history-ref n) ) )
+
 
 ;;; Chop terminating separator from pathname:
 
+(define (dirseparator? c)
+  (or (char=? c #\\) (char=? c #\/)))
+
 (define chop-separator 
-  (let ([pds ##sys#pathname-directory-separator]
-	[substring substring] )
+  (let ([substring substring] )
     (lambda (str)
-      (let ([len (sub1 (##sys#size str))])
-	(if (and (fx> len 0) (char=? (string-ref str len) pds))
+      (let* ((len (sub1 (##sys#size str)))
+	     (c (string-ref str len)))
+	(if (and (fx> len 0) (dirseparator? c))
 	    (substring str 0 len)
 	    str) ) ) ) )
 
@@ -162,13 +157,13 @@ EOF
     (lambda (name)
       (let ([path (getenv "PATH")])
 	(and (> (##sys#size name) 0)
-	     (cond [(char=? ##sys#pathname-directory-separator (string-ref name 0)) (addext name)]
-		   [(string-index (lambda (c) (char=? c ##sys#pathname-directory-separator)) name)
+	     (cond [(dirseparator? (string-ref name 0)) (addext name)]
+		   [(string-index dirseparator? name)
 		    (and-let* ([p (_getcwd buf 256)])
-		      (addext (string-append (chop-separator p) (string ##sys#pathname-directory-separator) name)) ) ]
+		      (addext (string-append (chop-separator p) "/" name)) ) ]
 		   [(addext name)]
 		   [else
-		    (let ([name2 (string-append (string ##sys#pathname-directory-separator) name)])
+		    (let ([name2 (string-append "/" name)])
 		      (let loop ([ps (string-split path ";")])
 			(and (pair? ps)
 			     (let ([name2 (string-append (chop-separator (##sys#slot ps 0)) name2)])
@@ -293,8 +288,9 @@ EOF
 			((utr) (do-untrace (map string->symbol (string-split (read-line)))))
 			((br) (do-break (map string->symbol (string-split (read-line)))))
 			((ubr) (do-unbreak (map string->symbol (string-split (read-line)))))
+			((uba) (do-unbreak-all))
 			((breakall) 
-			 (set! ##sys#break-in-thread #t) ) 
+			 (set! ##sys#break-in-thread #f) ) 
 			((breakonly)
 			 (set! ##sys#break-in-thread (eval (read))) )
 			((info)
@@ -331,15 +327,16 @@ EOF
  ,du EXP           Dump data of expression EXP
  ,dur EXP N        Dump range
  ,q                Quit interpreter
- ,l FILENAME ...   Load file with given filename
- ,ln FILENAME ...  Load file and print result of each top-level expression
+ ,l FILENAME ...   Load one or more files
+ ,ln FILENAME ...  Load one or more files and print result of each top-level expression
  ,r                Show system information
  ,s TEXT ...       Execute shell-command
  ,tr NAME ...      Trace procedures
  ,utr NAME ...     Untrace procedures
  ,br NAME ...      Set breakpoints
  ,ubr NAME ...     Remove breakpoints
- ,breakall         Break in all threads
+ ,uba              Remove all breakpoints
+ ,breakall         Break in all threads (default)
  ,breakonly THREAD Break only in specified thread
  ,c                Continue from breakpoint
  ,info             List traced procedures and breakpoints
@@ -481,6 +478,13 @@ EOF
 		(set! broken-procedures (del p broken-procedures eq?) ) ) ) ) )
      names) ) )
 
+(define do-unbreak-all
+  (lambda ()
+    (for-each (lambda (bp)
+                (##sys#setslot (car bp) 0 (cdr bp)))
+              broken-procedures)
+    (set! broken-procedures '())
+    (##sys#void)))
 
 ;;; Parse options from string:
 
@@ -501,10 +505,11 @@ EOF
 ;;; Print status information:
 
 (define report
-  (let ([printf printf]
+  (let ((printf printf)
 	(chop chop)
-	[with-output-to-port with-output-to-port] 
-	[current-output-port current-output-port] )
+	(sort sort)
+	(with-output-to-port with-output-to-port)
+	(current-output-port current-output-port) )
     (lambda port
       (with-output-to-port (if (pair? port) (car port) (current-output-port))
 	(lambda ()
@@ -515,9 +520,12 @@ EOF
 	    (printf "Features:")
 	    (for-each
 	     (lambda (lst) 
-	       (display "\n\t")
-	       (for-each (cut print* #\space <>) lst) )
-	     (chop (map keyword->string ##sys#features) 10) )
+	       (display "\n  ")
+	       (for-each 
+		(lambda (f)
+		  (printf "~a~a" f (make-string (fxmax 1 (fx- 16 (string-length f))) #\space)) )
+		lst) )
+	     (chop (sort (map keyword->string ##sys#features) string<?) 5))
 	    (printf "~%~
                    Machine type:    \t~A ~A~%~
                    Software type:   \t~A~%~
@@ -578,19 +586,28 @@ EOF
       (let ([out (:optional port ##sys#standard-output)])
 
 	(define (descseq name plen pref start)
-	  (##sys#call-with-current-continuation
-	   (lambda (return)
-	     (let ([len (fx- (plen x) start)])
-	       (when name (fprintf out "~A of length ~S~%" name len))
-	       (do ([i 0 (fx+ i 1)])
-		   ((fx>= i len))
-		 (when (fx>= i max-describe-lines)
-		   (fprintf out "~% (~A elements not displayed)~%" (fx- len i))
-		   (return #f) )
-		 (fprintf out " ~S: ~S~%" i (pref x (fx+ start i))) ) ) ) ) )
+	  (let ((len (fx- (plen x) start)))
+	    (when name (fprintf out "~A of length ~S~%" name len))
+	    (let loop1 ((i 0))
+	      (cond ((fx>= i len))
+		    ((fx>= i max-describe-lines)
+		     (fprintf out "~% (~A elements not displayed)~%" (fx- len i)) )
+		    (else
+		     (let ((v (pref x (fx+ start i))))
+		       (let loop2 ((n 1) (j (fx+ i (fx+ start 1))))
+			 (cond ((fx>= j len)
+				(fprintf out " ~S: ~S" i v)
+				(if (fx> n 1)
+				    (fprintf out "\t(followed by ~A identical instance~a)~% ...~%" 
+					     (fx- n 1)
+					     (if (eq? n 2) "" "s"))
+				    (newline out) )
+				(loop1 (fx+ i n)) )
+			       ((eq? v (pref x j)) (loop2 (fx+ n 1) (fx+ j 1)))
+			       (else (loop2 n len)) ) ) ) ) ) ) ) )
 
 	(when (##sys#permanent? x)
-	  (fprintf out "statically allocated (~X) " (##sys#block-address x)) )
+	  (fprintf out "statically allocated (0x~X) " (##sys#block-address x)) )
 	(cond [(char? x)
 	       (let ([code (char->integer x)])
 		 (fprintf out "character ~S, code: ~S, #x~X, #o~O~%" x code code code) ) ]
@@ -620,7 +637,7 @@ EOF
 	       (let ([len (##sys#size x)])
 		 (if (and (> len 3)
 			  (memq #:tinyclos ##sys#features)
-			  (eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) )
+			  (eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) ) ;XXX handle this in tinyclos egg (difficult)
 		     (describe-object x out)
 		     (descseq 
 		      (sprintf "procedure with code pointer ~X" (##sys#peek-unsigned-integer x 0))
@@ -632,7 +649,7 @@ EOF
 			(##sys#slot x 7)
 			(##sys#slot x 3)
 			(##sys#peek-unsigned-integer x 0) ) ]
-	      [(and (memq #:tinyclos ##sys#features) (instance? x))
+	      [(and (memq #:tinyclos ##sys#features) (instance? x)) ; XXX put into tinyclos egg
 	       (describe-object x out) ]
 	      [(##sys#locative? x)
 	       (fprintf out "locative~%  pointer ~X~%  index ~A~%  type ~A~%"
@@ -652,7 +669,7 @@ EOF
 	      [(##sys#pointer? x) (fprintf out "machine pointer ~X~%" (##sys#peek-unsigned-integer x 0))]
 	      [(##sys#bytevector? x)
 	       (let ([len (##sys#size x)])
-		 (fprintf out "byte vector of size ~S:~%" len)
+		 (fprintf out "blob of size ~S:~%" len)
 		 (hexdump x len ##sys#byte out) ) ]
 	      [(##core#inline "C_lambdainfop" x)
 	       (fprintf out "lambda information: ~s~%" (##sys#lambda-info->string x)) ]
@@ -664,20 +681,6 @@ EOF
 	       (hash-table-walk
 		x
 		(lambda (k v) (fprintf out " ~S\t-> ~S~%" k v)) ) ]
-	      [(##sys#structure? x 'array)
-	       (let* ([shp (##sys#slot x 3)]
-		      [vec (##sys#slot x 1)] 
-		      (n (##sys#size vec)) )
-		 (fprintf out "array of rank ~S with ~A element~a~% shape: "
-			  (fx/ (##sys#size shp) 2)
-			  n
-			  (if (fx= n 1) "" "s") )
-		 (let loop ([s (vector->list shp)])
-		   (unless (null? s)
-		     (let ([next (##sys#slot s 1)])
-		       (fprintf out " ~A..~A " (##sys#slot s 0) (##sys#slot next 0))
-		       (loop (##sys#slot next 1)) ) ) )
-		 (##sys#write-char-0 #\newline out) ) ]
 	      [(##sys#structure? x 'condition)
 	       (fprintf out "condition: ~s~%" (##sys#slot x 1))
 	       (for-each
@@ -689,7 +692,7 @@ EOF
 			(fprintf out "\t~s: ~s~%" (cdar props) (cadr props)) )
 		      (loop (cddr props)) ) ) )
 		(##sys#slot x 1) ) ]
-	      [(and (##sys#structure? x 'meroon-instance) (provided? 'meroon))
+	      [(and (##sys#structure? x 'meroon-instance) (provided? 'meroon)) ; XXX put this into meroon egg (really!)
 	       (unveil x out) ]
 	      [(##sys#generic-structure? x)
 	       (let ([st (##sys#slot x 0)])
@@ -767,53 +770,6 @@ EOF
 	(##sys#write-char-0 #\newline out) ) ) ) )
 
 
-;;; emacs interface (suggested by Linh Dang)
-
-(define ##csi#find-symbol-table (foreign-lambda c-pointer "C_find_symbol_table" c-string))
-(define ##csi#enum-symbols! (foreign-lambda scheme-object "C_enumerate_symbols" c-pointer scheme-object))
-
-(define (##csi#oblist)
-  (let ([it (cons -1 '())]
-	[ns (##csi#find-symbol-table ".")] )
-    (let loop ([lst '()])
-      (let ([s (##csi#enum-symbols! ns it)])
-	(if s
-	    (loop (cons s lst))
-	    lst) ) ) ) )
-
-(define (##csi#oblist->strings)
-  (let ([it (cons -1 '())]
-	[ns (##csi#find-symbol-table ".")] )
-    (let loop ([lst '()])
-      (let ([s (##csi#enum-symbols! ns it)])
-	(if s
-	    (loop (cons (->string s) lst))
-	    lst) ) ) ) )
-
-(define (##csi#name-of-symbols-starting-with prefix)
-  (let ([it (cons -1 '())]
-	[ns (##csi#find-symbol-table ".")]
-	[re (string-append "^" prefix)] )
-    (let loop ([lst '()])
-      (let ([s (##csi#enum-symbols! ns it)])
-	(if s 
-	    (loop (if (string-search re (symbol->string s))
-		      (cons (symbol->string s) lst)
-		      lst))
-	    lst) ) ) ) )
-
-(define (##csi#symbols-matching re-string)
-  (let ([it (cons -1 '())]
-	[ns (##csi#find-symbol-table ".")] )
-    (let loop ([lst '()])
-      (let ([s (##csi#enum-symbols! ns it)])
-	(if s 
-	    (loop (if (string-search re-string (symbol->string s))
-		      (cons s lst)
-		      lst))
-	    lst) ) ) ) )
-
-
 ;;; Start interpreting:
 
 (define (deldups lis . maybe-=)
@@ -834,28 +790,32 @@ EOF
 		 (else (find (cdr ks))) ) ) ) ) )
 
 (define-constant short-options 
-  '(#\k #\s #\v #\h #\D #\e #\i #\R #\b #\n #\q #\w #\- #\I) )
+  '(#\k #\s #\v #\h #\D #\e #\i #\R #\b #\n #\q #\w #\- #\I #\p #\P) )
 
 (define-constant long-options
-  '("-keyword-style" "-script" "-version" "-help" "--help" "--" "-feature" "-eval" "-case-insensitive"
-    "-require-extension" "-batch" "-quiet" "-no-warnings" "-no-init" "-include-path") )
+  '("-keyword-style" "-script" "-version" "-help" "--help" "--" "-feature" 
+    "-eval" "-case-insensitive"
+    "-require-extension" "-batch" "-quiet" "-no-warnings" "-no-init" 
+    "-include-path" "-release" "-ss"
+    "-print" "-pretty-print") )
 
 (define (canonicalize-args args)
   (let loop ((args args))
     (if (null? args)
 	'()
 	(let ((x (car args)))
-	  (cond ((or (string=? "-s" x) (string=? "-script" x)) args)
-		((and (fx> (##sys#size x) 2)
-		      (char=? #\- (##core#inline "C_subchar" x 0))
-		      (not (member x long-options)) )
-		 (if (char=? #\: (##core#inline "C_subchar" x 1))
-		     (loop (cdr args))
-		     (let ((cs (string->list (substring x 1))))
-		       (if (findall cs short-options)
-			   (append (map (cut string #\- <>) cs) (loop (cdr args)))
-			   (##sys#error "invalid option" x) ) ) ) )
-		(else (cons x (loop (cdr args)))))))))
+	  (cond 
+	   ((member x '("-s" "-ss" "-script" "--")) args)
+	   ((and (fx> (##sys#size x) 2)
+		 (char=? #\- (##core#inline "C_subchar" x 0))
+		 (not (member x long-options)) )
+	    (if (char=? #\: (##core#inline "C_subchar" x 1))
+		(loop (cdr args))
+		(let ((cs (string->list (substring x 1))))
+		  (if (findall cs short-options)
+		      (append (map (cut string #\- <>) cs) (loop (cdr args)))
+		      (##sys#error "invalid option" x) ) ) ) )
+	   (else (cons x (loop (cdr args)))))))))
 
 (define (findall chars clist)
   (let loop ((chars chars))
@@ -865,28 +825,30 @@ EOF
 
 (define (run)
   (let* ([extraopts (parse-option-string (or (getenv "CSI_OPTIONS") ""))]
-	 [args (canonicalize-args (cdr (argv)))]
-	 [loadlater #f]
+	 [args (canonicalize-args (command-line-arguments))]
 	 [kwstyle (member* '("-k" "-keyword-style") args)]
-	 [script (member* '("-s" "-script") args)])
+	 [script (member* '("-s" "-ss" "-script") args)])
     (cond [script
-	   (unless (pair? (cdr script)) (##sys#error "missing argument to `-script' option"))
+	   (when (or (not (pair? (cdr script)))
+		     (zero? (string-length (cadr script)))
+		     (char=? #\- (string-ref (cadr script) 0)) )
+	     (##sys#error "missing or invalid script argument"))
+	   (program-name (cadr script))
 	   (command-line-arguments (cddr script))
 	   (register-feature! 'script)
 	   (set-cdr! (cdr script) '()) 
-	   (when (and (memq (software-type) '(windows msdos))
-		      (not (eq? (build-platform) 'cygwin)) )
-	     (and-let* ([sname (lookup-script-file (cadr script))])
+	   (when ##sys#windows-platform
+	     (and-let* ((sname (lookup-script-file (cadr script))))
 	       (set-car! (cdr script) sname) ) ) ]
 	  [else
 	   (set! args (append (canonicalize-args extraopts) args))
 	   (and-let* ([p (member "--" args)])
 	     (set-cdr! p '()) ) ] )
-    (let* ([eval? (member* '("-e" "-eval") args)]
+    (let* ([eval? (member* '("-e" "-p" "-P" "-eval" "-print" "-pretty-print") args)]
 	   [batch (or script (member* '("-b" "-batch") args) eval?)]
-	   [quiet (or script (member* '("-q" "-quiet") args) eval?)]
-	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )
-      
+	   [quietflag (member* '("-q" "-quiet") args)]
+	   [quiet (or script quietflag eval?)]
+	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )      
       (define (collect-options opt)
 	(let loop ([opts args])
 	  (cond [(member opt opts) 
@@ -895,22 +857,29 @@ EOF
 			  (##sys#error "missing argument to command-line option" opt)
 			  (cons (cadr p) (loop (cddr p)))) ) ]
 		[else '()] ) ) )
-
       (define (loadinit)
 	(let ([fn (##sys#string-append "./" init-file)])
 	  (if (file-exists? fn)
 	      (load fn)
 	      (let* ([prefix (chop-separator (or (getenv "HOME") "."))]
-		     [fn (string-append prefix (string ##sys#pathname-directory-separator) init-file)] )
+		     [fn (string-append prefix "/" init-file)] )
 		(when (file-exists? fn) 
 		  (load fn) ) ) ) ) )
-
+      (define (evalstring str #!optional (rec (lambda _ (void))))
+	(let ((in (open-input-string str)))
+	  (do ([x (read in) (read in)])
+	      ((eof-object? x))
+	    (rec (receive (eval x))) ) ) )
+      (when quietflag (set! ##sys#eval-debug-level 0))
       (when (member* '("-h" "-help" "--help") args)
 	(print-usage)
 	(exit 0) )
       (when (member* '("-v" "-version") args)
 	(print-banner)
 	(exit 0) )
+      (when (member "-release" args)
+	(print (chicken-version))
+	(exit 0) )
       (when (member* '("-w" "-no-warnings") args)
 	(unless quiet (display "Warnings are disabled\n"))
 	(set! ##sys#warnings-enabled #f) )
@@ -944,7 +913,6 @@ EOF
       (unless (or (member* '("-n" "-no-init") args) script) (loadinit))
       (do ([args args (cdr args)])
 	  ((null? args)
-	   (when loadlater (load loadlater))
 	   (unless batch 
 	     (repl)
 	     (##sys#write-char-0 #\newline ##sys#standard-output) ) )
@@ -952,19 +920,36 @@ EOF
 	       [len (string-length arg)] )
 	  (cond ((member 
 		  arg 
-		  '("--" "-batch" "-quiet" "-no-init" "-no-warnings" "-script" "-b" "-q" "-n" "-w" "-s" "-i"
-		    "-case-insensitive") ) )
-		((member arg '("-feature" "-include-path" "-keyword-style" "-D" "-I" "-k"))
+		  '("--" "-batch" "-quiet" "-no-init" "-no-warnings" "-script"
+		    "-b" "-q" "-n" "-w" "-s" "-i"
+		    "-case-insensitive" "-ss") ) )
+		((member arg '("-feature" "-include-path" "-keyword-style" 
+			       "-D" "-I" "-k"))
 		 (set! args (cdr args)) )
 		((or (string=? "-R" arg) (string=? "-require-extension" arg))
 		 (eval `(##core#require-extension ',(string->symbol (cadr args))))
 		 (set! args (cdr args)) )
 		((or (string=? "-e" arg) (string=? "-eval" arg))
-		 (let ([in (open-input-string (cadr args))])
-		   (do ([x (read in) (read in)])
-		       ((eof-object? x))
-		     (eval x) )
-		   (set! args (cdr args)) ) )
-		(else (load arg) ) ) ) ) ) ) )
+		 (evalstring (cadr args))
+		 (set! args (cdr args)) )
+		((or (string=? "-p" arg) (string=? "-print" arg))
+		 (evalstring 
+		  (cadr args)
+		  (cut for-each print <...>) )
+		 (set! args (cdr args)) )
+		((or (string=? "-P" arg) (string=? "-pretty-print" arg))
+		 (evalstring 
+		  (cadr args)
+		  (cut for-each pretty-print <...>) )
+		 (set! args (cdr args)) )
+		(else
+		 (load arg) 
+		 (when (and script (string=? "-ss" (car script)))
+		   (call-with-values (cut main (command-line-arguments))
+		     (lambda results
+		       (exit
+			(if (and (pair? results) (fixnum? (car results)))
+			    (car results)
+			    0) ) ) ) ) ) ) ) ) ) ) )
 
 (run)
