@@ -1,6 +1,7 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,29 +23,20 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare (unit optimizer))
 
-#{compiler
+(private compiler
   compiler-arguments process-command-line perform-lambda-lifting!
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home decompose-lambda-list external-to-pointer
-  foreign-type-table-size copy-node! export-list inline-list not-inline-list
+  copy-node! export-list inline-list not-inline-list
   unit-name insert-timer-checks used-units external-variables
   debug-info-index debug-info-vector-name profile-info-vector-name
-  foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
+  foreign-declarations emit-trace-info block-compilation line-number-database-size
   always-bound-to-procedure block-globals make-block-variable-literal block-variable-literal? block-variable-literal-name
   target-heap-size target-stack-size constant-declarations
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
@@ -58,7 +50,7 @@
   reorganize-recursive-bindings substitution-table simplify-named-call compiler-warning
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! expand-foreign-callback-lambda* debug-lambda-list debug-variable-list debugging
-  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -72,14 +64,15 @@
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   make-random-name final-foreign-type inline-max-size simplified-ops
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result}
+  foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result)
 
 
 (eval-when (compile eval)
   (match-error-control #:fail) )
 
 (include "tweaks")
-(include "parameters")
+
+(define-constant maximal-number-of-free-variables-for-liftable 16)
 
 
 ;;; Scan toplevel expressions for assignments:
@@ -851,6 +844,10 @@
 
 (define (simplify-named-call db params name cont class classargs callargs)
   (define (test sym prop) (get db sym prop))
+  (define (defarg x)
+    (cond ((symbol? x) (varnode x))
+	  ((and (pair? x) (eq? 'quote (car x))) (qnode (cadr x)))
+	  (else (qnode x))))
 
   (case class
 
@@ -893,7 +890,7 @@
     ((3) ; classargs = (<var>)
      (and inline-substitutions-enabled
 	  (null? callargs)
-	  (test name 'standard-binding)
+	  (or (test name 'standard-binding) (test name 'extended-binding))
 	  (make-node '##core#call '(#t) (list cont (varnode (first classargs)))) ) )
 
     ;; (<op> a b) -> (<primitiveop> a (quote <i>) b)
@@ -1216,6 +1213,32 @@
 			(list (second classargs) w)
 			callargs) ) ) ) ) ) )
 
+    ;; (<op> <arg1> ... <argN>) -> (<primitiveop> ...)
+    ;; (<op> <arg1> ... <argN-I> <defargN-I>) -> (<primitiveop> ...)
+    ;; - default args in classargs should be either symbol or (optionally) 
+    ;;   quoted literal
+    ((23) ; classargs = (<minargc> <primitiveop> <literal1>|<varable1> ...)
+     (and inline-substitutions-enabled
+	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (let ([argc (first classargs)])
+	    (and (>= (length callargs) (first classargs))
+		 (make-node 
+		  '##core#call (list #t (second classargs))
+		  (cons*
+		   (varnode (second classargs))
+		   cont
+		   (let-values (((req opt) (split-at callargs argc)))
+		     (append
+		      req
+		      (let loop ((ca opt) 
+				 (da (cddr classargs)) )
+			(cond ((null? ca)
+			       (if (null? da)
+				   '()
+				   (cons (defarg (car da)) (loop '() (cdr da))) ) )
+			      ((null? da) '())
+			      (else (cons (car ca) (loop (cdr ca) (cdr da))))))))))))))
+
     (else (bomb "bad type (optimize)")) ) )
 
 
@@ -1369,9 +1392,9 @@
 			     (set! ksites (alist-cons #f n ksites))
 			     (cond [(eq? kvar (first arg0p))
 				    (unless (= argc (length (cdr subs)))
-				      (compiler-warning
-				       'call
-				       "known procedure called recursively with wrong number of arguments: `~A'" fnvar) )
+				      (quit
+				       "known procedure called recursively with wrong number of arguments: `~A'" 
+				       fnvar) )
 				    (node-class-set! n '##core#recurse)
 				    (node-parameters-set! n (list #t id))
 				    (node-subexpressions-set! n (cddr subs)) ]
@@ -1380,9 +1403,9 @@
 					 (let* ([klam (cdr a)]
 						[kbody (first (node-subexpressions klam))] )
 					   (unless (= argc (length (cdr subs)))
-					     (compiler-warning
-					      'call
-					      "known procedure called recursively with wrong number of arguments: `~A'" fnvar) )
+					     (quit
+					      "known procedure called recursively with wrong number of arguments: `~A'" 
+					      fnvar) )
 					   (node-class-set! n 'let)
 					   (node-parameters-set! n (take (third (node-parameters klam)) 1))
 					   (node-subexpressions-set!
@@ -1412,9 +1435,9 @@
 	      (let* ([n (cdr site)]
 		     [nsubs (node-subexpressions n)] )
 		(unless (= argc (length (cdr nsubs)))
-		  (compiler-warning 
-		   'call 
-		   "known procedure called with wrong number of arguments: `~A'" fnvar) )
+		  (quit
+		   "known procedure called with wrong number of arguments: `~A'"
+		   fnvar) )
 		(node-subexpressions-set!
 		 n
 		 (list (second nsubs)
