@@ -1,6 +1,7 @@
 ; scheduler.scm - Basic scheduler for multithreading
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,15 +23,6 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare
@@ -44,7 +36,8 @@
 	##sys#update-thread-state-buffer ##sys#restore-thread-state-buffer
 	##sys#remove-from-ready-queue ##sys#unblock-threads-for-i/o ##sys#force-primordial
 	##sys#fdset-input-set ##sys#fdset-output-set ##sys#fdset-clear
-	##sys#fdset-select-timeout ##sys#fdset-restore) 
+	##sys#fdset-select-timeout ##sys#fdset-restore
+	##sys#clear-i/o-state-for-thread!) 
   (foreign-declare #<<EOF
 #ifdef HAVE_ERRNO_H
 # include <errno.h>
@@ -114,6 +107,7 @@ EOF
     (##sys#update-thread-state-buffer ct)
     ;; Put current thread on ready-queue:
     (when (or (eq? cts 'running) (eq? cts 'ready)) ; should ct really be 'ready? - normally not.
+      (##sys#setislot ct 13 #f)			   ; clear timeout-unblock flag
       (##sys#add-to-ready-queue ct) )
     (let loop1 ()
       ;; Unblock threads waiting for timeout:
@@ -130,6 +124,9 @@ EOF
 		  (if (eq? tmo1 tmo2)
 		      (if (>= now tmo1)
 			  (begin
+			    (##sys#setislot tto 13 #t) ; mark as being unblocked by timeout
+			    (##sys#clear-i/o-state-for-thread! tto)
+			    ;;(pp `(CLEARED: ,tto ,@##sys#fd-list) ##sys#standard-error) ;***
 			    (##sys#thread-basic-unblock! tto)
 			    (loop (cdr lst)) )
 			  (begin
@@ -225,6 +222,7 @@ EOF
 	    (set! ##sys#timeout-list (cons (cons tm t) tl)) )
 	(loop (cdr tl) tl) ) ) 
   (##sys#setslot t 3 'blocked)
+  (##sys#setislot t 13 #f)
   (##sys#setislot t 4 tm) )
 
 (define (##sys#thread-block-for-termination! t t2)
@@ -233,6 +231,7 @@ EOF
     (unless (or (eq? state 'dead) (eq? state 'terminated))
       (##sys#setslot t2 12 (cons t (##sys#slot t2 12)))
       (##sys#setslot t 3 'blocked) 
+      (##sys#setislot t 13 #f)
       (##sys#setslot t 11 t2) ) ) )
 
 (define (##sys#thread-kill! t s)
@@ -296,7 +295,7 @@ EOF
   (foreign-lambda* int ([bool to] [unsigned-long tm])
     "struct timeval timeout;"
     "timeout.tv_sec = tm / 1000;"
-    "timeout.tv_usec = tm % 1000;"
+    "timeout.tv_usec = (tm % 1000) * 1000;"
     "C_fdset_input_2 = C_fdset_input;"
     "C_fdset_output_2 = C_fdset_output;"
     "return(select(FD_SETSIZE, &C_fdset_input, &C_fdset_output, NULL, to ? &timeout : NULL));") )
@@ -332,10 +331,14 @@ EOF
 	  (if (fx= fd (car a)) 
 	      (##sys#setslot a 1 (cons t (cdr a)))
 	      (loop (cdr lst)) ) ) ) )
-  (if i/o
-      (##sys#fdset-input-set fd)
-      (##sys#fdset-output-set fd) )
+  (case i/o
+    ((#t #:input) (##sys#fdset-input-set fd))
+    ((#f #:output) (##sys#fdset-output-set fd))
+    ((#:all)
+     (##sys#fdset-input-set fd)
+     (##sys#fdset-output-set fd) ) )
   (##sys#setslot t 3 'blocked)
+  (##sys#setislot t 13 #f)
   (##sys#setslot t 11 (cons fd i/o)) )
 
 (define (##sys#unblock-threads-for-i/o)
@@ -344,17 +347,18 @@ EOF
 	 [rq? (pair? ##sys#ready-queue-head)]
 	 [n (##sys#fdset-select-timeout	; we use FD_SETSIZE, but really should use max fd
 	     (or rq? to?)
-	     (if (and to? (not rq?))
+	     (if (and to? (not rq?))	; no thread was unblocked by timeout, so wait
 		 (let* ([tmo1 (caar ##sys#timeout-list)]
 			[now (##sys#fudge 16)])
 		   (fxmax 0 (- tmo1 now)) )
 		 0) ) ] )		; otherwise immediate timeout.
     (dbg n " fds ready")
-    (cond [(eq? -1 n) (##sys#force-primordial)]
+    (cond [(eq? -1 n) 
+	   (##sys#force-primordial)]
 	  [(fx> n 0)
 	   (set! ##sys#fd-list
 	     (let loop ([n n] [lst ##sys#fd-list])
-	       (if (zero? n)
+	       (if (or (zero? n) (null? lst))
 		   lst
 		   (let* ([a (car lst)]
 			  [fd (car a)]
@@ -369,13 +373,39 @@ EOF
 				 (loop (sub1 n) (cdr lst)) )
 			       (let* ([t (car threads)]
 				      [p (##sys#slot t 11)] )
-				 (when (and (pair? p) (eq? fd (car p)))
+				 (when (and (pair? p)
+					    (eq? fd (car p))
+					    (not (##sys#slot t 13) ) ) ; not unblocked by timeout
 				   (##sys#thread-basic-unblock! t) )
 				 (loop2 (cdr threads)) ) ) )
 			 (cons a (loop n (cdr lst))) ) ) ) ) ) ] )
     (##sys#fdset-restore) ) )
 
 
+;;; Clear I/O state for unblocked thread
+
+(define (##sys#clear-i/o-state-for-thread! t)
+  (when (pair? (##sys#slot t 11))
+    (let ((fd (##sys#slot (##sys#slot t 11) 0)))
+      (set! ##sys#fd-list
+	(let loop ([lst ##sys#fd-list])
+	  (if (null? lst)
+	      '()
+	      (let* ([a (##sys#slot lst 0)]
+		     [fd2 (##sys#slot a 0)] )
+		(if (eq? fd fd2)
+		    (let ((ts (##sys#delq t (##sys#slot a 1)))) ; remove from fd-list entry
+		      (cond ((null? ts)
+			     ;;(pp `(CLEAR FD: ,fd ,t) ##sys#standard-error)
+			     (##sys#fdset-clear fd) ; no more threads waiting for this fd
+			     (##sys#fdset-restore)
+			     (##sys#slot lst 1) )
+			    (else
+			     (##sys#setslot a 1 ts) ; fd-list entry is list with t removed
+			     lst) ) )
+		    (cons a (loop (##sys#slot lst 1)))))))))))
+
+
 ;;; Get list of all threads that are ready or waiting for timeout or waiting for I/O:
 
 (define (##sys#all-threads)
