@@ -16,6 +16,11 @@
  [paranoia]
  [else
   (declare
+    (no-procedure-checks-for-usual-bindings)
+    (bound-to-procedure
+     char-set char-set-complement ucs-range->char-set! ucs-range->char-set char-set-union
+     char-set-adjoin string->char-set list->char-set string-copy make-char-set char-set-copy
+     char-set? char-set-size char-set:s)
     (no-bound-checks) ) ] )
 
 (cond-expand
@@ -32,7 +37,8 @@
     (define-macro (##sys#check-port . _) '(##core#undefined))
     (define-macro (##sys#check-number . _) '(##core#undefined))
     (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
- [else] )
+ [else
+  (declare (emit-exports "srfi-14.exports"))] )
 
 (register-feature! 'srfi-14)
 
