@@ -1,6 +1,7 @@
-;;; srfi-18.scm - Simple thread unit - felix
+;;;; srfi-18.scm - Simple thread unit - felix
 ;
-; Copyright (c) 2000-2006, Felix L. Winkelmann
+; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,15 +23,6 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare
@@ -48,6 +40,7 @@
     (no-bound-checks)
     (no-procedure-checks-for-usual-bindings)
     (bound-to-procedure
+     ##sys#thread-yield!
      condition-property-accessor ##sys#tty-port? ##sys#thread-block-for-i/o thread-yield! ##sys#thread-unblock!
      ##sys#thread-basic-unblock! gensym ##sys#thread-block-for-timeout! ##sys#thread-kill!
      ##sys#thread-block-for-termination! make-thread ##sys#exact->inexact ##sys#flonum-fraction truncate
@@ -121,6 +114,11 @@ EOF
   (##sys#check-structure tm 'time 'time->seconds)
   (+ (##sys#slot tm 2) (/ (##sys#slot tm 3) 1000)) )
 
+(define (time->milliseconds tm)
+  (##sys#check-structure tm 'time 'time->milliseconds)
+  (+ (inexact->exact (* (- (##sys#slot tm 2) C_startup_time_seconds) 1000))
+     (##sys#slot tm 3) ) )
+
 (define (seconds->time n)
   (##sys#check-number n 'seconds->time)
   (let* ([n2 (max 0 (- n C_startup_time_seconds))] ; seconds since startup
@@ -128,6 +126,11 @@ EOF
 	 [n3 (inexact->exact (truncate (+ (* n2 1000) ms)))] ) ; milliseconds since startup
     (##sys#make-structure 'time n3 (truncate n) (inexact->exact ms)) ) )
 
+(define (milliseconds->time nms)
+  (##sys#check-exact nms 'milliseconds->time)
+  (let ((s (+ C_startup_time_seconds (/ nms 1000))))
+    (##sys#make-structure 'time nms s 0) ) )
+
 (define (time? x) (##sys#structure? x 'time))
 
 (define srfi-18:time? time?)
@@ -219,12 +222,7 @@ EOF
       (##sys#add-to-ready-queue thread) 
       thread) ) )
 
-(define (thread-yield!)
-  (##sys#call-with-current-continuation
-   (lambda (return)
-     (let ((ct ##sys#current-thread))
-       (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
-       (##sys#schedule) ) ) ) )
+(define thread-yield! ##sys#thread-yield!) ;In library.scm
 
 (define thread-join!
   (lambda (thread . timeout)
@@ -241,7 +239,7 @@ EOF
 	    ct 1
 	    (lambda ()
 	      (case (##sys#slot thread 3)
-		[(dead) (##sys#apply-values (##sys#slot thread 2))]
+		[(dead) (apply return (##sys#slot thread 2))]
 		[(terminated)
 		 (return 
 		  (##sys#signal
@@ -281,16 +279,16 @@ EOF
     (##sys#setslot thread 3 'ready)
     (##sys#add-to-ready-queue thread) ) )
 
-(define thread-sleep!
-  (lambda (tm)
-    (unless tm (##sys#signal-hook #:type-error 'thread-sleep! "invalid timeout argument" tm))
+(define (thread-sleep! tm)
+  (define (sleep limit loc)
     (##sys#call-with-current-continuation
      (lambda (return)
-       (let ([limit (##sys#compute-time-limit tm)]
-	     [ct ##sys#current-thread] )
+       (let ((ct ##sys#current-thread))
 	 (##sys#setslot ct 1 (lambda () (return (##core#undefined))))
 	 (##sys#thread-block-for-timeout! ct limit)
-	 (##sys#schedule) ) ) ) ) )
+	 (##sys#schedule) ) ) ) )
+  (unless tm (##sys#signal-hook #:type-error 'thread-sleep! "invalid timeout argument" tm))
+  (sleep (##sys#compute-time-limit tm) 'thread-sleep!) )
 
 
 ;;; Mutexes:
@@ -473,8 +471,6 @@ EOF
 	   (old) ) )
 	(##sys#thread-unblock! thread) ) ) )
 
-(define thread-deliver-signal! thread-signal!)
-
 
 ;;; Don't block in the repl: (by Chris Double)
 
@@ -487,3 +483,11 @@ EOF
 	  (old)
 	  (##sys#thread-block-for-i/o! ##sys#current-thread 0 #t)
 	  (thread-yield!)))) ) )
+
+
+;;; Waiting for I/O on file-descriptor
+
+(define (thread-wait-for-i/o! fd #!optional (mode #:all))
+  (##sys#check-exact fd 'thread-wait-for-i/o!)
+  (##sys#thread-block-for-i/o! ##sys#current-thread fd mode) 
+  (thread-yield!) )
