@@ -1,6 +1,7 @@
 ;;;; tcp.scm - Networking stuff
 ;
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
+; Copyright (c) 2008, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -22,25 +23,17 @@
 ; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 ; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 ; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Unter den Gleichen 1
-; 37130 Gleichen
-; Germany
 
 
 (declare
   (unit tcp)
-  (uses extras scheduler)
+  (uses data-structures ports extras scheduler)
   (usual-integrations)
   (fixnum-arithmetic)
   (no-bound-checks)
   (export tcp-close tcp-listen tcp-connect tcp-accept tcp-accept-ready? ##sys#tcp-port->fileno tcp-listener? tcp-addresses
-	  tcp-abandon-port tcp-listener-port tcp-listener-fileno tcp-port-numbers tcp-buffer-size)
+	  tcp-abandon-port tcp-listener-port tcp-listener-fileno tcp-port-numbers tcp-buffer-size
+	  tcp-read-timeout tcp-write-timeout tcp-accept-timeout tcp-connect-timeout)
   (no-procedure-checks-for-usual-bindings)
   (bound-to-procedure
    ##net#socket ##net#bind ##net#connect ##net#listen ##net#accept make-parameter ##sys#string-append ##sys#tcp-port->fileno
@@ -105,10 +98,9 @@ EOF
     (define-macro (##sys#check-symbol x) '(##core#undefined))
     (define-macro (##sys#check-string x) '(##core#undefined))
     (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x) '(##core#undefined))
-    (define-macro (##sys#check-port x) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector x) '(##core#undefined)) ) )
+    (define-macro (##sys#check-exact x . _) '(##core#undefined))
+    (define-macro (##sys#check-port x . _) '(##core#undefined))
+    (define-macro (##sys#check-number x) '(##core#undefined))))
  (else
   (declare (emit-exports "tcp.exports"))) )
 
@@ -135,11 +127,14 @@ EOF
 (define ##net#listen (foreign-lambda int "listen" int int))
 (define ##net#accept (foreign-lambda int "accept" int c-pointer c-pointer))
 (define ##net#close (foreign-lambda int "closesocket" int))
-(define ##net#send (foreign-lambda int "send" int scheme-pointer int int))
 (define ##net#recv (foreign-lambda int "recv" int scheme-pointer int int))
 (define ##net#shutdown (foreign-lambda int "shutdown" int int))
 (define ##net#connect (foreign-lambda int "connect" int scheme-pointer int))
 
+(define ##net#send
+  (foreign-lambda* int ((int s) (scheme-pointer msg) (int offset) (int len) (int flags))
+		   "return(send(s, (char *)msg+offset, len, flags));"))
+
 (define ##net#make-nonblocking
   (foreign-lambda* bool ((int fd))
     "int val = fcntl(fd, F_GETFL, 0);"
@@ -273,6 +268,11 @@ EOF
 
 (define (##net#bind-socket port style host)
   (##sys#check-exact port)
+  (cond-expand
+   (unsafe)
+   (else
+    (when (or (fx< port 0) (fx>= port 65535))
+      (##sys#signal-hook #:domain-error 'tcp-listen "invalid port number" port) ) ) )
   (let ((s (##net#socket _af_inet style 0)))
     (when (eq? _invalid_socket s)
       (##sys#update-errno)
@@ -322,6 +322,20 @@ EOF
 (define-constant +output-chunk-size+ 8192)
 
 (define tcp-buffer-size (make-parameter #f))
+(define tcp-read-timeout)
+(define tcp-write-timeout)
+(define tcp-connect-timeout)
+(define tcp-accept-timeout)
+
+(let ()
+  (define ((check loc) x)
+    (when x (##sys#check-exact x loc))
+    x)
+  (define minute (* 60 1000))
+  (set! tcp-read-timeout (make-parameter minute (check 'tcp-read-timeout)))
+  (set! tcp-write-timeout (make-parameter minute (check 'tcp-write-timeout))) 
+  (set! tcp-connect-timeout (make-parameter #f (check 'tcp-connect-timeout))) 
+  (set! tcp-accept-timeout (make-parameter #f (check 'tcp-accept-timeout))) )
 
 (define ##net#io-ports
   (let ((make-input-port make-input-port)
@@ -340,14 +354,24 @@ EOF
 	     (oclosed #f)
 	     (outbufsize (tbs))
 	     (outbuf (and outbufsize (fx> outbufsize 0) ""))
+	     (tmr (tcp-read-timeout))
+	     (tmw (tcp-write-timeout))
 	     (read-input
 	      (lambda ()
 		(let loop ()
 		  (let ((n (##net#recv fd buf +input-buffer-size+ 0)))
 		    (cond ((eq? -1 n)
 			   (cond ((eq? errno _ewouldblock) 
+				  (when tmr
+				    (##sys#thread-block-for-timeout! 
+				     ##sys#current-thread
+				     (fx+ (##sys#fudge 16) tmr) ) )
 				  (##sys#thread-block-for-i/o! ##sys#current-thread fd #t)
 				  (yield)
+				  (when (##sys#slot ##sys#current-thread 13)
+				    (##sys#signal-hook
+				     #:network-error
+				     "read operation timed out" fd) )
 				  (loop) )
 				 (else
 				  (##sys#update-errno)
@@ -429,22 +453,31 @@ EOF
 			      #!eof) ) ) ) ) ) )
 	     (output
 	      (lambda (s)
-		(let loop ((len (##sys#size s)) 
-			   (s s))
+		(let loop ((len (##sys#size s))
+			   (offset 0))
 		  (let* ((count (fxmin +output-chunk-size+ len))
-			 (n (##net#send fd s count 0)) )
+			 (n (##net#send fd s offset count 0)) )
 		    (cond ((eq? -1 n)
-			   (if (eq? errno _ewouldblock)
-			       (begin
-				 (##sys#thread-block-for-i/o! ##sys#current-thread fd #f)
-				 (yield) 
-				 (loop len s) )
-			       (begin
-				 (##sys#update-errno)
-				 (##sys#signal-hook 
-				  #:network-error (##sys#string-append "can not write to socket - " strerror) fd) ) ) )
+			   (cond ((eq? errno _ewouldblock)
+				  (when tmw
+				    (##sys#thread-block-for-timeout! 
+				     ##sys#current-thread
+				     (fx+ (##sys#fudge 16) tmw) ) )
+				  (##sys#thread-block-for-i/o! ##sys#current-thread fd #f)
+				  (yield) 
+				  (when (##sys#slot ##sys#current-thread 13)
+				    (##sys#signal-hook
+				     #:network-error
+				     "write operation timed out" fd) )
+				  (loop len offset) )
+				 (else
+				  (##sys#update-errno)
+				  (##sys#signal-hook 
+				   #:network-error
+				   (##sys#string-append "can not write to socket - " strerror) 
+				   fd) ) ) )
 			  ((fx< n len)
-			   (loop (fx- len n) (##sys#substring s n (##sys#size s))) ) ) ) ) ) )
+			   (loop (fx- len n) (fx+ offset n)) ) ) ) ) ) )
 	     (out
 	      (make-output-port
 	       (if outbuf
@@ -476,13 +509,14 @@ EOF
 	(##sys#setslot out 3 "(tcp)")
 	(##sys#setslot in 7 'socket)
 	(##sys#setslot out 7 'socket)
-	(##sys#setslot (##sys#port-data in) 0 data)
-	(##sys#setslot (##sys#port-data out) 0 data)
+	(##sys#setslot in 9 data)
+	(##sys#setslot out 9 data)
 	(values in out) ) ) ) )
 
 (define (tcp-accept tcpl)
   (##sys#check-structure tcpl 'tcp-listener)
-  (let ((fd (##sys#slot tcpl 1)))
+  (let ((fd (##sys#slot tcpl 1))
+	(tma (tcp-accept-timeout)))
     (let loop ()
       (if (eq? 1 (##net#select fd))
 	  (let ((fd (##net#accept fd #f #f)))
@@ -493,8 +527,17 @@ EOF
 	       tcpl) )
 	    (##net#io-ports fd) )
 	  (begin
+	    (when tma
+	      (##sys#thread-block-for-timeout! 
+	       ##sys#current-thread
+	       (fx+ (##sys#fudge 16) tma) ) )
 	    (##sys#thread-block-for-i/o! ##sys#current-thread fd #t)
 	    (yield)
+	    (when (##sys#slot ##sys#current-thread 13)
+	      (##sys#signal-hook
+	       #:network-error
+	       'tcp-accept
+	       "accept operation timed out" fd) )
 	    (loop) ) ) ) ) )
 
 (define (tcp-accept-ready? tcpl)
@@ -509,16 +552,17 @@ EOF
 
 (define get-socket-error
   (foreign-lambda* int ((int socket))
-		   "int err, optlen;"
-		   "optlen = sizeof(err);"
-		   "if (typecorrect_getsockopt(socket, SOL_SOCKET, SO_ERROR, &err, (socklen_t *)&optlen) == -1)"
-		   "return(-1);"
-		   "return(err);"))
+    "int err, optlen;"
+    "optlen = sizeof(err);"
+    "if (typecorrect_getsockopt(socket, SOL_SOCKET, SO_ERROR, &err, (socklen_t *)&optlen) == -1)"
+    "return(-1);"
+    "return(err);"))
 
 (define general-strerror (foreign-lambda c-string "strerror" int))
 
 (define (tcp-connect host . more)
-  (let ((port (:optional more #f)))
+  (let ((port (:optional more #f))
+	(tmc (tcp-connect-timeout)))
     (##sys#check-string host)
     (unless port
       (set!-values (host port) (##net#parse-host host "tcp"))
@@ -546,8 +590,17 @@ EOF
 	      (let ((f (##net#select-write s)))
 		(when (eq? f -1) (fail))
 		(unless (eq? f 1)
-		  ;(##sys#thread-block-for-i/o! ##sys#current-thread s #t)
+		  (when tmc
+		    (##sys#thread-block-for-timeout!
+		     ##sys#current-thread
+		     (fx+ (##sys#fudge 16) tmc) ) )
+		  (##sys#thread-block-for-i/o! ##sys#current-thread s #:all)
 		  (yield)
+		  (when (##sys#slot ##sys#current-thread 13)
+		    (##sys#signal-hook
+		     #:network-error
+		     'tcp-connect
+		     "connect operation timed out" s) )
 		  (loop) ) ) )
 	    (fail) ) )
       (let ((err (get-socket-error s)))
@@ -558,14 +611,7 @@ EOF
       (##net#io-ports s) ) ) )
 
 (define (##sys#tcp-port->fileno p)
-  (##sys#slot (##sys#tcp-port-data p) 0) )
-
-(define (##sys#tcp-port-data p)
-  (##sys#check-port p)
-  (let ((d (##sys#port-data p)))
-    (if d
-	(##sys#slot d 0)
-	(##sys#signal-hook #:type-error "bad argument type - not a TCP port - " p) ) ) )
+  (##sys#slot (##sys#port-data p) 0) )
 
 (define (tcp-addresses p)
   (let ((fd (##sys#tcp-port->fileno p)))
@@ -594,10 +640,11 @@ EOF
     port) )
 
 (define (tcp-abandon-port p)
+  (##sys#check-port p 'tcp-abandon-port)
   (##sys#setislot
-   (##sys#tcp-port-data p)
+   (##sys#port-data p)
    (if (##sys#slot p 1) 2 1)
-   #t) ) 
+   #t) )
 
 (define (tcp-listener-fileno l)
   (##sys#check-structure l 'tcp-listener 'tcp-listener-fileno)
